
/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/
#include "seamless3d.h"

void inline sweepColorY(const M_Triangle& triangle, M_Color3f &color, const Color3f beginColor,const Color3f endColor, float beginOpacity, float endOpacity, float beginHeight, float endHeight, M_Int& mask){
    float opacityDiff=endOpacity-beginOpacity;
    float rColorDiff=endColor.r-beginColor.r;
    float gColorDiff=endColor.g-beginColor.g;
    float bColorDiff=endColor.b-beginColor.b;
    float heightDiff=endHeight-beginHeight;
    for(int i = triangle.getTriangleLength(); --i >= 0;){
        //if(triangle.refPart(i)==part){
        int ci=triangle.refTexIndex(i);
        if(mask[ci]==0){
            Art* prt = triangle.refPart(i);
            Vec3f vec=prt->matrix * prt->coord[triangle.refIndex(i)];
            float length=vec.y;
            if(length>=beginHeight && length<=endHeight){

                float frac=(length-beginHeight)/heightDiff;
                float opacity=(beginOpacity+frac*opacityDiff);
                float clarity=1.0-opacity;

                color[ci]=Color3f(
                    color[ci].r*clarity+opacity*(beginColor.r+frac*rColorDiff),
                    color[ci].g*clarity+opacity*(beginColor.g+frac*gColorDiff),
                    color[ci].b*clarity+opacity*(beginColor.b+frac*bColorDiff)
                    );
                mask[ci]=1;
            }
        }
        //}
    }
}
//void sweepColorSpherical(float beginHeight, float endHeight, Color3f beginColor,Color3f endColor){
void inline sweepColorSpherical(const Art* part, const M_Triangle& triangle, M_Color3f &color, const Color3f beginColor,const Color3f endColor, float beginOpacity, float endOpacity, float beginHeight, float endHeight, M_Int& mask){

    float opacityDiff=endOpacity-beginOpacity;
    float rColorDiff=endColor.r-beginColor.r;
    float gColorDiff=endColor.g-beginColor.g;
    float bColorDiff=endColor.b-beginColor.b;
    float heightDiff=endHeight-beginHeight;
    for(int i=triangle.length*3;--i>=0;){
        if(triangle.refPart(i)==part){
            int ci=triangle.refTexIndex(i);
            if(mask[ci]==0){
                Art* prt = triangle.refPart(i);
                Vec3f vec=prt->matrix * prt->coord[triangle.refIndex(i)];

                float length = vec.length();
                if(length >= beginHeight && length <= endHeight){
                    float frac=(length-beginHeight)/heightDiff;
                    float opacity=(beginOpacity+frac*opacityDiff);
                    float clarity=1.0-opacity;
                    color[ci]=Color3f(
                        color[ci].r*clarity+opacity*(beginColor.r+frac*rColorDiff),
                        color[ci].g*clarity+opacity*(beginColor.g+frac*gColorDiff),
                        color[ci].b*clarity+opacity*(beginColor.b+frac*bColorDiff)
                        );
                    mask[ci]=1;
                }
            }
        }
    }
}


void ColorSweep::doBuild(){
    static M_Int mask;

    if(part.length==0||color.length==0)return;

    M_Float& opacity=this->opacity.refValue();
    const M_Color3f& color=this->color.getValue();

    Art* prt=part.refValue()[0];
    ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(prt->workEffect);
    if(colorEffect==0)return;
    SeamlessPtr smls=prt->smls;
    mask.length=0;
    mask.length=colorEffect->color.length;

    Matrix parentMatrix=translationMatrix(-location) * rotationMatrix(-orientation);

    for(int k=0; k<smls->skeleton.length;k++){
        smls->skeleton.refValue()[k]->catMatrix(parentMatrix);
    }

    if(opacity.length < color.length){
        opacity.length=color.length;
        for(int i=0; i<opacity.length; i++)opacity[i]=1;
    }
    for(int k=0; k<part.length; k++){
        Art* prt=part.refValue()[k];

        float posFrom=0;
        float posTo=distance.refValue()[0];

        for(int i=0;;){
            M_Color3f& eColor=dynamic_cast<ColorEffect*>(prt->workEffect)->color;
            if(straight){
                sweepColorY(prt->triangle, eColor, color[i], color[i+1], opacity[i], opacity[i+1], posFrom, posTo,mask);
            }
            else if(spherical){
                sweepColorSpherical(prt, prt->triangle, eColor, color[i], color[i+1], opacity[i], opacity[i+1], posFrom, posTo,mask);
            }

            if(i>=distance.length-1)break;


            posFrom=posTo;
            i++;
            posTo+=distance.refValue()[i];

        }
    }
    smls->catMatrix();
}
void inline sweepColorY(const M_Vec3f &coord, Matrix& matrix, M_Color3f &color, const Color3f beginColor,const Color3f endColor, float beginOpacity, float endOpacity, float beginHeight, float endHeight, const int h, const int texCoordBase){
    float opacityDiff=endOpacity-beginOpacity;
    float rColorDiff=endColor.r-beginColor.r;
    float gColorDiff=endColor.g-beginColor.g;
    float bColorDiff=endColor.b-beginColor.b;
    float heightDiff=endHeight-beginHeight;
    int coordLen = coord.length;
    for(int i = h, j = texCoordBase; i < coordLen; i++, j++){
        Vec3f vec=matrix*coord[i];
        float length=vec.y;
        if(length>=beginHeight && length<=endHeight){

            float frac=(length-beginHeight)/heightDiff;
            float opacity=(beginOpacity+frac*opacityDiff);
            float clarity=1.0-opacity;

            color[j]=Color3f(
                color[j].r*clarity+opacity*(beginColor.r+frac*rColorDiff),
                color[j].g*clarity+opacity*(beginColor.g+frac*gColorDiff),
                color[j].b*clarity+opacity*(beginColor.b+frac*bColorDiff)
            );
        }
    }
}
void inline sweepColorSpherical(const M_Vec3f &coord, M_Color3f &color, const Color3f beginColor,const Color3f endColor, float beginOpacity, float endOpacity, float beginHeight, float endHeight, const int h, const int texCoordBase){
    float opacityDiff=endOpacity-beginOpacity;
    float rColorDiff=endColor.r-beginColor.r;
    float gColorDiff=endColor.g-beginColor.g;
    float bColorDiff=endColor.b-beginColor.b;
    float heightDiff=endHeight-beginHeight;
    int coordLen = coord.length;
    for(int i = h, j = texCoordBase; i < coordLen; i++, j++){
        float length = coord[i].length();
        if(length>=beginHeight && length<=endHeight){
            float frac=(length-beginHeight)/heightDiff;
            float opacity=(beginOpacity+frac*opacityDiff);
            float clarity=1.0-opacity;
            color[j]=Color3f(
                color[j].r*clarity+opacity*(beginColor.r+frac*rColorDiff),
                color[j].g*clarity+opacity*(beginColor.g+frac*gColorDiff),
                color[j].b*clarity+opacity*(beginColor.b+frac*bColorDiff)
                );
        }
    }
}

void ColorSweep::doColorForStem(Art* prt, M_Vec3f& coord, const int h, const int texCoordBase){//why did we delete this function once?
    if(color.length==0)return;

    M_Float& opacity=this->opacity.refValue();
    const M_Color3f& color=this->color.getValue();

    ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(prt->workEffect);
    if(colorEffect==0)return;
    M_Color3f& eColor=dynamic_cast<ColorEffect*>(prt->workEffect)->color;

    if(opacity.length < color.length){
        opacity.length=color.length;
        for(int i=0; i<opacity.length; i++)opacity[i]=1;
    }
    {
        float posFrom=0;
        float posTo=distance.refValue()[0];
        Matrix parentMatrix=translationMatrix(-location) * rotationMatrix(-orientation);

        for(int i=0;;){
            if(straight){
                sweepColorY(coord, parentMatrix, eColor, color[i], color[i+1], opacity[i], opacity[i+1], posFrom, posTo, h, texCoordBase);
            }
            else if(spherical){
                sweepColorSpherical(coord, eColor, color[i], color[i+1], opacity[i], opacity[i+1], posFrom, posTo, h, texCoordBase);
            }

            if(i>=distance.length-1)break;


            posFrom=posTo;
            i++;
            posTo+=distance.refValue()[i];

        }
    }
}

void ColorSweep::processFields(){
    matrix=rotationMatrix(orientation)*translationMatrix(location);
    imatrix=translationMatrix(-location)*rotationMatrix(-orientation);
    if(!(spherical || straight)){
        straight.refValue()=true;
    }
}

