/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include "seamless3d.h"


void addRelevantPart(M_Art& relevant, Art*& part);
void removeNonRelevantParts(Art* source);
bool commonIllegalImmediateRef(const M_SArt& m_art, const Art* part, M_Art& accumOthers);

void setUpLookUp(Art* part){
    Effect* effect=part->workEffect;
    M_Int& lookup=effect->lookup;
    if(effect->getEffectLength()>=lookup.length){
        int oldLen=lookup.length;
        lookup.length=effect->getEffectLength();
        for(int i=lookup.length; --i>=oldLen;){
            lookup[i]=-1;
        }
    }
}

bool verifyPartRelivanceToTriangles(Art* source, M_Triangle& triangle){
    for(int i=triangle.length; --i>=0; ){
        S_Triangle tri=triangle[i];
        if(	tri.pa==source || tri.pb==source ||	tri.pc==source){
            return true;
        }
    }
    return false;
}

inline void removeNonRelevant(Art* source, M_Art& relevantPart){
    for(int k=relevantPart.length; --k>=0; ){
        if(relevantPart[k]==source){
            int relLen=relevantPart.length-1;
            for(;k<relLen; k++){
                relevantPart[k]=relevantPart[k+1];
            }
            relevantPart.length--;
            return;
        }
    }
    throw source;
}
void addRelevantPart(M_Art& relevant, Art*& part){
    for(int r=relevant.length; --r>=0;){
        if(relevant[r]==part)return;
    }
    relevant+=part;
}
bool partAllreadyIsRelevant(M_Art& relevant, Art*& part){
    for(int r=relevant.length; --r>=0;){
        if(relevant[r]==part)return true;
    }
    return false;;
}

// remove all parts that dont reference source from source reletive list
void removeNonRelevantParts(Art* source){
    M_Art& reletivePart=source->relevantPart;

    for(int k=reletivePart.length;--k>=0;){
        Art* part=reletivePart[k];
        if(verifyPartRelivanceToTriangles(source, part->triangle)==false){
            removeNonRelevant(part, reletivePart); //remove us from some one's relevant list
        }
    }
}

inline bool checkForFastAddSelfCopy(Art& part, M_Triangle& triangle, M_Int& destMap, M_Art& partList){
    int i;

    int markTriangle=triangle.length;
    triangle.length+=triangle.length;

    for(i=markTriangle*3; --i>=0;){
        triangle.refIndex(i+markTriangle*3) = triangle.refIndex(i)+triangle.refPart(i)->coord.length;
        int h=destMap[triangle.refPart(i)->plidx];
        if(h<0){
            triangle.length=markTriangle;
            return false;
        }
        triangle.refPart(i+markTriangle*3) = partList[h];
    }
    for(i=part.relevantPart.length; --i>=0;){
        if(destMap[part.relevantPart[i]->plidx]<0){ //***********do not remove
            triangle.length=markTriangle;
            return false;
        }
    }

    return true;
}


void commonPasteAddCopyOfMirror(CopyPart* cp, Art* part, M_Int& destMap, M_Art& partList){

    const M_Triangle& srcTriangle=part->triangle;
    M_Triangle& triangle=part->triangle;
///
    if(checkForFastAddSelfCopy(*part, triangle, destMap, partList)){
        cp->addPart(part, part, false);

        removeNonRelevantParts(part);
    }
    else {
        M_Art& partList=part->smls->partList;
        static ArtPtr temp = new Part;
        temp->workEffect=part->workEffect;
        temp->smls=part->smls;
        temp->plidx=partList.length;
        partList+=temp;
        Mesh mesh;
        part->extractGeometryGS(mesh, g_tmpIndex);
        temp->workEffect->convertGeomSourceToPart(temp, mesh);
        cp->source.refValue()+=temp;
        cp->dest.refValue()+=part;
        cp->doBuild();
        partList.length--;
    }
}

void commonReplaceWithCopyOfRememberedPart(CopyPart& cp, Art* src, Art* part){
    part->deleteGeometry();
    Mesh mesh;
    src->extractGeometryGS(mesh, g_tmpIndex);
    part->workEffect->convertGeomSourceToPart(part, mesh);
    Vec3f translation=src->translation;
    if(cp.mirrorX){
        translation.x=-part->translation.refValue().x;
    }
    if(cp.mirrorY){
        translation.y=-part->translation.refValue().y;
    }
    if(cp.mirrorZ){
        translation.z=-part->translation.refValue().z;
    }
    part->translation.refValue()=translation;
    cp.source=M_SArt(part);
    cp.dest=M_SArt(part);
    cp.doBuild();
}

inline bool checkForFastCopy(Art& dst, Art& src, const M_Triangle& srcTriangle, M_Triangle& dstTriangle, M_Int& destMap, M_Art& partList){
    int i;

    dstTriangle.length=srcTriangle.length;

    for(i=dstTriangle.length*3; --i>=0;){
        dstTriangle.refIndex(i) = srcTriangle.refIndex(i);
        int h=destMap[srcTriangle.refPart(i)->plidx];
        if(h<0){
            return false;
        }
        dstTriangle.refPart(i) = partList[h];
    }
    for(i=src.relevantPart.length; --i>=0;){
        if(destMap[src.relevantPart[i]->plidx]<0){ //***********do not remove
            return false;
        }
    }
    dst.relevantPart.length=src.relevantPart.length;
    for(i=src.relevantPart.length; --i>=0;){
        dst.relevantPart[i]=partList[destMap[src.relevantPart[i]->plidx]];
    }
    return true;
}

void CopyPart::addPart(Art* src,Art* dst, bool copyTriangles){
    if(copyTriangles){
        //dst->relevantPart=M_Art(dst);
        addRelevantPart(dst->relevantPart, dst);
    }

    const M_Vec3f& srcCoord=src->coord;
    const M_Vec3f& srcNormal=src->normal;
    const M_Triangle& srcTriangle=src->triangle;

    M_Vec3f& dstCoord=dst->coord.refValue();
    M_Vec3f& dstNormal=dst->normal.refValue();
    M_Triangle& dstTriangle=dst->triangle;

    int srcCoordLen=srcCoord.length;
    int srcNormalLen=srcNormal.length;
    int srcTriangleLen=srcTriangle.length;

    int markCoord=dstCoord.length;
    int markNormal=dstNormal.length;
    int markTriangle=dstTriangle.length*3;
    if(!copyTriangles){
        markTriangle/=2;
    }


    int edif=0;
/*
    if(src->workEffect->getType()==TEXTUREEFFECT){
        M_Vec2f& texCoord=TextureEffectPtr(src->workEffect)->texCoord.refValue();
        int srcFromEffectCoord=src->fromEffectCoord;
        int srcToEffectCoord=src->toEffectCoord;
        int length=srcToEffectCoord-srcFromEffectCoord;
        edif=texCoord.length-srcFromEffectCoord;
        int d=dst->fromEffectCoord=texCoord.length;
        texCoord.addLength(length);
        dst->toEffectCoord+=texCoord.length;

        for(int s=srcFromEffectCoord; s<srcToEffectCoord; s++, d++){
            texCoord[d]=texCoord[s];
        }
    }
    else if(src->workEffect->getType()==COLOREFFECT){
        M_Color3f& color=ColorEffectPtr(src->workEffect)->color.refValue();
        int srcFromEffectCoord=src->fromEffectCoord;
        int srcToEffectCoord=src->toEffectCoord;
        int length=srcToEffectCoord-srcFromEffectCoord;
        edif=color.length-srcFromEffectCoord;
        int d=dst->fromEffectCoord=color.length;
        color.addLength(length);
        dst->toEffectCoord=color.length;

        for(int s=srcFromEffectCoord; s<srcToEffectCoord; s++, d++){
            color[d]=color[s];
        }
    }
    else throw this;
    */
    dstCoord.length+=srcCoordLen;
    for(int i=0; i<srcCoordLen;i++){
        dstCoord[markCoord+i] = srcCoord[i];
    }
    if(copyTriangles){
        dstTriangle.length+=srcTriangleLen;
        for(int i=0; i<srcTriangleLen*3; i++){;
            dstTriangle.refIndex(markTriangle+i) = srcTriangle.refIndex(i)+markCoord;
            dstTriangle.refPart(markTriangle+i) = dst;
        }
    }
    src->workEffect->copyAddColor(src, dst, srcTriangle, dstTriangle,markTriangle);

//	if(staticTransform==true){
    if(scale != Vec3f(1, 1, 1) || rotation != Rotation() || translation != Vec3f()){
        Matrix matrix;
        matrix.setTransform( dst->translation, dst->rotation, dst->scale);
        for(int i=dstCoord.length; --i>=markCoord;){;
            dstCoord[i] = matrix * dstCoord[i];
        }
    }
//	}

    if(dst->smls->surfaceNormals){
        dstNormal.length+=srcNormalLen;
        for(int i=0; i<srcNormalLen; i++){
            dstNormal[markNormal+i] = srcNormal[i];
        }
    }
    if(copyTransform==true){//////to be fixed
        if(staticTransform==true){
            dst->translation=Vec3f();
            dst->rotation=Rotation();
            dst->scale=Vec3f(1,1,1);
        }
        else {
            dst->translation	= src->translation;
            dst->rotation	= src->rotation;
            dst->scale		= src->scale;
        }
    }
    if(stemNurbsMirror){
        for(int i=0; i<srcCoordLen;i++){
            dstCoord[markCoord+i].x = -(dstCoord[markCoord+i].x - 0.5) + 0.5;
        }
        dst->insideOut(markTriangle/3);
    }
    if(mirrorX){
        dst->mirrorFuncX(markTriangle,markCoord);
    }
    if(mirrorY){
        dst->mirrorFuncY(markTriangle,markCoord);
    }
    if(mirrorZ){
        dst->mirrorFuncZ(markTriangle,markCoord);
    }
    if(insideOut){
        dst->insideOut(markTriangle/3);
    }
    if(scale != Vec3f(1, 1, 1) || rotation != Rotation() || translation != Vec3f()){
        Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
        dst->staticTransform(markCoord, matrix);
    }
    if(resetTransform){
        dst->translation=Vec3f();
        dst->rotation=Rotation();
        dst->scale=Vec3f(1,1,1);
    }
}

void CopyPart::doBuild(){

    static CopyPart cp;
    M_SArt& dest = this->dest.refValue();
    M_SArt* sourcePtr = &this->source.refValue();
    if(sourcePtr->length == 0){
        sourcePtr = &dest;
    }
    M_SArt& source=*sourcePtr;


    int length=source.length;
    if(length>dest.length){
        length=dest.length;
    }
    if(!length)return;
    M_Art& partList=source[0]->smls->partList;
    static M_Int destMap;

    {

        destMap.length=partList.length;
        {
            for(int k=destMap.length; --k>=0;){
                destMap[k]=-1;
            }
        }
        {
            for(int k=length; --k>=0;){
                Art* src=source[k];
                Art* dst=dest[k];

                destMap[src->plidx]=dst->plidx;
            }
        }
    }

    for(int k=0; k<length;k++){
        Art* src=source[k];
        Art* dst=dest[k];

        setUpLookUp(dst);
        if(src==dst){
            if(add){
                cp.source.refValue().length=0;
                cp.dest.refValue().length=0;
                cp.add=add;
                cp.mirrorX = mirrorX;
                cp.mirrorY = mirrorY;
                cp.mirrorZ = mirrorZ;
                cp.insideOut = insideOut;
                cp.stemNurbsMirror = stemNurbsMirror;
                cp.scale = scale;
                cp.rotation = rotation;
                cp.translation = translation;
                commonPasteAddCopyOfMirror(&cp, dst, destMap, partList);
                cp.stemNurbsMirror=false;
            }
            else {
sharedVertices:
                if(src!=dst){
                    cp.source.refValue().length=0;
                    cp.dest.refValue().length=0;
                    cp.add=add;
                    cp.mirrorX=mirrorX;
                    cp.mirrorY=mirrorY;
                    cp.mirrorZ=mirrorZ;
                    cp.insideOut=insideOut;
                    cp.copyTransform=copyTransform;
                    commonReplaceWithCopyOfRememberedPart(cp, src,dst);
                }
                else {
                    if(mirrorX){
                        dst->mirrorFuncX();
                    }
                    if(mirrorY){
                        dst->mirrorFuncY();
                    }
                    if(mirrorZ){
                        dst->mirrorFuncZ();
                    }
                    if(insideOut==true){
                        dst->insideOut();
                    }
                    if(scale != Vec3f(1, 1, 1) || rotation != Rotation() || translation != Vec3f()){
                        Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
                        dst->staticTransform(0, matrix);
                    }
                    if(resetTransform==true){
                        dst->translation=Vec3f();
                        dst->rotation=Rotation();
                        dst->scale=Vec3f(1,1,1);
                    }
                }
            }
        }
        else if(add){
            addPart(src,dst, true);
        }
        else{
            int i;
            const M_Triangle& srcTriangle=src->triangle;
            M_Triangle& dstTriangle=dst->triangle;
///
            if(checkForFastCopy(*dst, *src, srcTriangle, dstTriangle, destMap, partList)){


                const M_Vec3f& srcCoord=src->coord;
                const M_Vec3f& srcNormal=src->normal;

                M_Vec3f& dstCoord=dst->coord.refValue();
                M_Vec3f& dstNormal=dst->normal.refValue();

                int edif=0;

                dstCoord.length=srcCoord.length;
                for(i=dstCoord.length; --i>=0;){
                    dstCoord[i] = srcCoord[i];
                }

                src->workEffect->copyColor(src, dst, srcTriangle, dstTriangle);

                if(staticTransform==true){
                    Matrix matrix;
                    matrix.setTransform( dst->translation, dst->rotation, dst->scale);
                    for(i=dstCoord.length; --i>=0;){
                        dstCoord[i] = matrix * dstCoord[i];
                    }
                }

                dstNormal.length=srcNormal.length;
                for(i=dstNormal.length; --i>=0;){
                    dstNormal[i] = srcNormal[i];
                }
                if(copyTransform==true){//////to be fixed
                    if(staticTransform==true){
                        dst->translation=Vec3f();
                        dst->rotation=Rotation();
                        dst->scale=Vec3f(1,1,1);
                    }
                    else {
                        dst->translation	= src->translation;
                        dst->rotation	= src->rotation;
                        dst->scale		= src->scale;
                    }
                }
                if(mirrorX==true){
                    dst->mirrorFuncX();
                    dst->translation.refValue().x=-src->translation.refValue().x;
                    dst->translation.refValue().y=src->translation.refValue().y;
                    dst->translation.refValue().z=src->translation.refValue().z;
                }
                if(mirrorY==true){
                    dst->mirrorFuncY();
                    dst->translation.refValue().x=src->translation.refValue().x;
                    dst->translation.refValue().y=-src->translation.refValue().y;
                    dst->translation.refValue().z=src->translation.refValue().z;
                }
                if(mirrorZ==true){
                    dst->mirrorFuncZ();
                    dst->translation.refValue().x=src->translation.refValue().x;
                    dst->translation.refValue().y=src->translation.refValue().y;
                    dst->translation.refValue().z=-src->translation.refValue().z;
                }
                if(insideOut==true){
                    dst->insideOut();
                }
                if(resetTransform==true){
                    dst->translation=Vec3f();
                    dst->rotation=Rotation();
                    dst->scale=Vec3f(1,1,1);
                }
                if(scale != Vec3f(1, 1, 1) || rotation != Rotation() || translation != Vec3f()){
                    Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
                    dst->staticTransform(0, matrix);
                }
                removeNonRelevantParts(dst);
            }
            else{
                goto sharedVertices;
            }
        }
    }
}

bool CopyPart::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
    return commonIllegalImmediateRef(dest, p, accumOthers);
}
