/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/


#include <windows.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "robot_aid.h"
#include "resource.h"


Ptr<Menu> g_selAnimMenu = 0;

extern const BIProto* builtInProtoList[];
extern const char* tip_toggleWireframe;
extern const char* tip_resetViewPosition;
extern const char* tip_resetViewOrientation;
extern const char* tip_centerViewToControlPoint;
extern const char* tip_centerViewToPivotPoint;
extern const char* tip_findThinTriangle;
extern const char* tip_zeroToX;
extern const char* tip_build_sphere_for_selected_part;
extern const char* tip_build_cylinder_for_selected_part;
extern const char* tip_build_cone_for_selected_part;
extern const char* tip_build_torus_for_selected_part;
extern const char* tip_build_bezier_lathe_for_selected_part;
extern const char* tip_build_box_for_selected_part;
extern const char* tip_build_rectangle_for_selected_part;
extern const char* tip_undo;
extern const char* tip_redo;

extern const char* menu_help;
extern const char* menu_tutorials;
extern const char* menu_about_seamless3d;
extern const char* menu_beeline_anim_demo;
extern const char* menu_poly_radial_tug_demo;
extern const char* menu_nurbs_lathe_avatar_demo;
extern const char* menu_fuse_surface_avatar_demo;
extern const char* menu_switch_anim_in_3d_demo;
extern const char* menu_immediate_avatar_demo;
extern const char* menu_bring_static_model_to_life_demo;
extern const char* menu_worm_animation_demo;
extern const char* menu_textured_object_demo;

extern const char* menu_file;
extern const char* menu_save;
extern const char* menu_save_as;
extern const char* menu_open;
extern const char* menu_new;
extern const char* menu_open_from_web;
extern const char* menu_import_to_Part;
extern const char* menu_export_from_Part;

extern const char* menu_import_bxx_av_studio;

extern const char* menu_toggle_background_color;
extern const char* menu_perspective;
extern const char* menu_copy;
extern const char* menu_paste;
extern const char* menu_coord;
extern const char* menu_join_move;
extern const char* menu_move_vertex;
extern const char* menu_mirrorX;
extern const char* menu_join_vertex_keep_dest_color;
extern const char* menu_join_vertex_keep_both_colors;
extern const char* menu_join_vertex_keep_source_color;
extern const char* menu_copy_poly_map_to_source;
extern const char* menu_set_x_to_zero;
extern const char* menu_color;
extern const char* menu_vertexColor;
extern const char* menu_triangleColor;
extern const char* menu_join;
extern const char* menu_join_matching_in_all_parts;
extern const char* menu_partColor;
extern const char* menu_vertices_colored;
extern const char* menu_vertices_in_all_parts_colored;
extern const char* menu_ownership;
extern const char* menu_vertexOwnership;
extern const char* menu_part;
extern const char* menu_close_vertices;
extern const char* menu_close_vertices_then_join;
extern const char* menu_part_mirrorX;
extern const char* menu_part_mirrorY;
extern const char* menu_part_mirrorZ;
extern const char* menu_add_part;
extern const char* menu_add_part_mirrorX;
extern const char* menu_add_part_mirrorY;
extern const char* menu_add_part_mirrorZ;
extern const char* menu_immediate_triangle_for_selected_part;
extern const char* menu_scene_tree;
extern const char* menu_open_scene_tree;
extern const char* menu_reset_view_position_and_orientation;
extern const char* menu_cursor_view_control;

extern const char* menu_break_join;
extern const char* menu_break_coord;
extern const char* menu_break_color;
extern const char* menu_delete;
extern const char* menu_vertex;
extern const char* menu_triangleDelete;
extern const char* menu_triangles_right_from_selected_coord;
extern const char* menu_triangles_left_from_selected_coord;
extern const char* menu_triangles_up_from_selected_coord;
extern const char* menu_triangles_down_from_selected_coord;
extern const char* menu_layered;
extern const char* menu_all_triangles_in_part;
extern const char* menu_mirror_triangle;
extern const char* menu_split_triangle;
extern const char* menu_swap_triangle_divide;
extern const char* menu_compile_transform;
const char* menu_rotate_uv_90 = "rotate UV 90";

const char* menu_convert_to_0_u_mid_points = "0 U mid points";
const char* menu_convert_to_1_u_mid_points = "1 U mid points";
const char* menu_convert_to_0_v_mid_points = "0 V mid points";
const char* menu_convert_to_1_v_mid_points = "1 V mid points";

void convert_to_0_u_mid_points();
void convert_to_1_u_mid_points();
void convert_to_0_v_mid_points();
void convert_to_1_v_mid_points();

const char* menu_mirror_half = "mirror half";
//const char* menu_mirror_modify_west = "mirror modify west";
extern const char* menu_scale;
extern const char* menu_scale_rotation;
extern const char* menu_scale_rotation_translation;
extern const char* menu_inside_out;
extern const char* menu_do_build;
extern const char* menu_settings;
extern const char* menu_reset_rotations;


extern const char* menu_element;
extern const char* menu_short_copy;
extern const char* menu_insert_before;
extern const char* menu_copy_of;
extern const char* menu_move;
extern const char* menu_replace;
extern const char* menu_insert_after;

extern const char* menu_beam_to;
extern const char* menu_import;
extern const char* menu_export;
extern const char* menu_shape_to_file;
extern const char* menu_shape_from_file;
extern const char* menu_IndexedFaceSet_from_file;
extern const char* menu_convertToSeamless;
extern const char* menu_convertToAnim;
extern const char* menu_optimize;
extern const char* menu_calculate_acceleration_for_return;
extern const char* menu_stretchCurve;
extern const char* menu_undo_IS;
extern const char* menu_delete_this_ROUTE;
extern const char* menu_sibling_above;
extern const char* menu_sibling_below;
extern const char* menu_Anim_nodes_from_file;
extern const char* menu_all_nodes_from_file;
extern const char* menu_Build_nodes_from_file;
extern const char* menu_convert;
extern const char* menu_convert_all;
extern const char* menu_fast_build;
extern const char* menu_full_build;
extern const char* menu_poly_map_mirrorX="part poly map mirrorX";

const char* menu_output_povray_movie="output povray movie";
const char* menu_output_batch_file="output batch file";
const char* menu_reduce_captured_poses="reduce captured poses";
const char* menu_convert_to_patch="convert to patch";
const char* menu_convert_to_bendy="convert to bendy";
const char* menu_convert_pivot_points_to_translations="convert pivot points to translations";
const char* menu_convert_translations_to_pivot_points="convert translations to pivot points";

extern Node* g_parent;

void preBreakAndJoinVerticesDemo();
void breakAndJoinVerticesDemo();
void preNurbsSurfacePolyEditingDemo();
void nurbsSurfacePolyEditingDemo();
void convertToSubdiv();
//void exportObj();
bool modifyVrmlX3dFilename(String& filename, bool writeX3DXML, bool writeX3DVRMLClassic);
void extractFileName(String& filename, const String& path);

void compileBuildTransform();
void createLathesFromPivotPoints();
void openRecentMenu(int recentFileIndex);
void importBxxAvStudio();
void exportAnimToBvh();
void outputVrmlMorph();
void setPeriodFromVideo();
void importPivotPoints();
void importPivotPointDistances();
void exportMovie();
void exportObj();
void exportWebGL();
void exportVrmlX3d();
void outputNodeBind();
void convertNurbsLinesToStrands();
void toggleTransformFallOff();
void simplifyCurve();

void importAllNodesField();
void importBuildNodesField();

void importBuildReplace(int fieldIndex);
void addImportBuildAbove(int fieldIndex);
void addImportBuildBelow(int fieldIndex);

void importAnimationReplace(int fieldIndex);
void addImportAnimAbove(int fieldIndex);
void addImportAnimBelow(int fieldIndex);

void importAllNodesReplace(int fieldIndex);
void addImportNodeAbove(int fieldIndex);
void addImportNodeBelow(int fieldIndex);

void addRouteToField(int dstFieldIndex, int srcFieldIndex);
void addIsField(int NodeFieldIndex, int protoFieldIndex);

void showCages();
void hideCages();
void addSelectedPatchesToGroup();
void deleteSelectedPatches();
void mirrorPatches();
void unmirrorPatches();

void switchViewNeutal(){
	Viewpoint::neutral = true;
	g_stw->paint();
}
void rotateCtrlStaves();
void calcLatheWeights();
void deleteMorpherChannels();
void toggleAnimNeutral();
void manualCaptureIKPoints();
void manualCaptureIKPoint();
void convertToNurbsTrack();
void newNurbsTrackSegment();

void beginCompensation();
void removeCompensation();
void addNiki();
void snapNikiPoint();
void displayNikiPanel();
void highlightWholeTimeLine();
void openAnimEditOptions();

void untickIkFields();
void generateRawStitchPolys();
void clearOwnership();
void clearOwnershipPatch();
void repeatSeperatorMap();
void autoLengthStrands();
void mergeLocks();
void copyLockMap();
void addRegions();
void selectLocksFromRegions();
void rotateRegion();
void deleteStrandsInSelRegions();
void seedStrandsInSelRegions();
void seed5Strands();
void connectLockToSelRegions();
//void extrudeRegion();
void splitLock();
void cutLock();
void pasteBorderRegion();
void scaleLockMap();
void hideBoundStrands();
void showBoundStrands();
void insertLock();
void ringletGenerator();
void thinner();
void mirrorLock();
void insertStrandControlPoints();
void seed();
void deleteUnboundStrands();
void createLocksFromStrands();
//void generateAllStrands();
//void generateOnlySelecteStrands();

//void decrementSeperatorMap();
void removeAllPolyMidPointsOwnership();
void exportMorphedPatch();
void importMorphedPatch();
void synchronizeCells();
void selectAnimBarFromList();
extern bool g_dontCloseMenu;

void subdivPatches();
void changeFormantKey();

void buildF10Menu(){
	const Vec2i localCursor = g_mousePos - g_tb->pos;
	g_dontCloseMenu = true;

	g_remMenu = new Menu;
	M_<Ptr<MenuItem> >& list = g_remMenu->list;
//		child+=new MenuItem_(menu_reset_view_position_and_orientation, reset3dScene, *(menu->menuPtr));
	list+=new MenuItem_(Vec2i(16,16), Vec2i(remember_x, remember_y), menu_copy, remember, *g_remMenu);
	{
		BranchMenuItem* menu=new BranchMenuItem(Vec2i(16,16), Vec2i(16, 192), menu_paste, *g_remMenu);
		list+=menu;
		M_<Ptr<MenuItem> >& child=menu->menuPtr->list;
		{
			BranchMenuItem* menu2=new BranchMenuItem(menu_coord, *(menu->menuPtr));
			child+=menu2;
			M_<Ptr<MenuItem> >& child2=menu2->menuPtr->list;

			child2+=new MenuItem_(menu_mirrorX,		forceToMirrorX, *(menu2->menuPtr));
		}
		{
			BranchMenuItem* menu2=new BranchMenuItem(menu_color, *(menu->menuPtr));
			child+=menu2;
			M_<Ptr<MenuItem> >& child2=menu2->menuPtr->list;

			child2+=new MenuItem_(Vec2i(24,24), Vec2i(colorVertex_x, colorVertex_y), menu_vertexColor,	colorVertex, *(menu2->menuPtr));
			child2+=new MenuItem_(Vec2i(24,24), Vec2i(joinMatchingColors_x, joinMatchingColors_y), menu_join_matching_in_all_parts,  joinMatchingColors, *(menu2->menuPtr));
			child2+=new MenuItem_(Vec2i(24,24), Vec2i(pasteColorToTriangle_x, pasteColorToTriangle_y), menu_triangleColor,           pasteColorToTriangle, *(menu2->menuPtr));
			child2+=new MenuItem_(Vec2i(24,24), Vec2i(colorPart_x, colorPart_y), menu_partColor,                 colorPart, *(menu2->menuPtr));
			child2+=new MenuItem_(Vec2i(24,24), Vec2i(replaceColorPart_x, replaceColorPart_y), menu_vertices_colored,   replaceColorPart, *(menu2->menuPtr));
			child2+=new MenuItem_(Vec2i(24,24), Vec2i(replaceColorSkeleton_x, replaceColorSkeleton_y), menu_vertices_in_all_parts_colored, replaceColorSkeleton, *(menu2->menuPtr));
		}
		{
			BranchMenuItem* menu2=new BranchMenuItem(menu_ownership, *(menu->menuPtr));
			child+=menu2;
			M_<Ptr<MenuItem> >& child2=menu2->menuPtr->list;

			child2+=new MenuItem_(Vec2i(24,24), Vec2i(transferSingleVertex_x, transferSingleVertex_y), menu_vertexOwnership,               transferSingleVertex, *(menu2->menuPtr));
			child2+=new MenuItem_(menu_part,                        transferPartVertex, *(menu2->menuPtr));
			child2+=new MenuItem_(menu_close_vertices,              transferCloseVertices, *(menu2->menuPtr));
			child2+=new MenuItem_(Vec2i(24,24), Vec2i(transferCloseVerticesJoin_x, transferCloseVerticesJoin_y), menu_close_vertices_then_join, transferCloseVerticesJoin, *(menu2->menuPtr));
		}
		child+=new Seperator(*(menu->menuPtr));
		child+=new MenuItem_(menu_part,			replaceWithCopyOfRememberedPart, *(menu->menuPtr));

		child+=new MenuItem_(Vec2i(24,24), Vec2i(replaceWithCopyOfRememberedPartMirrorX_x, replaceWithCopyOfRememberedPartMirrorX_y), menu_part_mirrorX,	replaceWithCopyOfRememberedPartMirrorX, *(menu->menuPtr));
		child+=new MenuItem_(menu_part_mirrorY,	replaceWithCopyOfRememberedPartMirrorY, *(menu->menuPtr));
		child+=new MenuItem_(menu_part_mirrorZ,	replaceWithCopyOfRememberedPartMirrorZ, *(menu->menuPtr));
		//child+=new Seperator(*(menu->menuPtr));
		child+=new MenuItem_(menu_poly_map_mirrorX, replaceWithCopyOfPolyMapMirrorX, *(menu->menuPtr));
		child+=new Seperator(*(menu->menuPtr));
		child+=new MenuItem_(menu_add_part,	pasteAddCopyOf, *(menu->menuPtr));
		child+=new MenuItem_(Vec2i(24,24), Vec2i(pasteAddCopyOfMirrorX_x, pasteAddCopyOfMirrorX_y), menu_add_part_mirrorX,	pasteAddCopyOfMirrorX, *(menu->menuPtr));
		child+=new MenuItem_(menu_add_part_mirrorY,	pasteAddCopyOfMirrorY, *(menu->menuPtr));
		child+=new MenuItem_(menu_add_part_mirrorZ,	pasteAddCopyOfMirrorZ, *(menu->menuPtr));
	}
	list+=new Seperator(*g_remMenu);
	/*
	{
		BranchMenuItem* menu=new BranchMenuItem("swap vertex order", *g_remMenu);
		list+=menu;
		M_<Ptr<MenuItem> >& child=menu->menuPtr->list;

		child+=new MenuItem_("with previous vertex (U)",     swapVertexOrderDec, *(menu->menuPtr));
		child+=new MenuItem_("with next vertex (I)",         swapVertexOrderInc, *(menu->menuPtr));
	}
	*/
	{
		BranchMenuItem* menu=new BranchMenuItem(menu_join_move, *g_remMenu);
		list+=menu;
		M_<Ptr<MenuItem> >& child=menu->menuPtr->list;

		child+=new MenuItem_(Vec2i(24,24), Vec2i(joinVertexKeepDestColor_x, joinVertexKeepDestColor_y), menu_join_vertex_keep_dest_color,				joinVertexKeepDestColor, *(menu->menuPtr));
		child+=new MenuItem_(Vec2i(24,24), Vec2i(joinVertexKeepBothColors_x, joinVertexKeepBothColors_y), menu_join_vertex_keep_both_colors,	joinVertexKeepBothColors, *(menu->menuPtr));
		child+=new MenuItem_(menu_join_vertex_keep_source_color,	joinVertexKeepSourceColor, *(menu->menuPtr));

		child+=new MenuItem_(menu_move_vertex,				moveVertex, *(menu->menuPtr));
		child+=new MenuItem_(Vec2i(24,24), Vec2i(zeroVetexX_x, zeroVetexX_y), menu_set_x_to_zero,					zeroVetexX, *(menu->menuPtr));
	}
	{
		BranchMenuItem* menu=new BranchMenuItem(menu_break_join, *g_remMenu);
		list+=menu;
		M_<Ptr<MenuItem> >& child=menu->menuPtr->list;

		child+=new MenuItem_(Vec2i(24,24), Vec2i(breakVertexJoin_x, breakVertexJoin_y), menu_break_coord,                   breakVertexJoin, *(menu->menuPtr));
		child+=new MenuItem_(Vec2i(24,24), Vec2i(breakColorJoin_x, breakColorJoin_y), menu_break_color,                breakColorJoin, *(menu->menuPtr));
	}
	{
		BranchMenuItem* menu=new BranchMenuItem(Vec2i(16,16), Vec2i(delete_x, delete_y), menu_delete, *g_remMenu);
		list+=menu;
		M_<Ptr<MenuItem> >& child=menu->menuPtr->list;
		child+=new MenuItem_(menu_vertex,                      immediateDeleteVertex, *(menu->menuPtr));
		child+=new MenuItem_(Vec2i(24,24), Vec2i(deleteTri_x, deleteTri_y), menu_triangleDelete,           deleteTriangle, *(menu->menuPtr));
		child+=new MenuItem_(menu_triangles_right_from_selected_coord, cutRightFromVertex, *(menu->menuPtr));
		child+=new MenuItem_(Vec2i(24,24), Vec2i(cutToTheLeftOfTheCursor_x, cutToTheLeftOfTheCursor_y), menu_triangles_left_from_selected_coord, cutLeftFromVertex, *(menu->menuPtr));
		child+=new MenuItem_(Vec2i(24,24), Vec2i(cutUpFromTheCursor_x, cutUpFromTheCursor_y), menu_triangles_up_from_selected_coord, cutUpFromVertex, *(menu->menuPtr));
		child+=new MenuItem_(Vec2i(24,24), Vec2i(cutDownFromTheCursor_x, cutDownFromTheCursor_y), menu_triangles_down_from_selected_coord, cutDownFromVertex, *(menu->menuPtr));
		child+=new MenuItem_(menu_layered,     findLayeredTriangle, *(menu->menuPtr));
		child+=new MenuItem_(menu_all_triangles_in_part,       cutWholePart, *(menu->menuPtr));
		child+=new MenuItem_("morpher channels for selected points",       deleteMorpherChannels, *(menu->menuPtr));
	}
	list+=new Seperator(*g_remMenu);
	list+=new MenuItem_(Vec2i(24,24), Vec2i(mirrorTriangleImmediate_x, mirrorTriangleImmediate_y), menu_mirror_triangle,			mirrorTriangleImmediate, *g_remMenu);
	list+=new MenuItem_(Vec2i(24,24), Vec2i(splitTriangleImmediate_x, splitTriangleImmediate_y), menu_split_triangle,			splitTriangleImmediate, *g_remMenu);
	list+=new MenuItem_(Vec2i(16,16), Vec2i(swapDivide_x, swapDivide_y), menu_swap_triangle_divide, swapDivide, *g_remMenu);
	/*
	{
		BranchMenuItem* menu=new BranchMenuItem("swap triangle divide", *g_remMenu);
		list+=menu;
		M_<Ptr<MenuItem> >& child=menu->menuPtr->list;
		child+=new MenuItem_("between triangle ? and triangle ? (D)", swapDivide, *(menu->menuPtr));
	}
	*/
	list+=new Seperator(*g_remMenu);
	/*
	{
		BranchMenuItem* menu=new BranchMenuItem("RGB color", *g_remMenu);
		list+=menu;
		M_<Ptr<MenuItem> >& child=menu->menuPtr->list;

		child+=new MenuItem_("lighten (Ctrl -)",             lightenVertex, *(menu->menuPtr));
		child+=new MenuItem_("darken (Ctrl =)",              darkenVertex, *(menu->menuPtr));
	}
	*/
	{
		BranchMenuItem* menu=new BranchMenuItem(menu_compile_transform, *g_remMenu);
		list+=menu;
		M_<Ptr<MenuItem> >& child=menu->menuPtr->list;

		child+=new MenuItem_(menu_scale,						compileScale, *(menu->menuPtr));
		child+=new MenuItem_(menu_scale_rotation,				compileScaleRotation, *(menu->menuPtr));
		child+=new MenuItem_(menu_scale_rotation_translation,	compileTransform, *(menu->menuPtr));
	}
	list+=new MenuItem_(menu_inside_out,           reverseIndices, *g_remMenu);
//	list+=new MenuItem_("generateNormals\tCtrl N",     GENERATE_NORMALS);
	list+=new MenuItem_(menu_do_build,                 doBuild, *g_remMenu);
	{
		BranchMenuItem* menu=new BranchMenuItem("add geometry", *g_remMenu);
		list+=menu;
		M_<Ptr<MenuItem> >& child=menu->menuPtr->list;
		child+=new MenuItem_(Vec2i(16,16), Vec2i(sphereBuildIcon_x, sphereBuildIcon_y), tip_build_sphere_for_selected_part,		buildSphereForSelPart, *(menu->menuPtr));
		child+=new MenuItem_(Vec2i(16,16), Vec2i(cylinderBuildIcon_x, cylinderBuildIcon_y), tip_build_cylinder_for_selected_part,		buildCylinderForSelPart, *(menu->menuPtr));
		child+=new MenuItem_(Vec2i(16,16), Vec2i(coneBuildIcon_x, coneBuildIcon_y), tip_build_cone_for_selected_part,		buildConeForSelPart, *(menu->menuPtr));
		child+=new MenuItem_(Vec2i(16,16), Vec2i(torusBuildIcon_x, torusBuildIcon_y), tip_build_torus_for_selected_part,		buildTorusForSelPart, *(menu->menuPtr));
		child+=new MenuItem_(Vec2i(16,16), Vec2i(bezierLatheBuildIcon_x, bezierLatheBuildIcon_y), tip_build_bezier_lathe_for_selected_part,		buildBezierLatheForSelPart, *(menu->menuPtr));
		child+=new MenuItem_(Vec2i(16,16), Vec2i(boxBuildIcon_x, boxBuildIcon_y), tip_build_box_for_selected_part,		buildBoxForSelPart, *(menu->menuPtr));
		child+=new MenuItem_(Vec2i(16,16), Vec2i(rectangleBuildIcon_x, rectangleBuildIcon_y), tip_build_rectangle_for_selected_part,		newRectangle, *(menu->menuPtr));
		child+=new MenuItem_(menu_immediate_triangle_for_selected_part,		genTriangle, *(menu->menuPtr));
	}


	g_remMenu->autoSizeAndStuff();
	g_remMenu->pos = g_mousePos - Vec2i(0, g_remMenu->size.y);
}


extern M_Int newNodeLookUp;
extern int g_fromFieldIdx;
extern int g_routeIndex;
NodePtr g_menueParent;
int g_menueElementForParent;

int g_menuGroupIndex=-1;



void addNewChild(int fieldIdx, int nodeType){
	addNewNodeAsChild(g_selNode.getRaw(), fieldIdx, nodeType);
}
void addNewAbove(int fieldIndex, int nodeType){
	addNewNodeAsSibling(g_menueParent, fieldIndex, g_menueElementForParent, nodeType, DROP_ABOVE);
}
void addNewReplace(int fieldIndex, int nodeType){
	replaceWithNewNode(g_menueParent, fieldIndex, g_menueElementForParent, nodeType);
}
void addNewBelow(int fieldIndex, int nodeType){
	addNewNodeAsSibling(g_menueParent, fieldIndex, g_menueElementForParent, nodeType, DROP_BELOW);
}
void addCopyOfChild(int fieldIndex){
	addCopyOfAsChild(g_selNode.getRaw(), fieldIndex, g_stw->copyLine, g_selCount);
}
void addCopyOfAbove(int fieldIndex){
	addCopyOfNodeAsSibling(g_stw->selLine, g_stw->copyLine, g_copySibCount, DROP_ABOVE);
}
void addCopyOfReplace(int fieldIndex){
	addCopyOfNodeAsSibling(g_stw->selLine, g_stw->copyLine, g_copySibCount, DROP_REPLACE);
}
void addCopyOfBelow(int fieldIndex){
	addCopyOfNodeAsSibling(g_stw->selLine, g_stw->copyLine, g_copySibCount, DROP_BELOW);
}

void addUSEChild(int fieldIndex){
	useNodeAsChild(g_selNode.getRaw(), g_copyNode.getRaw(), fieldIndex, g_stw->copyLine, g_selCount, g_copySibCount);
}
void addUSEAbove(int fieldIndex){
	useSiblingNode(g_selNode.getRaw(), g_stw->selLine, g_stw->copyLine, g_selCount, g_copySibCount, DROP_ABOVE);
}
void addUSEReplace(int fieldIndex){
	useSiblingNode(g_selNode.getRaw(), g_stw->selLine, g_stw->copyLine, g_selCount, g_copySibCount, DROP_REPLACE);
}
void addUSEBelow(int fieldIndex){
	useSiblingNode(g_selNode.getRaw(), g_stw->selLine, g_stw->copyLine, g_selCount, g_copySibCount, DROP_BELOW);
}

void addMoveChild(int fieldIndex){
	moveNodeAsChild(g_selNode.getRaw(), fieldIndex, g_stw->selLine, g_stw->copyLine, g_selCount, g_copySibCount, g_copyAllCount);
}
void addMoveAbove(int fieldIndex){
	replaceNodeWithMove(g_stw->selLine, g_stw->copyLine, g_selCount, g_copySibCount, g_copyAllCount, DROP_ABOVE);
}
void addMoveReplace(int fieldIndex){
	replaceNodeWithMove(g_stw->selLine, g_stw->copyLine, g_selCount, g_copySibCount, g_copyAllCount, DROP_REPLACE);
}
void addMoveBelow(int fieldIndex){
	replaceNodeWithMove(g_stw->selLine, g_stw->copyLine, g_selCount, g_copySibCount, g_copyAllCount, DROP_BELOW);
}
void importNodeField(Node* parentSelNode, int selFieldIdx, int selElement, M_Node& src, unsigned type);
void finishImportNodesFromFile();
void importNode(int importWhere){
	M_Node import;
	String& str=g_loadStr;
	//str=openFile( hWndMain);
	if(str!=""){
		createSmlsFromString(import, str);
		if(import.length){
			if(Seamless* smls=dynamic_cast<Seamless*>(import[0].p)){
				
				int selFieldIdx=0;
				int selElement=0;
				
				int parentSelLine=getParent(g_stw->selLine);
				
				Node* parentSelNode=g_sceneTreeList[parentSelLine].node;
				
				getFNodeElementForParent(parentSelLine, g_stw->selLine, g_sceneTreeList, selFieldIdx, selElement );
				
				importNodeField(parentSelNode, selFieldIdx, selElement, smls->build, importWhere);
				finishImportNodesFromFile();
			}
		}
	}
}

void importNodesFromFileField(const String& filename, M_Node& children, bool onlyAnimation = false, bool recalc = false);
void importNodesFromFileAndInsert(const String& filename, int importWhere, bool onlyAnimation);
void importBuildNode(int importWhere);
void importBuildNodesFromFileField(M_Node& m_node);

void addRouteToSubMenu(Menu& window, Field* fieldPtr, int srcIdx){
	BranchMenuItem* branch=0;
	if(g_copyNode.getRaw()){
		const M_Properties &mprop=g_copyNode->getProperties();

		for(int i=0; i<mprop.length; i++){
			const Properties& sprop=mprop[i];
	 
			Field* destField=g_copyNode->getField(i);
			if(fieldPtr->getType()==destField->getType()&&sprop.isWritable()){
				if(!branch){
					Button::str="TO ";
					Button::str+=g_copyNode->name;
					branch=new BranchMenuItem(Button::str, window);
					window.list+=branch;
				}
				branch->menuPtr->list+=new MenuItem_int_int(sprop.name, addRouteToField, i, srcIdx, *(branch->menuPtr));
			}
		}
	}
}
ProtoDeclare* getSceneProtoDeclare(int selLine);


void importAnimation();//IMPORT_ANIMATION
void importAnimRecalc();


void addISToSubMenu(Menu& window, Field* fieldPtr, int srcIdx, const Properties& sprop, ProtoInterface* proto){
	BranchMenuItem* branch=0;
	if(proto){

		M_Properties &mprop=proto->properties;

		for(int i=0; i<mprop.length; i++){
			Properties& spropProto=mprop[i];
 
			Field* destField=proto->getField(i);
			if((fieldPtr->getType()==destField->getType()) && ((spropProto.isWritable()==sprop.isWritable()) || (spropProto.isReadable()==sprop.isReadable()))){
				if(!branch){
					Button::str="IS ";
					Button::str+=proto->name;
					branch=new BranchMenuItem(Button::str, window);
					window.list+=branch;
				}
				branch->menuPtr->list+=new MenuItem_int_int(spropProto.name, addIsField, srcIdx, i, *(branch->menuPtr));
			}
		}
	}
}
void elementBeforeCopyOf();//ELEMENT_BEFORE_COPYOF
void elementBeforeMove();//ELEMENT_BEFORE_MOVE
void elementReplaceCopyOf();//ELEMENT_REPLACE_COPYOF
void elementReplaceMove();//ELEMENT_REPLACE_MOVE
void elementAfterCopyOf();//ELEMENT_AFTER_COPYOF
void elementAfterMove();//ELEMENT_AFTER_MOVE
void deleteElement();//DELETE_ELEMENT

void addNewToSubMenu(Menu& window, Field* fieldPtr, int fieldIndex, const Properties& sprop, void (*function)(int fieldIndex, int nodeType)){
	if(
		sprop.blockNew()||
		fieldPtr->getType() != SFNODE&&
		fieldPtr->getType() != MFNODE||
		sprop.promptNode.length == 0)
		{
		return;
	}

//	BranchMenuItem* newMenu=new BranchMenuItem(Vec2i(24,24), Vec2i(435,24), "new", window);
	BranchMenuItem* newMenu=new BranchMenuItem(Vec2i(16,16), Vec2i(464,128), "new", window);
	{
		for(int j=0; j<sprop.promptNode.length; j++){
			int nodeType=sprop.promptNode[j];
			int iconNodeType = nodeType;
			if(nodeType>=g_userProtoBegin){
				iconNodeType=PROTOINSTANCE;
			}
			newMenu->menuPtr->list+=new MenuItem_int_int(builtInProtoList[iconNodeType]->iconSize, builtInProtoList[iconNodeType]->iconSrcPos, nodeLookUpStr[nodeType].c_str(), function, fieldIndex, nodeType, *(newMenu->menuPtr));
		}
	}
	window.list+=newMenu;
}
void addUSEToSubMenu(Menu& window, Field* fieldPtr, int fieldIndex, const Properties& sprop, void (*function)(int fieldIndex)){
	if(
		sprop.blockUse()||
		fieldPtr->getType()!=SFNODE&&
		fieldPtr->getType()!=MFNODE||
		g_copyNode.getRaw()==0||
		g_copyNode->name=="")
		{
		return;
	}
	for(int j=0; j<sprop.promptNode.length; j++){
		if(sprop.promptNode[j]==g_copyNode->getType()){
			Button::str="USE ";
			Button::str+=g_copyNode->name;
			window.list+=new MenuItem_int(Button::str, function, fieldIndex, window);
			return;
		}
	}
}
void addCopyOfToSubMenu(Menu& window, Field* fieldPtr, int fieldIndex, const Properties& sprop, void (*function)(int fieldIndex)){
	if(
		sprop.blockNew()||
		fieldPtr->getType()!=SFNODE&&
		fieldPtr->getType()!=MFNODE||
		g_copyNode.getRaw()==0)
		{
		return;
	}

	for(int j=0; j<sprop.promptNode.length; j++){
		if(sprop.promptNode[j]==g_copyNode->getType()){
			window.list+=new MenuItem_int("copy of" , function, fieldIndex, window);
		}
	}
}
void addMoveToSubMenu(Menu& window, Field* fieldPtr, int fieldIndex, const Properties& sprop, void (*function)(int fieldIndex)){
	if(
		sprop.blockNew()||
		fieldPtr->getType()!=SFNODE&&
		fieldPtr->getType()!=MFNODE||
		g_copyNode.getRaw()==0)
		{
		return;
	}

	for(int j=0; j<sprop.promptNode.length; j++){
		if(sprop.promptNode[j]==g_copyNode->getType()){
			Seamless* smls = getSeamless2();
			//if(smls == 0 || smls->workAnimCP == 0){
			window.list+=new MenuItem_int(menu_move , function, fieldIndex, window);
			//}
		}
	}
}
void addElementToSubMenu(Menu& window, Field* fieldPtr, int srcIdx, const Properties& sprop){
	if(g_menuGroupIndex!=-1){
		String str;

		BranchMenuItem* branchMenuItem = new BranchMenuItem(menu_element, window);
		branchMenuItem->menuPtr->list+=new MenuItem_(Vec2i(16,16), Vec2i(remember_x, remember_y), menu_short_copy, remember, *(branchMenuItem->menuPtr));

		BranchMenuItem* pasteMenu = new BranchMenuItem(Vec2i(16,16), Vec2i(16, 192), menu_paste, *(branchMenuItem->menuPtr));
		branchMenuItem->menuPtr->list+=pasteMenu;
		M_<Ptr<MenuItem> >& pasteChild=pasteMenu->menuPtr->list;
		{
			BranchMenuItem* insertElement = new BranchMenuItem(menu_insert_before, *(pasteMenu->menuPtr));
			insertElement->menuPtr->list+=new MenuItem_(menu_copy_of, elementBeforeCopyOf, *(insertElement->menuPtr));
			insertElement->menuPtr->list+=new MenuItem_(menu_move, elementBeforeMove, *(insertElement->menuPtr));
			pasteChild+=insertElement;
		}
		{
			BranchMenuItem* insertElement = new BranchMenuItem(menu_replace, *(pasteMenu->menuPtr));
			insertElement->menuPtr->list+=new MenuItem_(menu_copy_of, elementReplaceCopyOf, *(insertElement->menuPtr));
			insertElement->menuPtr->list+=new MenuItem_(menu_move, elementReplaceMove, *(insertElement->menuPtr));
			pasteChild+=insertElement;
		}
		{
			BranchMenuItem* insertElement = new BranchMenuItem(menu_insert_after, *(pasteMenu->menuPtr));
			insertElement->menuPtr->list+=new MenuItem_(menu_copy_of, elementAfterCopyOf, *(insertElement->menuPtr));
			insertElement->menuPtr->list+=new MenuItem_(menu_move, elementAfterMove, *(insertElement->menuPtr));
			pasteChild+=insertElement;
		}
		branchMenuItem->menuPtr->list+=new Seperator(*(branchMenuItem->menuPtr));
		branchMenuItem->menuPtr->list+=new MenuItem_(Vec2i(16,16), Vec2i(delete_x, delete_y), menu_delete, deleteElement, *(branchMenuItem->menuPtr));
		window.list+=branchMenuItem;
	}
}

Ptr<BranchMenuItem> addFieldMenu(Menu& home, Field* fieldPtr, int fieldIdx, const Properties& sprop, ProtoInterface* protoInterface){
	Ptr<BranchMenuItem> branchMenuItem=new BranchMenuItem(sprop.name, home);
	M_<MenuItemPtr>& list=branchMenuItem->menuPtr->list;
	int insertOrder=0;
	static String str;
	Menu& menu=*(branchMenuItem->menuPtr);
	//insert new
	addNewToSubMenu(menu, fieldPtr, fieldIdx, sprop, addNewChild);
	addUSEToSubMenu(menu, fieldPtr, fieldIdx, sprop, addUSEChild);
	//insert copyOf
	addCopyOfToSubMenu(menu, fieldPtr, fieldIdx, sprop, addCopyOfChild);
	//insert move
	addMoveToSubMenu(menu, fieldPtr, fieldIdx, sprop, addMoveChild);
//	if(sprop.isReadable()){
//		addRouteToSubMenu(menu, fieldPtr, fieldIdx);
//	}
	addISToSubMenu(menu, fieldPtr, fieldIdx, sprop, protoInterface);

	addElementToSubMenu(menu, fieldPtr, fieldIdx, sprop);

	//insert import
	if(dynamic_cast<Progenitor*>(g_selNode.getRaw()) && dynamic_cast<MFNode*>(fieldPtr)){
		Ptr<BranchMenuItem> importMenu=new BranchMenuItem(Vec2i(16,16), Vec2i(464,80), menu_import, menu);
		menu.list+=importMenu.p;
		importMenu->menuPtr->list+=new MenuItem_(menu_Anim_nodes_from_file, importAnimation, *(importMenu->menuPtr));
		importMenu->menuPtr->list+=new MenuItem_("Anim nodes & recalc", importAnimRecalc, *(importMenu->menuPtr));
		if(dynamic_cast<GestureGroup*>(g_selNode.getRaw())==0){
			importMenu->menuPtr->list+=new MenuItem_(menu_all_nodes_from_file, importAllNodesField, *(importMenu->menuPtr));
		}
	}
	if((dynamic_cast<Seamless*>(g_selNode.getRaw())||dynamic_cast<GroupBuild*>(g_selNode.getRaw())) && dynamic_cast<MFBuild*>(fieldPtr)){
		Ptr<BranchMenuItem> importMenu=new BranchMenuItem(Vec2i(16,16), Vec2i(464,80), menu_import, menu);
		menu.list+=importMenu.p;
		importMenu->menuPtr->list+=new MenuItem_(menu_Build_nodes_from_file, importBuildNodesField, *(importMenu->menuPtr));
	}

	return branchMenuItem;
}

void extendChildrenLookUp(Node* node, int& remproptNodeLen){
	if(!node)return;
	Progenitor* c=dynamic_cast<Progenitor*>(node);
	if(!c)return;
	const Properties& sprop=node->getProperties(c->children);
	remproptNodeLen=sprop.promptNode.length;
	for(int fieldIdx=g_userProtoBegin; fieldIdx<g_lookupProtoInterface.length; fieldIdx++){
		const_cast<Properties&>(sprop).promptNode+=fieldIdx;
	}
}
void restoreChildrenLookUp(Node* node, int& remproptNodeLen){
	if(!node)return;
	Progenitor* c=dynamic_cast<Progenitor*>(node);
	if(!c)return;
	const Properties& sprop=node->getProperties(c->children);
	const_cast<Properties&>(sprop).promptNode.length=remproptNodeLen;
}

Ptr<BranchMenuItem> addPasteMenu(const Vec2i i_iconSize, const Vec2i i_srcPos, const char* label, Menu& parentWin, Field* fieldPtr, int fieldIndex, const Properties& sprop, ProtoInterface* protoInterface, int& cntr, void (newOption)(int fieldIndex, int nodeType), void (*copyOfOption)(int fieldIndex), void (*moveOption)(int fieldIndex), void (*useOption)(int fieldIndex), void (*importOption)(int fieldIndex), Node* parentNode){
	Ptr<BranchMenuItem> branchMenuItem=new BranchMenuItem(i_iconSize, i_srcPos, label, parentWin);
	//M_<MenuItemPtr>& menuChild=branchMenuItem->menuPtr->list;
	int insertOrder=0;
	static String str;
	addNewToSubMenu(*(branchMenuItem->menuPtr), fieldPtr, fieldIndex, sprop, newOption);
	addUSEToSubMenu(*(branchMenuItem->menuPtr), fieldPtr, fieldIndex, sprop, useOption);
	addCopyOfToSubMenu(*(branchMenuItem->menuPtr), fieldPtr, fieldIndex, sprop, copyOfOption);
	addMoveToSubMenu(*(branchMenuItem->menuPtr), fieldPtr, fieldIndex, sprop, moveOption);
	if(sprop.name=="build"){
		BranchMenuItem* importMenu = new BranchMenuItem(Vec2i(16,16), Vec2i(464,80), menu_import, *(branchMenuItem->menuPtr));
		branchMenuItem->menuPtr->list+=importMenu;
		if(importOption==addImportNodeAbove){
			importOption=addImportBuildAbove;
		}
		else if(importOption==importAllNodesReplace){
			importOption=importBuildReplace;
		}
		else if(importOption==addImportNodeBelow){
			importOption=addImportBuildBelow;
		}
		importMenu->menuPtr->list+=new MenuItem_int(menu_Build_nodes_from_file, importOption, fieldIndex, *(importMenu->menuPtr));
	}
	else if(sprop.name=="children"){
		BranchMenuItem* importMenu = new BranchMenuItem(Vec2i(16,16), Vec2i(464,80), menu_import, *(branchMenuItem->menuPtr));
		branchMenuItem->menuPtr->list += importMenu;

		if(dynamic_cast<GestureGroup*>(parentNode)==0){
			importMenu->menuPtr->list+=new MenuItem_int(menu_all_nodes_from_file, importOption, fieldIndex, *(importMenu->menuPtr));
		}
		if(importOption==addImportNodeAbove){
			importOption=addImportAnimAbove;
		}
		else if(importOption==importAllNodesReplace){
			importOption = importAnimationReplace;
		}
		else if(importOption==addImportNodeBelow){
			importOption=addImportAnimBelow;
		}
		importMenu->menuPtr->list+=new MenuItem_int(menu_Anim_nodes_from_file, importOption, fieldIndex, *(importMenu->menuPtr));
	}

	return branchMenuItem;
}




void deleteNode();//DELETE_NODE
void convertGroupToSmls();//CONVERT_GROUP_TO_SMLS
void convertHAnimToSmls();
void convertTimeSensorToAnim();
void optimize();//OPTIMIZE
void convertToCNames();
void homeToInit();//HOME_TO_INIT
void stretchCurve();//STRETCH_CURVE
void samplePose();//SAMPLE_POSE:
void newPanel();
void hide();
void toggleTrack();
void setTrackAndDescendants();
void hideAndDescendants();
void setTrackAndDescendants();
void selectTrackExclusive();
void calcLookAtFromOrientation();
void selectModelTracks();
void selectModelTracksExclusive();
void selectSceneTracks();

void importIndexedFaceSet(){//IMPORT_INDEXEDFACESET
}
void deleteRoute(int fieldIndex, int routeIndex);//DELETE_ROUTE

void undoIs();//UNDO_IS

void SceneTree::buildRoutePopupMenu(Menu& menu){
	M_<MenuItemPtr>& list=menu.list;
	String str;

	ProtoInterface* protoInterface=0;
	ProtoDeclare* protoDeclare=getSceneProtoDeclare(selLine);
	if(protoDeclare!=0){
		protoInterface=protoDeclare->protoInterface;
	}
	const M_Properties& mprop=g_selNode->getProperties();

	Node* parentNode=0;
	int parentSelLine=-1;
	if(g_stw->selLine!=0){
		parentSelLine=getParent(g_stw->selLine);
		parentNode=g_sceneTreeList[parentSelLine].node;
	}
	int remproptNodeLenA=0;
	int remproptNodeLenB=0;
	extendChildrenLookUp(parentNode, remproptNodeLenA);	//must not return untill restored
	extendChildrenLookUp(g_selNode.getRaw(), remproptNodeLenB);

	int cntr=0;


	Ptr<BranchMenuItem> pasteMenu=new BranchMenuItem(Vec2i(16,16), Vec2i(16, 192), menu_paste, menu);
	list+=pasteMenu.p;
	if(parentNode){
		int selFieldIdx=0;

		g_menueParent=g_sceneTreeList[parentSelLine].node;

		getFNodeElementForParent(parentSelLine, selLine, g_sceneTreeList, selFieldIdx, g_menueElementForParent );

		const Properties& sprop=g_menueParent->getProperties()[selFieldIdx];

		Field* fieldPtr=g_menueParent->getField(selFieldIdx);

		if(fieldPtr->getType()==MFNODE){
			Ptr<BranchMenuItem> subMenu=addPasteMenu(Vec2i(16,16), Vec2i(dropSiblingAbove_x, dropSiblingAbove_y), menu_sibling_above, *(pasteMenu->menuPtr), fieldPtr, selFieldIdx, sprop, protoInterface, cntr, addNewAbove, addCopyOfAbove, addMoveAbove, addUSEAbove, addImportNodeAbove, g_menueParent);
			if(subMenu->menuPtr->list.length){
				pasteMenu->menuPtr->list+=subMenu.p;
			}
		}
		Ptr<BranchMenuItem> subMenu=addPasteMenu(Vec2i(16,16), Vec2i(dropSiblingReplace_x, dropSiblingReplace_y), menu_replace, *(pasteMenu->menuPtr), fieldPtr, selFieldIdx, sprop, protoInterface, cntr, addNewReplace, addCopyOfReplace, addMoveReplace, addUSEReplace, importAllNodesReplace, g_menueParent);
		if(subMenu->menuPtr->list.length){
			pasteMenu->menuPtr->list+=subMenu.p;
		}
		if(fieldPtr->getType()==MFNODE){
			Ptr<BranchMenuItem> subMenu=addPasteMenu(Vec2i(16,16), Vec2i(dropSiblingBelow_x, dropSiblingBelow_y), menu_sibling_below, *(pasteMenu->menuPtr), fieldPtr, selFieldIdx, sprop, protoInterface, cntr, addNewBelow, addCopyOfBelow, addMoveBelow, addUSEBelow, addImportNodeBelow, g_menueParent);
			if(subMenu->menuPtr->list.length){
				pasteMenu->menuPtr->list+=subMenu.p;
			}
		}
	}
	pasteMenu->menuPtr->list+=new Seperator(*(pasteMenu->menuPtr));
	if(g_selNode->getType()!=PROTODECLARE){
		for(int i=0; i<mprop.length; i++){
			const Properties& sprop=mprop[i];
			Field* fieldPtr=g_selNode->getField(i);
			if(dynamic_cast<SFNode*>(fieldPtr) || dynamic_cast<MFNode*>(fieldPtr)){
				Ptr<BranchMenuItem> subMenu=addFieldMenu(*(pasteMenu->menuPtr), fieldPtr, i, sprop, protoInterface);
				if(subMenu->menuPtr->list.length){
					pasteMenu->menuPtr->list+=subMenu.p;
				}
			}
		}
	}
	switch(g_selNode->getType()){
		case SCENE:
		case PROTOINTERFACE:
		case PROTOBODY:
			break;
		default:
			list+=new Seperator(menu);
			Seamless* smls = getSeamless2();
			//if(smls == 0 || smls->workAnimCP == 0){
			list+=new MenuItem_(Vec2i(16,16), Vec2i(delete_x, delete_y), menu_delete, deleteNode, menu);
			//}
			list+=new Seperator(menu);
	}

	bool extra=false;
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(part->show){
			list+=new MenuItem_(Vec2i(hiddenPart_w,hiddenPart_h), Vec2i(hiddenPart_x, hiddenPart_y), "hide", hide, menu);
			list+=new MenuItem_(Vec2i(hiddenPart_w,hiddenPart_h), Vec2i(hiddenPart_x, hiddenPart_y), "hide & descendants", hideAndDescendants, menu);
		}
		else{
			int type=part->getType();
			list+=new MenuItem_(builtInProtoList[type]->iconSize, builtInProtoList[type]->iconSrcPos, "show", hide, menu);
			list+=new MenuItem_(builtInProtoList[type]->iconSize, builtInProtoList[type]->iconSrcPos, "show & descendants", hideAndDescendants, menu);
		}
		if(part->track){
			list+=new MenuItem_("unselect track", toggleTrack,  menu);
			list+=new MenuItem_("unselect track & descendants", setTrackAndDescendants,  menu);
		}
		else{
			list+=new MenuItem_("select track",toggleTrack,  menu);
			list+=new MenuItem_("select track & descendants", setTrackAndDescendants,  menu);
		}
		list+=new MenuItem_("select track exclusive", selectTrackExclusive,  menu);
		list+=new MenuItem_("select model tracks", selectModelTracks,  menu);
		list+=new MenuItem_("select model tracks exclusive", selectModelTracksExclusive,  menu);
		list+=new MenuItem_("select all scene tracks", selectSceneTracks,  menu);

		list+=new Seperator(menu);
		list+=new MenuItem_("clear obscure ownership", removeAllPolyMidPointsOwnership, menu);
		list+=new MenuItem_("clear ownership", clearOwnership, menu);
		if(Window::animMode()){
			list+=new MenuItem_("capture IK points", manualCaptureIKPoints, menu);
		}
		else{
			list+=new MenuItem_("paste to poly map", pasteToPolyMap, menu);
			list+=new MenuItem_("mirror poly map coords", mirrorHalfPolyMap, menu);
			list+=new MenuItem_("add mirror half poly map", addMirrorHalfPolyMap, menu);
			list+=new MenuItem_("delete mirror poly map", deleteHalfPolyMap, menu);
		}
		list+=new Seperator(menu);
		extra=true;
	}
	if(dynamic_cast<ImmediatePart*>(g_selNode.getRaw()) || dynamic_cast<Morpher*>(g_selNode.getRaw())){
		BranchMenuItem* import=new BranchMenuItem(Vec2i(16,16), Vec2i(464,80), menu_import, menu);
		list+=import;
		import->menuPtr->list+=new MenuItem_(menu_shape_from_file, importPart, *(import->menuPtr));
		BranchMenuItem* exportMenu=new BranchMenuItem(menu_export, menu);
		list+=exportMenu;
		exportMenu->menuPtr->list+=new MenuItem_(menu_shape_to_file, exportPart, *(exportMenu->menuPtr));
//		exportMenu->menuPtr->list+=new MenuItem_("pivot point position to clip board", pivotPointToClipBoard, *(exportMenu->menuPtr));
		extra=true;
	}
	if(dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		list+=new MenuItem_(builtInProtoList[PART]->iconSize, builtInProtoList[PART]->iconSrcPos, menu_convert, convertToWhitePart, menu);
		BranchMenuItem* convertMap=new BranchMenuItem("convert poly map", menu);

		convertMap->menuPtr->list+=new MenuItem_("mercator to north pole", mercatorToNorthPole, *(convertMap->menuPtr));
		convertMap->menuPtr->list+=new MenuItem_("mercator to south pole", mercatorToSouthPole, *(convertMap->menuPtr));
		convertMap->menuPtr->list+=new MenuItem_("north pole to mercator", northPoleToMercator, *(convertMap->menuPtr));
		convertMap->menuPtr->list+=new MenuItem_("south pole to mercator", southPoleToMercator, *(convertMap->menuPtr));
		convertMap->menuPtr->list+=new MenuItem_("pole to pole", poleToPole, *(convertMap->menuPtr));
		list+=convertMap;
		extra=true;
	}
	if(dynamic_cast<Part*>(g_selNode.getRaw()) || dynamic_cast<Morpher*>(g_selNode.getRaw())){
		BranchMenuItem* exportMenu=new BranchMenuItem(menu_export, menu);
		list+=exportMenu;
		exportMenu->menuPtr->list+=new MenuItem_(menu_shape_to_file, exportPart, *(exportMenu->menuPtr));
		if(Art* part = dynamic_cast<Part*>(g_selNode.getRaw())){
			if(part->smls->animCP){
				BranchMenuItem* importMenu=new BranchMenuItem(menu_import, menu);
				list+=importMenu;
				importMenu->menuPtr->list+=new MenuItem_("morphed patch", importMorphedPatch, *(importMenu->menuPtr));

				exportMenu->menuPtr->list+=new MenuItem_("morphed patch", exportMorphedPatch, *(exportMenu->menuPtr));
			}
		}

		exportMenu->menuPtr->list+=new MenuItem_("pivot point position to clip board", pivotPointToClipBoard, *(exportMenu->menuPtr));

		extra=true;
	}
	if(dynamic_cast<Part*>(g_selNode.getRaw())){
		list+=new MenuItem_(builtInProtoList[IMMEDIATEPART]->iconSize, builtInProtoList[IMMEDIATEPART]->iconSrcPos, menu_convert, convertToImmediatePart, menu);
		extra=true;
	}
	else if(Seamless* smls=dynamic_cast<Seamless*>(g_selNode.getRaw())){
		//String path=g_settings->file;
		//String filename;
		//extractFileName(filename, path);
		//list+=new MenuItem_("export to " + filename, exportObj, menu);

		list+=new MenuItem_("compile build transform", compileBuildTransform, menu);
		list+=new MenuItem_(menu_scale_rotation_translation,	compileTransform, menu);
		list+=new MenuItem_("import build nodes", importBuildNodesField, menu);
		list+=new MenuItem_("import pivotPoints", importPivotPoints, menu);
		list+=new MenuItem_("import pivotPoint distances", importPivotPointDistances, menu);
		list+=new MenuItem_("create lathes from pivotPoints", createLathesFromPivotPoints, menu);
		list+=new MenuItem_(menu_convert_pivot_points_to_translations, convertPivotPointsToTranslations, menu);
		list+=new MenuItem_(menu_convert_translations_to_pivot_points, convertTranslationsToPivotPoints, menu);
		list+=new MenuItem_("untick ik fields", untickIkFields, menu);
		list+=new MenuItem_(builtInProtoList[IMMEDIATEPART]->iconSize, builtInProtoList[IMMEDIATEPART]->iconSrcPos, menu_convert_all, convertAllToImmediateParts, menu);
		list+=new MenuItem_("convert to color patches", convertToColorPatches, menu);
		list+=new MenuItem_("generate raw stitch polys", generateRawStitchPolys, menu);

		if(smls->track){
			list+=new MenuItem_("unselect track", toggleTrack,  menu);
			list+=new MenuItem_("unselect track & descendants", setTrackAndDescendants,  menu);
		}
		else{
			list+=new MenuItem_("select track",toggleTrack,  menu);
			list+=new MenuItem_("select track & descendants", setTrackAndDescendants,  menu);
		}
		list+=new MenuItem_("select track exclusive", selectTrackExclusive,  menu);
		extra=true;
	}
	else if(TextureEffect* textureEffect = dynamic_cast<TextureEffect*>(g_selNode.getRaw())){
		list+=new MenuItem_("save texture file", saveTextureFile, menu);
	}
	else if(Stem* stem=dynamic_cast<Stem*>(g_selNode.getRaw())){
		list+=new MenuItem_("store generated poly map", storeGeneratedPolyMap, menu);
		list+=new MenuItem_("paste to poly map", pasteToPolyMap, menu);
		list+=new MenuItem_("copy color to poly map", copyColorToPolyMap, menu);
		list+=new MenuItem_("export color to texture", exportColorToTexture, menu);
		list+=new MenuItem_("delete poly map", deletePolyMap, menu);
		list+=new MenuItem_("delete mirror poly map", deleteHalfPolyMap, menu);
		list+=new MenuItem_("synchronize cells", synchronizeCells, menu);
	}
	else if(Viewpoint* view=dynamic_cast<Viewpoint*>(g_selNode.getRaw())){
		list+=new MenuItem_("neutral", switchViewNeutal, menu);
		if(view->track){
			list+=new MenuItem_("unselect track", toggleTrack,  menu);
		}
		else{
			list+=new MenuItem_("select track",toggleTrack,  menu);
		}
		list+=new MenuItem_("select track exclusive", selectTrackExclusive,  menu);
		list+=new MenuItem_("clac lookAt", calcLookAtFromOrientation,  menu);
	}
	else if(MidiFormant* formant=dynamic_cast<MidiFormant*>(g_selNode.getRaw())){
		list+=new MenuItem_("change formant key", changeFormantKey, menu);
	}
	else if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		if(getHair()){
			list += new MenuItem_("repeat lock map", repeatSeperatorMap, menu);
			list += new MenuItem_("auto length strands", autoLengthStrands, menu);
			//list += new MenuItem_("copy lock map", copyLockMap, menu);

			list += new MenuItem_("merge locks", mergeLocks, menu);

			list += new MenuItem_("add regions", addRegions, menu);
			list += new MenuItem_("paste border & region", pasteBorderRegion, menu);
			//list += new MenuItem_("scale map", scaleLockMap, menu);
			//list += new MenuItem_("rotate region", rotateRegion, menu);
			list += new MenuItem_("select locks from regions", selectLocksFromRegions, menu);
			list += new MenuItem_("delete strands in sel regions", deleteStrandsInSelRegions, menu);
			list += new MenuItem_("seed strands in sel regions", seedStrandsInSelRegions, menu);
			list += new MenuItem_("connect lock to sel regions", connectLockToSelRegions, menu);
			//list += new MenuItem_("extrude region", extrudeRegion, menu);
			list += new MenuItem_("split lock", splitLock, menu);
			//list += new MenuItem_("hide bound strands", hideBoundStrands, menu);
			//list += new MenuItem_("show bound strands", showBoundStrands, menu);
			list += new MenuItem_("cut lock", cutLock, menu);
			list += new MenuItem_("insert strand points", insertStrandControlPoints, menu);
			list += new MenuItem_("insert lock", insertLock, menu);
			list += new MenuItem_("ringlet generator", ringletGenerator, menu);
			//list += new MenuItem_("thinner", thinner, menu);
			list += new MenuItem_("mirror lock", mirrorLock, menu);
			//list += new MenuItem_("generate all strands", generateAllStrands, menu);
			//list += new MenuItem_("generate only selecte strands", generateOnlySelecteStrands, menu);
			list+=new Seperator(menu);
		}

		BranchMenuItem* import=new BranchMenuItem(Vec2i(16,16), Vec2i(464,80), menu_import, menu);
		list += import;
		import->menuPtr->list+=new MenuItem_(menu_shape_from_file, importPart, *(import->menuPtr));
		BranchMenuItem* exportMenu=new BranchMenuItem(menu_export, menu);
		list += exportMenu;
		exportMenu->menuPtr->list+=new MenuItem_("Patch to file", exportPatch, *(exportMenu->menuPtr));
		list += new MenuItem_("convert to subdiv", convertToSubdiv, menu);
		int uMid = checkNurbsType(patch->dim.x, patch->uKnot);
		int vMid = checkNurbsType(patch->dim.y, patch->vKnot);
		if(NurbsPatch::MID_0 == uMid){
			list += new MenuItem_(menu_convert_to_1_u_mid_points, convert_to_1_u_mid_points, menu);
		}
		else if(NurbsPatch::MID_1 == uMid){
			list += new MenuItem_(menu_convert_to_0_u_mid_points, convert_to_0_u_mid_points, menu);
		}
		if(NurbsPatch::MID_0 == vMid){
			list += new MenuItem_(menu_convert_to_1_v_mid_points, convert_to_1_v_mid_points, menu);
		}
		else if(NurbsPatch::MID_1 == vMid){
			list += new MenuItem_(menu_convert_to_0_v_mid_points, convert_to_0_v_mid_points, menu);
		}
		list += new MenuItem_(menu_rotate_uv_90, rotate_uv_90, menu);
		list += new MenuItem_("rotate control staves", rotateCtrlStaves, menu);
		list += new MenuItem_("mirror control points", mirrorCtrlPoints, menu);
		list += new MenuItem_("calc lathe weights", calcLatheWeights, menu);
		

		//list += new MenuItem_(menu_mirror_modify_west, mirrorModifyWest, menu);
		//list += new MenuItem_("mirror all control points", mirrorAllCtrlPoints, menu);
		list += new MenuItem_("paste control points", pasteCtrlPoints, menu);
		list += new MenuItem_("paste control points mirror", pasteCtrlPointsMirror, menu);
		list += new MenuItem_("set all weights to 1", setAllWeightsToOne, menu);
		list += new MenuItem_("flip polarity", flipPolarity, menu);
		list += new MenuItem_("inside out", insideOutPatch, menu);
		list += new MenuItem_("clear ownership", clearOwnershipPatch, menu);

		extra=true;
	}
	else if(Build* build=dynamic_cast<Build*>(g_selNode.getRaw())){
		if(
			dynamic_cast<NurbsLathe*>(g_selNode.getRaw())||
			dynamic_cast<BoxBuild*>(g_selNode.getRaw())||
			dynamic_cast<RectangleBuild*>(g_selNode.getRaw())||
			dynamic_cast<BendyLathe*>(g_selNode.getRaw())||
			dynamic_cast<TorusBuild*>(g_selNode.getRaw())||
			dynamic_cast<ConeBuild*>(g_selNode.getRaw())||
			dynamic_cast<SphereBuild*>(g_selNode.getRaw())||
			dynamic_cast<CylinderBuild*>(g_selNode.getRaw())||
			dynamic_cast<BezierLatheBuild*>(g_selNode.getRaw())
			){
			list+=new MenuItem_(menu_convert_to_patch, convertToPatch, menu);
			if(dynamic_cast<NurbsLathe*>(g_selNode.getRaw())){
				list+=new MenuItem_(menu_convert_to_bendy, convertToBendy, menu);
				list+=new MenuItem_("flip polarity", flipPolarity, menu);
			}
			//list+=new MenuItem_("expose Stems", exposeStem, menu);
		}
		list+=new MenuItem_(menu_fast_build, fastBuild, menu);
		list+=new MenuItem_(menu_full_build, fullBuild, menu);
		list+=new MenuItem_("show cages", showCages, menu);
		list+=new MenuItem_("hide cages", hideCages, menu);
		list+=new MenuItem_("add selected patches", addSelectedPatchesToGroup, menu);
		list+=new MenuItem_("mirror patches", mirrorPatches, menu);
		list+=new MenuItem_("unmirror patches", unmirrorPatches, menu);
		if(Hair* hair = getHair()){
			//list += new MenuItem_("generate all strands", generateAllStrands, menu);
			//list += new MenuItem_("generate only caged strands", generateOnlySelecteStrands, menu);
			list += new MenuItem_("seed", seed, menu);
			list += new MenuItem_("seed 5 strands", seed5Strands, menu);
			list += new MenuItem_("delete unbound strands", deleteUnboundStrands, menu);
			list += new MenuItem_("create locks from sel strands", createLocksFromStrands, menu);
			list += new MenuItem_("auto length strands", autoLengthStrands, menu);
		}
		extra=true;
	}
	else if(Anim* anim=dynamic_cast<Anim*>(g_selNode.getRaw())){
		list+=new MenuItem_("output screenshot movie", outputScreenShotMovie, menu);
		list+=new MenuItem_(menu_output_povray_movie, outputPovRayMovie, menu);
		list+=new MenuItem_(menu_output_batch_file, outputBatchFile, menu);
		list+=new MenuItem_("output coord interpolator", outputVrmlMorph, menu);
		list+=new MenuItem_(menu_reduce_captured_poses, reduceCapturedPoses, menu);
		list+=new MenuItem_("export to bvh", exportAnimToBvh, menu);
		list+=new MenuItem_("set period from video", setPeriodFromVideo, menu);
		if(Anim::autoCapture){
			list+=new MenuItem_("disable auto capture", toggleAnimNeutral, menu);
		}
		else{
			list+=new MenuItem_("enable auto capture", toggleAnimNeutral, menu);
		}

		extra=true;
	}
	else if(PositionInterpolator* pi=dynamic_cast<PositionInterpolator*>(g_selNode.getRaw())){
		list+=new MenuItem_("convert to NurbsTrack", convertToNurbsTrack, menu);

		extra=true;
	}
	else if(NurbsTrack* pi=dynamic_cast<NurbsTrack*>(g_selNode.getRaw())){
		list+=new MenuItem_("new curve segment", newNurbsTrackSegment, menu);

		extra=true;
	}
	if(g_selNode->getType()==INDEXEDFACESET){
		BranchMenuItem* import=new BranchMenuItem(Vec2i(16,16), Vec2i(464,80), menu_import, menu);
		import->menuPtr->list+=new MenuItem_(menu_IndexedFaceSet_from_file, importIndexedFaceSet, *(import->menuPtr));
		extra=true;
	}
	else if(g_selNode->getType() == TRANSFORM || g_selNode->getType() == _GROUP || dynamic_cast<Shape*>(g_selNode.getRaw())){
		if(g_selNode->getType()==TRANSFORM){
			list+=new MenuItem_(menu_beam_to, beamTo, menu);
		}
		list+=new MenuItem_(menu_convertToSeamless, convertGroupToSmls, menu);
		extra=true;
	}
	else if(dynamic_cast<HAnimHumanoid*>(g_selNode.getRaw())){
		list+=new MenuItem_(menu_convertToSeamless, convertHAnimToSmls, menu);
		extra=true;
	}
	else if(dynamic_cast<TimeSensor*>(g_selNode.getRaw())){
		list+=new MenuItem_(menu_convertToAnim, convertTimeSensorToAnim, menu);
		extra=true;
	}

	if(dynamic_cast<Scene*>(g_selNode.getRaw())){
		list+=new MenuItem_("convert to C names", convertToCNames, menu);
		extra=true;
	}
	if(dynamic_cast<AGroup*>(g_selNode.getRaw())){
		list+=new MenuItem_(menu_optimize, optimize, menu);
		extra=true;
	}
	restoreChildrenLookUp(parentNode, remproptNodeLenA);
	restoreChildrenLookUp(g_selNode.getRaw(), remproptNodeLenB);
	if(g_selNode->getType()==SCENE){
		list+=new Seperator(menu);
	}
	{
		RotAccelerator* rotAccelerator=dynamic_cast<RotAccelerator*>(g_selNode.getRaw());
		if(rotAccelerator!=0){
			list+=new MenuItem_(menu_calculate_acceleration_for_return, homeToInit, menu);
		}
	}
	{
		if(dynamic_cast<EC2*>(g_selNode.getRaw())){
			list+=new MenuItem_(menu_stretchCurve, stretchCurve, menu);
		}
	}
}



void Menu::paint(){
	//clear();

	int dstWidth=size.x;
	int dstHeight=size.y;
	int* dstPix=pixBuf.getArray();

	const int edgeSrcY=200;

	paintColorSweepForCurvedBar(&iconSrcBuf[(iconSrcBufCY-1-edgeSrcY)*480+334+8], &iconSrcBuf[(iconSrcBufCY-1-edgeSrcY)*480+334+9], 9, 0, Vec2i(dstWidth-9-9, 9), dstWidth, dstHeight, dstPix);
	pasteSprite2( 0, 0, 334, edgeSrcY, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 9, 9, dstPix);
	pasteSprite2( dstWidth-9, 0, 334+9, edgeSrcY, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 9, 9, dstPix);

	blitHorizontalColorSweep(Vec2i(9, dstHeight-9), Vec2i(316+8, edgeSrcY), Vec2i(316+9, edgeSrcY), Vec2i(dstWidth-9-9, 9));
	pasteSprite2( 0, dstHeight-9, 316, edgeSrcY, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 9, 9, dstPix);
	pasteSprite2( dstWidth-9, dstHeight-9, 316+9, edgeSrcY, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 9, 9, dstPix);

	paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1-edgeSrcY)*480+316],		9,	0,			9, dstHeight-9, dstWidth, dstHeight, dstPix);
	paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1-edgeSrcY)*480+316+9],	9,	dstWidth-9,	9, dstHeight-9, dstWidth, dstHeight, dstPix);

	paintBlock2(0xefefef, 9,9, dstWidth-9, dstHeight-9, dstWidth, dstHeight, dstPix);

	if(blisterItem != -1){
		list[blisterItem]->paintHoverBlister();
	}
	for(int i=0; i < list.length; i++){
		list[i]->paint();
	}
	if(g_settings->invertWindowColor){
		invert();
	}
	updateShaderTexture = true;
}
void Menu::buttonHover(Vec2i ptCursor){
	for(int k=list.length; --k>=0;){
		MenuItem* menuItem=list[k];
		if(menuItem->initHover(ptCursor)){
			blisterItem = k;
			paint();
			g_remMenu->ensurePaintIntegrity(&menuItem->win);
			return;
		}
	}
}
void Menu::autoSizeAndStuff(){
	blisterItem =-1;
	openedBranchItem=0;

	int numOfColums=1;
	int lastBeginIndex=0;
	int margin=0;
	const int itemHeight = 22;

	int width=0;
	for(int i=list.length; --i>=0;){
		int length = list[i]->label.length;
		if(width < length){
			width=length;
		}
	}

	width=width*8+32;
	int yPos=0;

	for(int i=list.length; --i>=0;){
		MenuItem* menuItem=list[i];
		if(dynamic_cast<Seperator*>(menuItem)){
			menuItem->bSize=Vec2i(width+15, 2);
			yPos+=2;
		}
		else{
			menuItem->bSize=Vec2i(width+15, itemHeight);
			yPos+=itemHeight;
		}
	}
	int guessHeight=yPos;
	Rect rect; getClientRect(rect);
	if(rect.bottom<yPos){
		guessHeight = (float)yPos/(float)rect.bottom+1.0f;
		guessHeight = (float)yPos/(float)guessHeight;
	}
	guessHeight+=20;


	yPos=0;
	int realHeight=0;

	for(int i=0;;){
		MenuItem* menuItem=list[i];
		menuItem->dstPos=Vec2i(margin, yPos+2);
		if(dynamic_cast<Seperator*>(menuItem)){
			yPos+=2;
		}
		else{
			yPos+=itemHeight;
		}
		if(yPos>realHeight){
			realHeight=yPos;
		}
		if(++i >= list.length){
			break;
		}
		if(yPos + itemHeight > guessHeight){
			yPos=0;
			numOfColums++;
			margin+=width;
		}
	}
	Window::size=Vec2i(width*numOfColums+15, realHeight+10);
	paint();
	Window::topWindow=this;
}
void Menu::ensurePaintIntegrity(const Menu* selectedMenu){
	if(isVisible()){
		if(openedBranchItem){
			if(selectedMenu != this){
				int localBlisterItem=openedBranchItem->getIndex(list);
				if(localBlisterItem != blisterItem){
					blisterItem=localBlisterItem;
					paint();
				}
			}
			openedBranchItem->menuPtr->ensurePaintIntegrity(selectedMenu);
		}
		else if(selectedMenu != this && blisterItem != -1){
			blisterItem=-1;
			paint();
		}
	}
}
bool CaretWindow::isInCell(const Vec2i localCursor, int& selFrom, int& selLen, int& newCaretPos)const{
	if(localCursor.y < cellPos.y + 13 - 16 || localCursor.y > cellPos.y + 13){
		return false;
	}

	selLen = abs(caretPos-highlightPos);
	selFrom = highlightPos;
	if(caretPos < selFrom){
		selFrom = caretPos;
	}

	if(selLen > 0){
		int newCaretPos=(localCursor.x+charWidth/2-(cellPos.x))/charWidth;

		int pixPosFrom = cellPos.x+selFrom*charWidth;
		int pixPosTo = pixPosFrom+selLen*charWidth;
		if(localCursor.x >= pixPosFrom && localCursor.x < pixPosTo){
			return true;
		}
	}
	return false;
}


bool AnimControlPanel::rightClick(const Vec2i ptCursor){
	if(ptCursor >= pos && ptCursor < pos+size){
		const Vec2i localCursor = ptCursor-pos;

		Window::activeWindow=this;

		g_menuGroupIndex=-1;

		g_remMenu=new Menu;
		M_<MenuItemPtr>& list=g_remMenu->list;
		g_remMenu->list.length=0;

		list+=new MenuItem_("output screenshot movie", outputScreenShotMovie, *g_remMenu);
		list+=new MenuItem_(menu_output_povray_movie, outputPovRayMovie, *g_remMenu);
		list+=new MenuItem_(menu_output_batch_file, outputBatchFile, *g_remMenu);
		list+=new MenuItem_("output coord interpolator", outputVrmlMorph, *g_remMenu);
		list+=new MenuItem_("export to bvh", exportAnimToBvh, *g_remMenu);
		list+=new Seperator(*g_remMenu);
		list+=new MenuItem_(menu_reduce_captured_poses, reduceCapturedPoses, *g_remMenu);
		list+=new MenuItem_("set period from video", setPeriodFromVideo, *g_remMenu);
		if(Anim::autoCapture){
			list+=new MenuItem_("disable auto capture", toggleAnimNeutral, *g_remMenu);
		}
		else{
			list+=new MenuItem_("enable auto capture", toggleAnimNeutral, *g_remMenu);
		}
		Node* node = g_selNode.getRaw();
		if(Art* part=dynamic_cast<Art*>(node)){
			list+=new Seperator(*g_remMenu);
			list+=new MenuItem_("capture IK points", manualCaptureIKPoints, *g_remMenu);
			list+=new MenuItem_("snap IK point to joint", manualCaptureIKPoint, *g_remMenu);
			list+=new MenuItem_("set compensation", beginCompensation, *g_remMenu);
			list+=new MenuItem_("remove compensation", removeCompensation, *g_remMenu);
			list+=new MenuItem_("snap niki point", snapNikiPoint, *g_remMenu);
			list+=new MenuItem_("display niki panel", displayNikiPanel, *g_remMenu);
			list+=new MenuItem_("add niki", addNiki, *g_remMenu);
			list+=new Seperator(*g_remMenu);
			if(part->track){
				list+=new MenuItem_("unselect track", toggleTrack,  *g_remMenu);
				list+=new MenuItem_("unselect track & descendants", setTrackAndDescendants,  *g_remMenu);
			}
			else{
				list+=new MenuItem_("select track",toggleTrack,  *g_remMenu);
				list+=new MenuItem_("select track & descendants", setTrackAndDescendants,  *g_remMenu);
			}
			list+=new MenuItem_("select track exclusive", selectTrackExclusive,  *g_remMenu);
			list+=new MenuItem_("select model tracks", selectModelTracks,  *g_remMenu);
			list+=new MenuItem_("select model tracks exclusive", selectModelTracksExclusive,  *g_remMenu);
			list+=new MenuItem_("select all scene tracks", selectSceneTracks,  *g_remMenu);
		}
		else if(Seamless* smls = dynamic_cast<Seamless*>(node)){
			list+=new Seperator(*g_remMenu);
			if(smls->track){
				list+=new MenuItem_("unselect track", toggleTrack,  *g_remMenu);
				list+=new MenuItem_("unselect track & descendants", setTrackAndDescendants,  *g_remMenu);
			}
			else{
				list+=new MenuItem_("select track",toggleTrack,  *g_remMenu);
				list+=new MenuItem_("select track & descendants", setTrackAndDescendants,  *g_remMenu);
			}
			list+=new MenuItem_("select track exclusive", selectTrackExclusive,  *g_remMenu);
		}
		else if(Viewpoint* view=dynamic_cast<Viewpoint*>(g_selNode.getRaw())){
			list+=new Seperator(*g_remMenu);
			if(view->track){
				list+=new MenuItem_("unselect track", toggleTrack,  *g_remMenu);
			}
			else{
				list+=new MenuItem_("select track",toggleTrack,  *g_remMenu);
			}
			list+=new MenuItem_("select track exclusive", selectTrackExclusive,  *g_remMenu);
		}

		list+=new Seperator(*g_remMenu);
		list+=new MenuItem_("timeline filter", openAnimEditOptions, *g_remMenu);
		list+=new MenuItem_("select all of timeline", highlightWholeTimeLine, *g_remMenu);

		{
			BranchMenuItem* selAnimMenu = new BranchMenuItem("select anim bar", *g_remMenu);
			list += selAnimMenu;
			M_<Ptr<MenuItem> >& selAnimList = selAnimMenu->menuPtr->list;

			for(int k = 0; k < g_animBarList.length; k++){
				selAnimList += new MenuItem_(g_animBarList[k]->name, selectAnimBarFromList, *(selAnimMenu->menuPtr));
			}
			g_selAnimMenu = selAnimMenu->menuPtr;
		}

		g_remMenu->autoSizeAndStuff();
		g_remMenu->pos=ptCursor-Vec2i(0, g_remMenu->size.y);
		return true;
	}
	return false;
}

bool ControlPanel::rightClick(const Vec2i ptCursor){
	if(ptCursor >= pos && ptCursor < pos+size){

		const Vec2i localCursor = ptCursor-pos;

		Window::activeWindow=this;

		Field *fieldPtr;
		g_menuGroupIndex=-1;

		g_remMenu=new Menu;
		M_<MenuItemPtr>& list=g_remMenu->list;
		g_remMenu->list.length=0;

		int selFrom=0;
		int selLen=0;
		int newCaretPos=0;
		if(textCell.isInCell(localCursor, selFrom, selLen, newCaretPos) && selLen > 0){
			list+=new MenuItem_("cut  (Ctrl X)", cut, *g_remMenu);
			list+=new MenuItem_(Vec2i(16,16), Vec2i(remember_x, remember_y), menu_short_copy, remember, *g_remMenu);
			list+=new MenuItem_(Vec2i(16,16), Vec2i(16, 192), "paste (Ctrl V)", pasteFromClipboard, *g_remMenu);
			g_remMenu->autoSizeAndStuff();
			g_remMenu->pos=ptCursor;
			return true;
		}
		else{
			int newCaretPos=0;
			for(int k=buttonList.length; --k>=0;){
				FieldButton* button=buttonList[k];
				if(button->checkClickTextCell(localCursor, newCaretPos)){
					if(button->editMode==false){
						textCell.cellActive=false;
						textCell.selectField(*button);
					}
					textCell.highlightPos=textCell.caretPos=newCaretPos+textCell.firstVisChar;
					textCell.updateCaret();

					paint();

					list+=new MenuItem_(Vec2i(16,16), Vec2i(16, 192), "paste (Ctrl V)", pasteFromClipboard, *g_remMenu);
					g_remMenu->autoSizeAndStuff();
					g_remMenu->pos=ptCursor;
					return true;
				}
			}
		}

		int routeYPos=startFieldLineYpos;
		for(int k=0; k<buttonList.length; k++){
			FieldButton* button=buttonList[k];
			if(routeYPos < button->dstPos.y){
				routeYPos = button->dstPos.y;
			}
			if(dynamic_cast<IndexButton*>(button) || dynamic_cast<LengthButton*>(button)){
				continue;
			}
			if(	localCursor.y > button->dstPos.y && localCursor.y < button->dstPos.y + button->bSize.y){
				/*
				if(i>=g_selNode->getFLen()){// must have clicked on a ROUTE line
					g_routeIndex=i-g_selNode->getFLen();
					removeRoutePopupMenu(parent, x, y);
					return;
				}
				*/
				fieldPtr=&button->bField;
				g_fromFieldIdx=node->getFieldIndex(fieldPtr);
	//			g_routeIndex=;

				ProtoInterface* protoInterface=0;
				ProtoDeclare* protoDeclare=getSceneProtoDeclare(g_stw->selLine);
				if(protoDeclare){
					protoInterface=protoDeclare->protoInterface;
				}
				int cntr=0;
				if(button->sprop.groupIndex!=-1){
					g_menuGroupIndex=button->sprop.groupIndex;
				}
				Ptr<BranchMenuItem> menuSim = addFieldMenu(*g_remMenu, fieldPtr, node->getFieldIndex(fieldPtr), button->sprop, protoInterface);
				if(menuSim->menuPtr->list.length){
					list+=menuSim.p;
				}
				if(fieldPtr->isRef()){
					g_remMenu->list+=new MenuItem_(menu_undo_IS, undoIs, *g_remMenu);
				}

				if(g_remMenu->list.length==0){
					g_remMenu=0;
					return false;
				}
				g_remMenu->autoSizeAndStuff();

				g_remMenu->pos=ptCursor;
				return true;
			}
		}
		routeYPos += 25;
		for(int k=0; k<node->getFLen(); k++){
			Field* field=node->getField(k);
			M_Route& m_route = field->refRoute();
			for(int r=0; r<m_route.length; r++){
				if(localCursor.y >= routeYPos && localCursor.y < routeYPos+pixSymbolCY){
					S_Route& s_route=m_route[r];
					list+=new MenuItem_int_int(menu_delete_this_ROUTE, deleteRoute, k, r, *g_remMenu);
					g_remMenu->autoSizeAndStuff();
					g_remMenu->pos=ptCursor;
					return true;
				}
				routeYPos+=pixSymbolCY;
			}
		}
		return true;
	}
	return false;
}
void openChatLog();

void searchWebFromCell0();
void searchWebFromCell1();
void searchWebFromCell2();

typedef void (*i_function)();

i_function searchWebFuncFromCell[3]={
	searchWebFromCell0,
	searchWebFromCell1,
	searchWebFromCell2
};
void ChatWin::CWTextCell::onRightClick(const Vec2i ptCursor){
	const Vec2i localCursor=ptCursor-win.pos;
	g_remMenu = new Menu;
	M_<Ptr<MenuItem> >& list = g_remMenu->list;
	list+=new MenuItem_("cut  (Ctrl X)", cut, *g_remMenu);
	list+=new MenuItem_(Vec2i(16,16), Vec2i(remember_x, remember_y), menu_short_copy, remember, *g_remMenu);
	list+=new MenuItem_(Vec2i(16,16), Vec2i(16, 192), "paste (Ctrl V)", win.pasteFromClipboard, *g_remMenu);
	list+=new MenuItem_("open chat log", openChatLog, *g_remMenu);

	int selFrom=0;
	int selLen=0;
	int newCaretPos=0;
	if(isInCell(localCursor, selFrom, selLen, newCaretPos) && selLen > 0){
		for(int k = 0; k < g_settings->searchEngine.length; k++){
			const String& searchEngine = g_settings->searchEngine[k];
			int i = searchEngine.indexOf("|");
			String str = searchEngine.substring(0, i - 1);
			str+=" \"";
			str+=cellStr.substr(selFrom, selLen);
			str+='\"';
			list+=new MenuItem_(str, searchWebFuncFromCell[k], *g_remMenu);
		}
	}
	else{
		cellActive = false;
		int newCaretPos=(localCursor.x+charWidth/2-(cellPos.x))/charWidth;
		highlightPos=caretPos=newCaretPos+firstVisChar;
		updateCaret();
		win.clearHighLight();

		win.paint();
	}

	g_remMenu->autoSizeAndStuff();
	g_remMenu->pos=ptCursor;
}
void searchWeb0();
void searchWeb1();
void searchWeb2();

typedef void (*i_function)();

i_function searchWebFunc[3]={
	searchWeb0,
	searchWeb1,
	searchWeb2
};

bool ChatWin::rightClick(const Vec2i ptCursor){
	if(ptCursor >= pos && ptCursor < pos+size){

		const Vec2i localCursor=ptCursor-pos;

		if(	localCursor.x >= textCell.cellPos.x && localCursor.x < textCell.cellPos.x + textCell.cellSize.x &&
			localCursor.y >= textCell.cellPos.y && localCursor.y < textCell.cellPos.y + textCell.cellSize.y
			){
			textCell.onRightClick(ptCursor);
			return true;
		}

		g_remMenu = new Menu;
		M_<Ptr<MenuItem> >& list = g_remMenu->list;
		list+=new MenuItem_(Vec2i(16,16), Vec2i(remember_x, remember_y), menu_short_copy, remember, *g_remMenu);
		list+=new MenuItem_("open chat log", openChatLog, *g_remMenu);

		String sel;
		getHighlightedString(sel);
		if(sel.length){
			for(int k = 0; k < g_settings->searchEngine.length; k++){
				const String& searchEngine = g_settings->searchEngine[k];
				int i = searchEngine.indexOf("|");
				String str = searchEngine.substring(0, i - 1);
				str+=" \"";
				getHighlightedString(str);
				str+='\"';
				list+=new MenuItem_(str, searchWebFunc[k], *g_remMenu);
			}
		}

	//	list+=new MenuItem_(Vec2i(16,16), Vec2i(paste_x, paste_y), menu_short_copy, remember, *g_remMenu);

		g_remMenu->autoSizeAndStuff();
		g_remMenu->pos=ptCursor;
		return true;
	}
	return false;
}


void helpMenu(){
	const Vec2i localCursor = g_mousePos - g_tb->pos;
	g_dontCloseMenu = true;
	g_remMenu = new Menu;
	M_<Ptr<MenuItem> >& list = g_remMenu->list;

	list += new MenuItem_(Vec2i(24,24), Vec2i(help_x,help_y), menu_tutorials, help, *g_remMenu);
	list += new MenuItem_(menu_about_seamless3d,		about, *g_remMenu);
	list += new Seperator(*g_remMenu);
	//list += new MenuItem_("YouTube Tut",		InputEvent::preYouTubeTut, *g_remMenu);
	//list += new MenuItem_(menu_beeline_anim_demo,		InputEvent::preBeelineAnimDemo, *g_remMenu);
	//list += new MenuItem_("model and animate a character demo",		InputEvent::preNurbsLatheAvDemo, *g_remMenu);
	//list += new MenuItem_("show art demo",		InputEvent::preYouTubeTut, *g_remMenu);
	//list += new MenuItem_("model a head demo",		InputEvent::preNurbsStitchedHeadDemo, *g_remMenu);
	//list += new MenuItem_("animate a limb demo",		InputEvent::preLimbAnimDemo, *g_remMenu);
	//list += new MenuItem_("multi control point edit demo",		InputEvent::preMultiCtrlPointEditDemo, *g_remMenu);
	//list += new MenuItem_(menu_worm_animation_demo,		InputEvent::preWormAnimationDemo, *g_remMenu);
	//list += new MenuItem_(menu_fuse_surface_avatar_demo,	InputEvent::preNurbsPatchAvDemo, *g_remMenu);
	//list += new MenuItem_("nurbs surface poly editing",		preNurbsSurfacePolyEditingDemo, *g_remMenu);
	//list += new MenuItem_(menu_switch_anim_in_3d_demo,		InputEvent::preSwitchAnimIn3dDemo, *g_remMenu);
	//list += new MenuItem_(menu_textured_object_demo,	InputEvent::preBuildTexturedObjectDemo, *g_remMenu);
	//list += new MenuItem_(menu_poly_radial_tug_demo,		InputEvent::preImmediateRadialTugDemo, *g_remMenu);
	//list += new MenuItem_("poly model avatar demo",	InputEvent::preImmediateAvDemo, *g_remMenu);
	//list += new MenuItem_(menu_bring_static_model_to_life_demo,	InputEvent::preBringToLifeDemo, *g_remMenu);
	//list += new MenuItem_("break join & color poly demo",		preBreakAndJoinVerticesDemo, *g_remMenu);

	g_remMenu->autoSizeAndStuff();
//	g_remMenu->pos = g_mousePos - g_remMenu->size;
	g_remMenu->pos = g_mousePos - Vec2i(g_remMenu->size.x, 0);
}


void fileMenu(){
	const Vec2i localCursor = g_mousePos - g_tb->pos;
	g_dontCloseMenu = true;
	g_remMenu = new Menu;
	M_<Ptr<MenuItem> >& list = g_remMenu->list;
	
	BranchMenuItem* importMenu = new BranchMenuItem(Vec2i(16,16), Vec2i(464,80), menu_import, *g_remMenu);
	g_remMenu->list += importMenu;
	M_<Ptr<MenuItem> >& importList = importMenu->menuPtr->list;
	importList += new MenuItem_(menu_import_bxx_av_studio,		importBxxAvStudio, *(importMenu->menuPtr));
	importList += new MenuItem_(menu_Anim_nodes_from_file, importAnimation, *(importMenu->menuPtr));
	importList += new MenuItem_("anim nodes & recalc", importAnimRecalc, *(importMenu->menuPtr));
	importList += new MenuItem_(menu_all_nodes_from_file, importAllNodesField, *(importMenu->menuPtr));
	importList += new MenuItem_("build nodes", importBuildNodesField, *(importMenu->menuPtr));
	importList += new MenuItem_("pivotPoints", importPivotPoints, *(importMenu->menuPtr));
	importList += new MenuItem_("pivotPoint distances", importPivotPointDistances, *(importMenu->menuPtr));
	if(dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Button::str = menu_import_to_Part;
		Button::str += g_selNode->name;
		importList += new MenuItem_(Vec2i(24,24), Vec2i(loadPartFromFile_x,loadPartFromFile_y), Button::str,	importPart, *(importMenu->menuPtr));
	}

	BranchMenuItem* exportMenu = new BranchMenuItem(Vec2i(16,16), Vec2i(464,80), menu_export, *g_remMenu);
	g_remMenu->list += exportMenu;
	M_<Ptr<MenuItem> >& exportList = exportMenu->menuPtr->list;
	exportList += new MenuItem_("movie",		exportMovie, *(exportMenu->menuPtr));
	exportList += new MenuItem_("obj",		exportObj, *(exportMenu->menuPtr));
	exportList += new MenuItem_("json",		exportWebGL, *(exportMenu->menuPtr));
	exportList += new MenuItem_("vrml/x3d",		exportVrmlX3d, *(exportMenu->menuPtr));
	exportList += new MenuItem_("cpp node bind",		outputNodeBind, *(exportMenu->menuPtr));
	if(dynamic_cast<Art*>(g_selNode.getRaw())){
		Button::str = menu_export_from_Part;
		Button::str += g_selNode->name;
		exportList += new MenuItem_(Vec2i(24,24), Vec2i(savePartAs_x,savePartAs_y), Button::str,	exportPart, *(exportMenu->menuPtr));
	}

	int fileLen = g_settings->file.refValue().length;
	if(fileLen){
		BranchMenuItem* recentBranch = new BranchMenuItem("recent files", *g_remMenu);
		list += recentBranch;
		M_<Ptr<MenuItem> >& recentList = recentBranch->menuPtr->list;
		{
			if(fileLen > 6){
				fileLen = 6;
			}
			for(int i = 0; i < fileLen; i++){
				String filename;
				String ext;
				splitPath(g_settings->file[i], String(), String(), filename, ext);
				filename += ext;
				recentList += new MenuItem_int(filename,	openRecentMenu, i, *(recentBranch->menuPtr));
			}
		}
	}

	list += new MenuItem_(Vec2i(24,24), Vec2i(openSmls_x,openSmls_y), menu_open,	openSmls, *g_remMenu);
	list += new MenuItem_(Vec2i(24,24), Vec2i(openNetFile_x,openNetFile_y),	menu_open_from_web,openNetFile, *g_remMenu);
	list += new MenuItem_(Vec2i(24,24), Vec2i(saveSmls_x,saveSmls_y), menu_save,	saveSmls, *g_remMenu);
	list += new MenuItem_(menu_save_as,				saveSmlsAs, *g_remMenu);
	list += new MenuItem_(Vec2i(24,24), Vec2i(newSmls_x,newSmls_y), menu_new,				newSmls, *g_remMenu);


	g_remMenu->autoSizeAndStuff();
	//g_remMenu->pos = g_mousePos - Vec2i(0, g_remMenu->size.y);
	g_remMenu->pos = g_mousePos;
}

void moreModellingMenu(){
	const Vec2i localCursor = g_mousePos - g_tb->pos;
	g_dontCloseMenu = true;
	g_remMenu = new Menu;
	M_<Ptr<MenuItem> >& list = g_remMenu->list;

	if(dynamic_cast<NurbsLine*>(g_selNode.getRaw())){
		list += new MenuItem_("convert lines to hair",	convertNurbsLinesToStrands, *g_remMenu);
	}
	list += new MenuItem_("subdive patches",		subdivPatches, *g_remMenu);
	list += new MenuItem_("create double surface",	createDoubleSurface, *g_remMenu);
	list += new MenuItem_("fillet double surface edge",	fillet, *g_remMenu);
	list += new MenuItem_("rotate control staves",	rotateCtrlStaves, *g_remMenu);
	list += new MenuItem_("rotate uv points 90",	rotate_uv_90, *g_remMenu);
	list += new MenuItem_("align selecteded points (Shift A)", alignMidSelectededPoints, *g_remMenu);
	list += new MenuItem_("hide/show mesh", hideShowMesh, *g_remMenu);
	list += new MenuItem_("paste control column", pasteColumn, *g_remMenu);
	list += new MenuItem_("transfer selected points", transferSelectedPoints, *g_remMenu);
	//list += new MenuItem_("move selected control point to nearest other (M)", moveVertex, *g_remMenu);
	list += new MenuItem_("paste control points", pasteCtrlPoints, *g_remMenu);
	list += new MenuItem_("radical patch", toggleTransformFallOff, *g_remMenu);
	list += new MenuItem_("smoothing patch", simplifyCurve, *g_remMenu);
	list += new MenuItem_("delete selected patches", deleteSelectedPatches, *g_remMenu);
	list += new MenuItem_("delete polies for selected patches", deletePoliesOwnedByPatch, *g_remMenu);
	list += new MenuItem_("generate polies for selected patches (Ctrl F4)", bnspeBuildPolies, *g_remMenu);
	//list += new MenuItem_(Vec2i(24,24), Vec2i(24, 305 - 24), "toggle zbuffer cage mode (Space)", toggleZBufferCageMode, *g_remMenu);


	g_remMenu->autoSizeAndStuff();
	g_remMenu->pos = g_mousePos - g_remMenu->size;
}
