/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/


#include <windows.h>

#include <sys/stat.h>
#include <share.h>
#include <fcntl.h>
#include <io.h>
#include <algorithm>    // std::max
using namespace std;

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "seamless3d_gui_render.h"
#include "gui.h"
#include "scene_tree.h"
#include "robot_aid.h"

bool clickLine(Vec2f click, Vec2f ptA, Vec2f ptB, Vec2f& crossPoint);
int getCorner(int corner, int offset, int len);
int getIndexContainedInMidPoint(const M_CtrlPolyPoint& ctrlPolyPoint, const int absoluteIdx, const NurbsPatch* checkPatch);
bool getMatchingMidPoint(const M_CtrlPolyPoint& ctrlPolyPoint, int absoluteIdx, const NurbsPatch* checkPatch, int& otherIdx);
bool getMatchingMidPoint2(const M_CtrlPolyPoint& ctrlPolyPoint, int absoluteIdx, const NurbsPatch* checkPatch, int& otherIdx, const int idx2);
void uGetEdgePoint(const M_CtrlPolyPoint& ctrlPolyPoint, int absoluteIdx, const NurbsPatch* altPatch, int& altIdx, int& altIdx2, bool& altLeaveNorthOrEast);
void vGetEdgePoint(const M_CtrlPolyPoint& ctrlPolyPoint, int absoluteIdx, const NurbsPatch* altPatch, int& altIdx, int& altIdx2, bool& altLeaveNorthOrEast);
NurbsPatch* anotherPatchContainedInMidPoint(const M_CtrlPolyPoint& ctrlPolyPoint, int midIdx, int centerIdx, NurbsPatch* checkPatch, int& otherMidIdx, int& otherCenterIdx);
NurbsPatch* anotherPatchContainedInMidPoint(const M_CtrlPolyPoint& ctrlPolyPoint, const NurbsPatch* checkPatch, int centerPoly, int midPoly, int& otherIdx, int& otherIdx2);
//NurbsPatch* anotherPointSamePatchContainedInMidPoint(const M_CtrlPolyPoint& ctrlPolyPoint, const NurbsPatch* checkPatch, int centerPoly, int midPoly, int& otherIdx, int& otherIdx2, const int checkIdx, const int checkIdx2);
bool isContainedInMidPoint(const M_CtrlPolyPoint& ctrlPolyPoint, const int absoluteIdx, const NurbsPatch* checkPatch, const int checkIdx);
void swapEastWest(M_Insert& m_insert);
bool clickLine(Vec2f click, Vec2f ptA, Vec2f ptB, Vec2f& crossPoint);
inline int evenEdge(int par);
bool allreadyInList(NurbsPatch* p, const int uOrVIdx, bool doU, float ratio, bool invertRatio, Vec2i cutLine, const M_Insert& m_insert);

bool beginsFromCenterLineFunc(int k, const Insert& insert);
bool lastEndsAtMirrorCenter(const M_Insert& m_insert);
void checkForLoop();
int getCenterBegin(const M_CtrlPolyPoint& ctrlPolyPoint, int poly);
int getMidPointAbsoluteIdx(const int midBeginIndex, int edge, const M_CtrlPolyPoint& ctrlPolyPoint);

extern Ptr<InsertComplete> g_insertComplete;

void swapEastWestIfNecessary(Vec3f dpA, Vec3f dpB){
	if(g_insertPatch.remLeftSide){
		if(dpA.x > 0 || dpB.x > 0){
			g_insertPatch.remLeftSide = false;
			swapEastWest(g_insertPatch.m_insert);
			swapEastWest(g_insertPatch.m_insertOD);
		}
	}
	else{
		if(dpA.x < 0 || dpB.x < 0){
			g_insertPatch.remLeftSide = true;
			swapEastWest(g_insertPatch.m_insert);
			swapEastWest(g_insertPatch.m_insertOD);
		}
	}
}

void NurbsPatch::setToCenterLine(Vec2i& cutLine){
	int uDim = dim.x;
	if(mirror && mirrorPatch == 0 && ((cutLine.x >= (uDim / 2) && cutLine.y <= (uDim / 2)) || (cutLine.x <= (uDim / 2) && cutLine.y >= (uDim / 2)))){
		cutLine.y = (uDim / 2);
	}
}

void NurbsPatch::setToCenterLine0(Vec2i& cutLine){
	int uDim = dim.x;
	if(mirror && mirrorPatch == 0 && ((cutLine.x >= (uDim / 2) && cutLine.y <= (uDim / 2)) || (cutLine.x <= (uDim / 2) && cutLine.y >= (uDim / 2)))){
		if(((uDim / 2) & 1) && cutLine.x == (uDim / 2)){
			return;
		}
		cutLine.y = (uDim / 2);
	}
}

int evenEdge(int par, const NurbsPatch* patch){
	if(patch->mirror && patch->mirrorPatch == 0){
		int uDim = patch->dim.x;
		if(par == uDim / 2){
			return par;
		}
	}
	if(par & 1){
		par++;
	}
	return par;
}

void uGetOthersideIndices(const int u, int& checkIdx, int& checkIdx2, const int idx, const int idx2, const NurbsPatch* patch, const Vec2i& cutLine, const int uDim){
	int ey = evenEdge(cutLine.y);
	checkIdx = evenEdge(u) + ey * uDim;
	checkIdx2 = checkIdx;
	if(idx < idx2){
		checkIdx2 += 1;
	}
	else{
		checkIdx2 -= 1;
	}
}

void vGetOthersideIndices(const int v, int& checkIdx, int& checkIdx2, const int idx, const int idx2, const NurbsPatch* patch, const Vec2i& cutLine, const int uDim){
	int ey = evenEdge(cutLine.y);
	checkIdx = evenEdge(v) * uDim + ey;
	checkIdx2 = checkIdx;
	if(idx < idx2){
		checkIdx2 += uDim;
	}
	else{
		checkIdx2 -= uDim;
	}
}


void NurbsPatch::uInsertLine(const int firstStitchPoly, bool& foundFirstPatch, const bool leftSide, const int idx, const int idx2, const float ratio, const bool invertRatio, M_Insert& m_insert){
	int checkIdx = 0;
	int checkIdx2 = 0;
	int uDim = dim.x;
	int vDim = dim.y;
	int u = idx % uDim;
	float rat = ratio;
	bool iRat = invertRatio;
	if(idx > idx2){
		u = idx2 % uDim;
		rat = 1.0 - ratio;
		iRat = !invertRatio;
	}
	bool leaveNorthOrEast = false;
	Vec2i cutLine = Vec2i(vDim - 2, 0);
	if(idx < uDim){
		leaveNorthOrEast = true;
		cutLine = Vec2i(0, vDim - 2);
	}

	if(allreadyInList(this, u, true, rat, iRat, cutLine, m_insert)){
		return;
	}
	m_insert += Insert(this, u, true, rat, iRat, leaveNorthOrEast, cutLine);

	uGetOthersideIndices(u, checkIdx, checkIdx2, idx, idx2, this, cutLine, uDim);

	smls->insertLineFindMatchingEdge(this, checkIdx, checkIdx2, true, leaveNorthOrEast, firstStitchPoly, foundFirstPatch, leftSide, ratio, invertRatio, m_insert);
}



void NurbsPatch::vInsertLine(const int firstStitchPoly, bool& foundFirstStitchPoly, const bool leftSide, const int idx, const int idx2, const float ratio, const bool invertRatio, M_Insert& m_insert){
	int checkIdx = 0;
	int checkIdx2 = 0;
	int uDim = dim.x;
	int vDim = dim.y;
	int u = idx % uDim;
	int v = idx / uDim;
	float rat = ratio;
	bool iRat = invertRatio;
	if(idx > idx2){
		v = idx2 / uDim;
		rat = 1.0 - ratio;
		iRat = !invertRatio;
	}
	bool leaveNorthOrEast = false;
	Vec2i cutLine = Vec2i(uDim - 2, 0);
	if(u == 0){
		leaveNorthOrEast = true;
		cutLine = Vec2i(0, uDim - 2);
	}
	
	setToCenterLine(cutLine);

	if(mirror){
		float x = dpoint[idx].x;
		if(x > -CROSS_OVER_AMOUNT && x < CROSS_OVER_AMOUNT){
			return;
		}
	}
	if(allreadyInList(this, v, false, rat, iRat, cutLine, m_insert)){
		return;
	}
	m_insert += Insert(this, v, false, rat, iRat, leaveNorthOrEast, cutLine);

	if(lastEndsAtMirrorCenter(m_insert)){
		return;
	}

	if(foundFirstStitchPoly){//obsolete ?
		return;
	}
	vGetOthersideIndices(v, checkIdx, checkIdx2, idx, idx2, this, cutLine, uDim);

	smls->insertLineFindMatchingEdge(this, checkIdx, checkIdx2, false, leaveNorthOrEast, firstStitchPoly, foundFirstStitchPoly, leftSide, ratio, invertRatio, m_insert);
}

bool isInList(NurbsPatch* patch, const M_Extra& m_extra){
	for(int k = m_extra.length; --k >= 0;){
		if(m_extra[k].p == patch){
			return true;
		}
	}
	return false;
}

void addPatchAndCenterIdx(const NurbsPatch* patch, const int centerLen, int centerBegin, const M_CtrlPolyPoint& ctrlPolyPoint, M_Extra& m_extra){
	for(int k = centerLen; --k>= 0;){
		NurbsPatch* p = ctrlPolyPoint[centerBegin + k].p;
		if(p == patch){
			int i = ctrlPolyPoint[centerBegin + k].i;
			int uDim = p->dim.x;
			int u = i % uDim;
			int v = i / uDim;
			m_extra += Extra(p, Vec2i(u, v));
		}
	}
}

void getMidPointPatch(int midIndex, const M_CtrlPolyPoint& ctrlPolyPoint, const NurbsPatch* blockA, const NurbsPatch* blockB, const NurbsPatch* blockC, M_Extra& m_extra, int centerBegin, int centerLen){
	int len = ctrlPolyPoint[midIndex].len;
	for(int c = 0; c < len; c++){
		NurbsPatch* patch = ctrlPolyPoint[midIndex + c].p;
		if(patch != blockA && patch != blockB && patch != blockC && false == isInList(patch, m_extra)){
			addPatchAndCenterIdx(patch, centerLen, centerBegin, ctrlPolyPoint, m_extra);
		}
	}
}



inline int decNextEdge(int& nextEdge, const int len){
	if(--nextEdge == -1){
		nextEdge = len - 1;
	}
	return nextEdge;
}

inline int incNextEdge(int& nextEdge, const int len){
	if(++nextEdge == len){
		nextEdge = 0;
	}
	return nextEdge;
}

void addExtras(M_Extra& m_extra, NurbsPatch* entryPatch, NurbsPatch* otherPatch, NurbsPatch* altPatch, int poly, int entryPolyEdge, int otherPolyEdge, int altPolyEdge, const bool aSide){
	M_CtrlPolyPoint& ctrlPolyPoint = altPatch->smls->ctrlPolyPoint;
	int len = ctrlPolyPoint[poly].len;
	const int centerBegin = getCenterBegin(ctrlPolyPoint, poly);
	int centerLen = ctrlPolyPoint[centerBegin].len;

	int nextEdge = otherPolyEdge;

	if(aSide){
		while(decNextEdge(nextEdge, len) != altPolyEdge){
			int midIndex = getMidPointAbsoluteIdx(poly + len, nextEdge, ctrlPolyPoint);
			getMidPointPatch(midIndex, ctrlPolyPoint, entryPatch, otherPatch, altPatch, m_extra, centerBegin, centerLen);
		}
	}
	else{
		while(incNextEdge(nextEdge, len) != altPolyEdge){
			int midIndex = getMidPointAbsoluteIdx(poly + len, nextEdge, ctrlPolyPoint);
			getMidPointPatch(midIndex, ctrlPolyPoint, entryPatch, otherPatch, altPatch, m_extra, centerBegin, centerLen);
		}
	}
}

void NurbsPatch::uInsertLineDivert(const int firstStitchPoly, bool& foundFirstStitchPoly, const bool leftSide, const int idx, const int idx2, const float ratio, const bool invertRatio, M_Insert& m_insert, int insertIdx, Vec2i cutLine, int poly, int entryPolyEdge, int otherPolyEdge, int altPolyEdge, const bool aSide, int alternativeCentreIdx, NurbsPatch* otherPatch, Vec2i otherUV, NurbsPatch* prevPatch, int& beginInsertIdx){
	int checkIdx = 0;
	int checkIdx2 = 0;
	int uDim = dim.x;
	int vDim = dim.y;
	int u = idx % uDim;
	float rat = ratio;
	bool iRat = invertRatio;
	if(idx > idx2){
		u = idx2 % uDim;
		rat = 1.0 - ratio;
		iRat = !invertRatio;
	}
	bool leaveNorthOrEast = false;
	if(cutLine.x < cutLine.y){
		leaveNorthOrEast = true;
	}
	beginInsertIdx = insertIdx + 1;
	m_insert.length = beginInsertIdx;
	
	if(allreadyInList(this, u, true, rat, iRat, cutLine, m_insert)){
		return;
	}
	Insert insert = Insert(this, u, true, rat, iRat, leaveNorthOrEast, cutLine, poly, entryPolyEdge, otherPolyEdge, altPolyEdge, alternativeCentreIdx, otherPatch, otherUV);
	addExtras(insert.m_extra, prevPatch, otherPatch, this, poly, entryPolyEdge, otherPolyEdge, altPolyEdge, aSide);
	m_insert += insert;

	M_CtrlPolyPoint& ctrlPolyPoint = smls->ctrlPolyPoint;
	//int len = ctrlPolyPoint[beginPoly].len;
	//int entryPolyEdge2 = entryPolyEdge + 1;
	//if(entryPolyEdge2 == len){
	//	entryPolyEdge2 = 0;
	//}
	//MultiSelect::pointList.length = 0;
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[beginPoly + entryPolyEdge].p, ctrlPolyPoint[beginPoly + entryPolyEdge].i);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[beginPoly + entryPolyEdge2].p, ctrlPolyPoint[beginPoly + entryPolyEdge2].i);
	//houseKeepVisualClue();


	uGetOthersideIndices(u, checkIdx, checkIdx2, idx, idx2, this, cutLine, uDim);

	smls->insertLineFindMatchingEdge(this, checkIdx, checkIdx2, true, leaveNorthOrEast, firstStitchPoly, foundFirstStitchPoly, leftSide, ratio, invertRatio, m_insert);
}

void NurbsPatch::vInsertLineDivert(const int firstStitchPoly, bool& foundFirstStitchPoly, const bool leftSide, const int idx, const int idx2, const float ratio, const bool invertRatio, M_Insert& m_insert, int insertIdx, Vec2i cutLine, int poly, int entryPolyEdge, int otherPolyEdge, int altPolyEdge, const bool aSide, int alternativeCentreIdx, NurbsPatch* otherPatch, Vec2i otherUV, NurbsPatch* prevPatch, int& beginInsertIdx){
	int checkIdx = 0;
	int checkIdx2 = 0;
	int uDim = dim.x;
	int vDim = dim.y;
	int u = idx % uDim;
	int v = idx / uDim;
	float rat = ratio;
	bool iRat = invertRatio;
	if(idx > idx2){
		v = idx2 / uDim;
		rat = 1.0 - ratio;
		iRat = !invertRatio;
	}
	bool leaveNorthOrEast = false;
	if(cutLine.x < cutLine.y){
		leaveNorthOrEast = true;
	}
	beginInsertIdx = insertIdx + 1;
	m_insert.length = beginInsertIdx;

	setToCenterLine(cutLine);
	
	if(allreadyInList(this, v, false, rat, iRat, cutLine, m_insert)){
		return;
	}

	Insert insert = Insert(this, v, false, rat, iRat, leaveNorthOrEast, cutLine, poly, entryPolyEdge, otherPolyEdge, altPolyEdge, alternativeCentreIdx, otherPatch, otherUV);
	addExtras(insert.m_extra, prevPatch, otherPatch, this, poly, entryPolyEdge, otherPolyEdge, altPolyEdge, aSide);
	m_insert += insert;
	

	//M_CtrlPolyPoint& ctrlPolyPoint = smls->ctrlPolyPoint;
	//int len = ctrlPolyPoint[poly].len;
	//int entryPolyEdge2 = entryPolyEdge + 1;
	//if(entryPolyEdge2 == len){
	//	entryPolyEdge2 = 0;
	//}
	//MultiSelect::pointList.length = 0;
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[poly + entryPolyEdge].p, ctrlPolyPoint[poly + entryPolyEdge].i);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[poly + entryPolyEdge2].p, ctrlPolyPoint[poly + entryPolyEdge2].i);
	//houseKeepVisualClue();

	if(lastEndsAtMirrorCenter(m_insert)){
		return;
	}
	if(foundFirstStitchPoly){//obsolete ?
		return;
	}

	vGetOthersideIndices(v, checkIdx, checkIdx2, idx, idx2, this, cutLine, uDim);

	smls->insertLineFindMatchingEdge(this, checkIdx, checkIdx2, false, leaveNorthOrEast, firstStitchPoly, foundFirstStitchPoly, leftSide, ratio, invertRatio, m_insert);
}

//void clearPrevCrossedMiddle(M_Insert& m_insert);
void NurbsPatch::uInsertLineDivertInnerPatch(const int firstStitchPoly, bool& foundFirstPatch, const bool leftSide, const int idx, const int idx2, const float ratio, const bool invertRatio, M_Insert& m_insert, int insertIdx, Vec2i cutLine, int& beginInsertIdx, Vec2i otherUV){
	g_insertPatch.clickedPolyEdge.length = 0;

	int checkIdx = 0;
	int checkIdx2 = 0;
	int uDim = dim.x;
	int vDim = dim.y;
	int u = idx % uDim;
	float rat = ratio;
	bool iRat = invertRatio;
	if(idx > idx2){
		u = idx2 % uDim;
		rat = 1.0 - ratio;
		iRat = !invertRatio;
	}
	bool leaveNorthOrEast = false;
	if(cutLine.x < cutLine.y){
		leaveNorthOrEast = true;
	}

	beginInsertIdx = insertIdx + 1;
	m_insert.length = beginInsertIdx;

	if(allreadyInList(this, u, true, rat, iRat, cutLine, m_insert)){
		return;
	}
	m_insert += Insert(this, u, true, this, otherUV, rat, iRat, leaveNorthOrEast, cutLine);


	uGetOthersideIndices(u, checkIdx, checkIdx2, idx, idx2, this, cutLine, uDim);

	smls->insertLineFindMatchingEdge(this, checkIdx, checkIdx2, true, leaveNorthOrEast, firstStitchPoly, foundFirstPatch, leftSide, ratio, invertRatio, m_insert);
}

void NurbsPatch::vInsertLineDivertInnerPatch(const int firstStitchPoly, bool& foundFirstPatch, const bool leftSide, const int idx, const int idx2, const float ratio, const bool invertRatio, M_Insert& m_insert, int insertIdx, Vec2i cutLine, int& beginInsertIdx, Vec2i otherUV){
	g_insertPatch.clickedPolyEdge.length = 0;

	int checkIdx = 0;
	int checkIdx2 = 0;
	int uDim = dim.x;
	int vDim = dim.y;
	int u = idx % uDim;
	int v = idx / uDim;
	float rat = ratio;
	bool iRat = invertRatio;
	if(idx > idx2){
		v = idx2 / uDim;
		rat = 1.0 - ratio;
		iRat = !invertRatio;
	}

	setToCenterLine(cutLine);

	bool leaveNorthOrEast = false;
	if(cutLine.x < cutLine.y){
		leaveNorthOrEast = true;
	}

	beginInsertIdx = insertIdx + 1;
	m_insert.length = beginInsertIdx;



	if(allreadyInList(this, v, false, rat, iRat, cutLine, m_insert)){
		return;
	}
	m_insert += Insert(this, v, false, this, otherUV, rat, iRat, leaveNorthOrEast, cutLine);

	if(foundFirstPatch){//obsolete ?
		return;
	}

	vGetOthersideIndices(v, checkIdx, checkIdx2, idx, idx2, this, cutLine, uDim);

	smls->insertLineFindMatchingEdge(this, checkIdx, checkIdx2, false, leaveNorthOrEast, firstStitchPoly, foundFirstPatch, leftSide, ratio, invertRatio, m_insert);
}


bool Seamless::addDiversionMark(){
	g_insertPatch.loop = false;
	g_insertPatch.m_insert = g_insertPatch.m_insert_tempDivertBackup;
	g_insertPatch.m_insertOD = g_insertPatch.m_insertOD_tempDivertBackup;
	g_insertPatch.remLeftSide = g_insertPatch.remLeftSide_tempDivertBackup;

	g_insertPatch.beginInsertIdx = g_insertPatch.beginInsertIdx_tempDivertBackup;
	g_insertPatch.beginInsertIdxOD = g_insertPatch.beginInsertIdxOD_tempDivertBackup;
	
	g_insertPatch.diversionHover = false;
	g_viewMatrix = g_scene->getViewMatrix();

	findClickedPolyEdge();

	if(addDiversionMark2(g_insertPatch.m_insert, g_insertPatch.beginInsertIdx)){
		g_insertPatch.diversionHover = true;
	}
	else{
		g_insertPatch.remLeftSide = !g_insertPatch.remLeftSide;
		swapEastWest(g_insertPatch.m_insert);
		//swapEastWest(g_insertPatch.m_insertOD);

		findClickedPolyEdge();

		if(addDiversionMark2(g_insertPatch.m_insert, g_insertPatch.beginInsertIdx)){
			g_insertPatch.diversionHover = true;
		}
		//if(g_insertPatch.m_insertOD.length){
		//	addDiversionMark2(g_insertPatch.m_insertOD, g_insertPatch.beginInsertIdxOD);
		//}

		g_insertPatch.remLeftSide = !g_insertPatch.remLeftSide;
		swapEastWest(g_insertPatch.m_insert);
		//swapEastWest(g_insertPatch.m_insertOD);
	}
	if(g_insertPatch.m_insert.length < 2){
		g_insertPatch.diversionHover = false;
		g_insertPatch.m_insert = g_insertPatch.m_insert_tempDivertBackup;
		g_insertPatch.m_insertOD = g_insertPatch.m_insertOD_tempDivertBackup;

		g_insertPatch.addInsertVisualClue();
		return g_insertPatch.diversionHover;
	}

	g_insertPatch.m_insert_loopBackup = g_insertPatch.m_insert;
	g_insertPatch.m_insertOD_loopBackup = g_insertPatch.m_insertOD;
	checkForLoop();

	g_insertPatch.addInsertVisualClue();
	return g_insertPatch.diversionHover;
}

bool checkMarkRightangle(NurbsPatch* checkPatch, int checkIdx, int checkIdx2, int uOrV, bool doU, bool leaveNorthOrEast, const int firstStitchPoly, bool& foundFirstPatch, const bool leftSide, float ratio, const bool invertRatio, M_Insert& m_insert, int insertIdx, int& beginInsertIdx){
	Vec2f click = Vec2f(g_pointer.translation.x, g_pointer.translation.y);
	int uDim = checkPatch->dim.x;
	int vDim = checkPatch->dim.y;
	Vec2i otherUV;

	int last = 0;
	int first = 0;
	if(m_insert[insertIdx].leaveNorthOrEast){
		first = m_insert[insertIdx].cutLine.x;
		if((first & 1) == 0){
			first++;
		}
		last = m_insert[insertIdx].cutLine.y - 1;
	}
	else{
		first = m_insert[insertIdx].cutLine.y;
		if((first & 1) == 0){
			first++;
		}
		last = m_insert[insertIdx].cutLine.x - 1;
	}

	if(doU){
		for(int v = first; v < last; v += 2){
			int idxA = v * uDim + uOrV + 1;
			int idxB = (v + 2) * uDim + uOrV + 1;
			Vec3f dpA = checkPatch->dpoint[idxA];
			Vec3f dpB = checkPatch->dpoint[idxB];
			Vec3f mdpA = g_viewMatrix * dpA;
			Vec3f mdpB = g_viewMatrix * dpB;
			Vec2f ptA = Vec2f(mdpA.x, mdpA.y);
			Vec2f ptB = Vec2f(mdpB.x, mdpB.y);
			Vec2f crossPoint;
			if(clickLine(click, ptA, ptB, crossPoint)){
#ifdef PROG_ROB
				g_storeCrossPoint.set(checkPatch->fromPoint + idxA, checkPatch->fromPoint + idxB, .5);
#endif
				//swapEastWestIfNecessary(dpA, dpB);
				leaveNorthOrEast = m_insert[insertIdx].leaveNorthOrEast;

				Vec2i cutLine = Vec2i(m_insert[insertIdx].uOrVIdx, uDim - 2);


				if(leaveNorthOrEast){
					if(checkIdx < checkIdx2){
						idxB = idxA;
					}
				}
				else{
					if(checkIdx > checkIdx2){
						idxB = idxA;
					}
				}
				idxA += uDim;
				idxA--;
				idxB--;

				if(checkIdx < checkIdx2){
					otherUV.x = checkIdx % uDim;
				}
				else{
					otherUV.x = checkIdx2 % uDim;
					
				}

				otherUV.y = idxA / uDim;
				
				if(leaveNorthOrEast == false){
					otherUV.y--;
				}

				int prevCutLineB = v;
				if(idxA < idxB){
					prevCutLineB++;
				}
				m_insert[insertIdx].cutLine.y = prevCutLineB;

				checkPatch->vInsertLineDivertInnerPatch(firstStitchPoly, foundFirstPatch, leftSide, idxA, idxB, ratio, invertRatio, m_insert, insertIdx, cutLine, beginInsertIdx, otherUV);
				return true;
			}
			idxA -= 2;
			idxB -= 2;
			dpA = checkPatch->dpoint[idxA];
			dpB = checkPatch->dpoint[idxB];
			mdpA = g_viewMatrix * dpA;
			mdpB = g_viewMatrix * dpB;
			ptA = Vec2f(mdpA.x, mdpA.y);
			ptB = Vec2f(mdpB.x, mdpB.y);
			if(clickLine(click, ptA, ptB, crossPoint)){
#ifdef PROG_ROB
				g_storeCrossPoint.set(checkPatch->fromPoint + idxA, checkPatch->fromPoint + idxB, .5);
#endif
				//swapEastWestIfNecessary(dpA, dpB);
				leaveNorthOrEast = m_insert[insertIdx].leaveNorthOrEast;

				Vec2i cutLine = Vec2i(m_insert[insertIdx].uOrVIdx, 0);

				if(leaveNorthOrEast){
					if(checkIdx > checkIdx2){
						idxB = idxA;
					}
				}
				else{
					if(checkIdx < checkIdx2){
						idxB = idxA;
					}
				}
				idxA += uDim;
				idxA++;
				idxB++;


				if(checkIdx < checkIdx2){
					otherUV.x = checkIdx % uDim;
				}
				else{
					otherUV.x = checkIdx2 % uDim;
					
				}

				otherUV.y = idxA / uDim;
				
				if(leaveNorthOrEast == false){
					otherUV.y--;
				}

				int prevCutLineB = v;
				if(idxA < idxB){
					prevCutLineB++;
				}
				m_insert[insertIdx].cutLine.y = prevCutLineB;

				checkPatch->vInsertLineDivertInnerPatch(firstStitchPoly, foundFirstPatch, leftSide, idxA, idxB, ratio, invertRatio, m_insert, insertIdx, cutLine, beginInsertIdx, otherUV);
				return true;
			}
		}
	}
	else{
		for(int u = first; u < last; u += 2){
			int idxA = (uOrV + 1) * uDim + u;
			int idxB = idxA + 2;
			Vec3f dpA = checkPatch->dpoint[idxA];
			Vec3f dpB = checkPatch->dpoint[idxB];
			Vec3f mdpA = g_viewMatrix * dpA;
			Vec3f mdpB = g_viewMatrix * dpB;
			Vec2f ptA = Vec2f(mdpA.x, mdpA.y);
			Vec2f ptB = Vec2f(mdpB.x, mdpB.y);
			Vec2f crossPoint;
			if(clickLine(click, ptA, ptB, crossPoint)){
#ifdef PROG_ROB
				g_storeCrossPoint.set(checkPatch->fromPoint + idxA, checkPatch->fromPoint + idxB, .5);
#endif
				//swapEastWestIfNecessary(dpA, dpB);
				leaveNorthOrEast = m_insert[insertIdx].leaveNorthOrEast;

				Vec2i cutLine = Vec2i(m_insert[insertIdx].uOrVIdx, vDim - 2);

				if(leaveNorthOrEast){
					if(checkIdx < checkIdx2){
						idxB = idxA;
					}
				}
				else{
					if(checkIdx > checkIdx2){
						idxB = idxA;
					}
				}
				idxA++;
				idxA -= uDim;
				idxB -= uDim;

				if(checkIdx < checkIdx2){
					otherUV.y = checkIdx / uDim;
				}
				else{
					otherUV.y = checkIdx2 / uDim;
					
				}
				otherUV.x = idxA % uDim;

				if(leaveNorthOrEast == false){
					otherUV.x--;
				}

				int prevCutLineB = u;
				if(idxA < idxB){
					prevCutLineB++;
				}
				m_insert[insertIdx].cutLine.y = prevCutLineB;

				checkPatch->uInsertLineDivertInnerPatch(firstStitchPoly, foundFirstPatch, leftSide, idxA, idxB, ratio, invertRatio, m_insert, insertIdx, cutLine, beginInsertIdx, otherUV);
				return true;
			}
			idxA -= 2 * uDim;
			idxB -= 2 * uDim;
			dpA = checkPatch->dpoint[idxA];
			dpB = checkPatch->dpoint[idxB];
			mdpA = g_viewMatrix * dpA;
			mdpB = g_viewMatrix * dpB;
			ptA = Vec2f(mdpA.x, mdpA.y);
			ptB = Vec2f(mdpB.x, mdpB.y);
			if(clickLine(click, ptA, ptB, crossPoint)){
#ifdef PROG_ROB
				g_storeCrossPoint.set(checkPatch->fromPoint + idxA, checkPatch->fromPoint + idxB, .5);
#endif
				//swapEastWestIfNecessary(dpA, dpB);
				leaveNorthOrEast = m_insert[insertIdx].leaveNorthOrEast;

				Vec2i cutLine = Vec2i(m_insert[insertIdx].uOrVIdx, 0);

				if(leaveNorthOrEast){
					if(checkIdx > checkIdx2){
						idxB = idxA;
					}
				}
				else{
					if(checkIdx < checkIdx2){
						idxB = idxA;
					}
					//ratio = 1.0 - ratio;
				}

				idxA++;
				idxA += uDim;
				idxB += uDim;

				if(checkIdx < checkIdx2){
					otherUV.y = checkIdx / uDim;
				}
				else{
					otherUV.y = checkIdx2 / uDim;
					
				}
				otherUV.x = idxA % uDim;

				if(leaveNorthOrEast == false){
					otherUV.x--;
				}

				int prevCutLineB = u;
				if(idxA < idxB){
					prevCutLineB++;
				}
				m_insert[insertIdx].cutLine.y = prevCutLineB;

				checkPatch->uInsertLineDivertInnerPatch(firstStitchPoly, foundFirstPatch, leftSide, idxA, idxB, ratio, invertRatio, m_insert, insertIdx, cutLine, beginInsertIdx, otherUV);
				return true;
			}
		}
	}
	return false;
}

void Seamless::findClickedPolyEdge(){
	g_insertPatch.clickedPolyEdge.length = 0;
	Vec2f click = Vec2f(g_pointer.translation.x, g_pointer.translation.y);
	int poly = 0;

	while(poly < ctrlPolyPoint.length){
		int len = ctrlPolyPoint[poly].len;
		for(int edgeA = 0; edgeA < len; edgeA++){
			int edgeB = edgeA + 1;
			if(edgeB == len){
				edgeB = 0;
			}
			NurbsPatch* patchA = ctrlPolyPoint[poly + edgeA].p;
			NurbsPatch* patchB = ctrlPolyPoint[poly + edgeB].p;
			int idxA = ctrlPolyPoint[poly + edgeA].i;
			int idxB = ctrlPolyPoint[poly + edgeB].i;

			Vec3f dpA = patchA->dpoint[idxA];
			Vec3f dpB = patchB->dpoint[idxB];
			Vec3f mdpA = g_viewMatrix * dpA;
			Vec3f mdpB = g_viewMatrix * dpB;
			Vec2f ptA = Vec2f(mdpA.x, mdpA.y);
			Vec2f ptB = Vec2f(mdpB.x, mdpB.y);
			Vec2f crossPoint;
			if(clickLine(click, ptA, ptB, crossPoint)){
				g_insertPatch.clickedPolyEdge += Vec2i(poly, edgeA);
#ifdef PROG_ROB
				g_storeCrossPoint.set(patchA->fromPoint + idxA, patchB->fromPoint + idxB, .5);
#endif

				//MultiSelect::pointList.length = 0;
				//MultiSelect::pointList += ControlPoint(patchA, idxA);
				//MultiSelect::pointList += ControlPoint(patchB, idxB);
				//houseKeepVisualClue();
				swapEastWestIfNecessary(dpA, dpB);

			}
		}
		poly += len;
		for(int k = 0; k < len + 1; k++){
			int midLen = ctrlPolyPoint[poly].len;
			if(midLen == 0){
				midLen++;
			}
			poly += midLen;
		}
	}
}
/*
NurbsPatch* getSharedEdge(int& sharedMarkIdx, const M_CtrlPolyPoint& ctrlPolyPoint, int beginPoly, int exitPolyEdge, int markPolyEdge, int len){
	if(exitPolyEdge == getCorner(markPolyEdge, 1, len)){
		sharedMarkIdx = ctrlPolyPoint[beginPoly + exitPolyEdge].i;
		return ctrlPolyPoint[beginPoly + exitPolyEdge].p;
	}
	if(exitPolyEdge == getCorner(markPolyEdge, len - 1, len)){
		sharedMarkIdx = ctrlPolyPoint[beginPoly + markPolyEdge].i;
		return ctrlPolyPoint[beginPoly + markPolyEdge].p;
	}
	else throw;
}
*/
int getIndexContainedInMidPoint(const M_CtrlPolyPoint& ctrlPolyPoint, const int absoluteIdx, const NurbsPatch* checkPatch, const int entryIdx2, const bool doU);

bool linePassesThroughClickedStitchPoly2(const M_CtrlPolyPoint& ctrlPolyPoint, const NurbsPatch* checkPatch, const int idx, const int originIdx, int& entryIdx, int& entryIdx2, const bool doU, const bool leaveNorthOrEast, const int uOrV, const int uOrV2, const int uDim, const int midPoly){
	entryIdx = getIndexContainedInMidPoint(ctrlPolyPoint, midPoly, checkPatch, idx, doU);
	if(entryIdx >= 0){
		entryIdx2 = idx;
		if(doU){
			int u = entryIdx % uDim;
			if(u == uOrV){
				u = entryIdx2 % uDim;
				if(u == uOrV2){
					if(leaveNorthOrEast){
						if(entryIdx >= originIdx){
							return true;
						}
					}
					else{
						if(entryIdx <= originIdx){
							return true;
						}
					}
				}
			}
		}
		else{
			int v = entryIdx / uDim;
			if(v == uOrV){
				v = entryIdx2 / uDim;
				if(v == uOrV2){
					if(leaveNorthOrEast){
						if(entryIdx >= originIdx){
							return true;
						}
					}
					else{
						if(entryIdx <= originIdx){
							return true;
						}
					}
				}
			}
		}
	}
	return false;
}

int getCenterBegin(const M_CtrlPolyPoint& ctrlPolyPoint, int poly){
	int len = ctrlPolyPoint[poly].len;
	poly += len;

	for(int k = 0; k < len; k++){
		int midLen = ctrlPolyPoint[poly].len;
		if(midLen == 0){
			midLen++;
		}
		poly += midLen;
	}
	return poly;
}

int getMidPoly(int edge, const M_CtrlPolyPoint& ctrlPolyPoint, int midPoly){
	for(int i = 0; i < edge; i++){
		int midLen = ctrlPolyPoint[midPoly].len;
		if(midLen == 0){
			midLen++;
		}
		midPoly += midLen;
	}
	return midPoly;
}

bool linePassesThroughClickedStitchPoly3(const M_CtrlPolyPoint& ctrlPolyPoint, const NurbsPatch* checkPatch, const int midBegin, const int originIdx, int& entryIdx, int& entryIdx2, int& poly, int& edge, int& otherPolyEdge, int& midPoly, int& crossOver, const bool doU, const bool leaveNorthOrEast, bool& aSide, const int uOrV, const int uOrV2, const int uDim, const int len){
	for(;edge < len; edge++){
		int edgeB = edge + 1;
		if(edgeB == len){
			edgeB = 0;
		}

		if(edge == 4){
			int a = 0;
		}

		NurbsPatch* patchA = ctrlPolyPoint[poly + edge].p;
		NurbsPatch* patchB = ctrlPolyPoint[poly + edgeB].p;
		int idxA = ctrlPolyPoint[poly + edge].i;
		int idxB = ctrlPolyPoint[poly + edgeB].i;

		if(checkPatch == patchA){
			if(linePassesThroughClickedStitchPoly2(ctrlPolyPoint, checkPatch, idxA, originIdx, entryIdx, entryIdx2, doU, leaveNorthOrEast, uOrV, uOrV2, uDim, midPoly)){
				int crossOverEdge = getCorner(edge, len - 1, len);
				crossOver = getMidPoly(crossOverEdge, ctrlPolyPoint, midBegin);
				otherPolyEdge = getCorner(crossOverEdge, len - 1, len);
				aSide = true;
				return true;
			}
		}
		if(checkPatch == patchB){
			if(linePassesThroughClickedStitchPoly2(ctrlPolyPoint, checkPatch, idxB, originIdx, entryIdx, entryIdx2, doU, leaveNorthOrEast, uOrV, uOrV2, uDim, midPoly)){
				crossOver = getMidPoly(edgeB, ctrlPolyPoint, midBegin);
				otherPolyEdge = getCorner(edgeB, 1, len);
				aSide = false;
				return true;
			}
		}

		int midLen = ctrlPolyPoint[midPoly].len;
		if(midLen == 0){
			midLen++;
		}
		midPoly += midLen;
	}
	return false;
}

void getAltPatchEdgePoints(const M_CtrlPolyPoint& ctrlPolyPoint, NurbsPatch*& altPatch, int& altIdx, int& altIdx2, bool& altDoU, bool& altLeaveNorthOrEast, const bool aSide, int poly, int polyIdx, int midBegin, int len, int& altPolyEdge){

	altPolyEdge = g_insertPatch.clickedPolyEdge[polyIdx].y;
	int altMidPoly = getMidPoly(altPolyEdge, ctrlPolyPoint, midBegin);

	if(aSide){
		int nextEdge = getCorner(altPolyEdge, 1, len);
		altPatch = ctrlPolyPoint[poly + nextEdge].p;
		altIdx2 = ctrlPolyPoint[poly + nextEdge].i;
		getMatchingMidPoint(ctrlPolyPoint, altMidPoly, altPatch, altIdx);
	}
	else{
		altPatch = ctrlPolyPoint[poly + altPolyEdge].p;
		altIdx2 = ctrlPolyPoint[poly + altPolyEdge].i;
		getMatchingMidPoint2(ctrlPolyPoint, altMidPoly, altPatch, altIdx, altIdx2);
	}

	int centerPoly = getCenterBegin(ctrlPolyPoint, poly);

	if(abs(altIdx - altIdx2) == 1){
		altDoU = true;
		uGetEdgePoint(ctrlPolyPoint, centerPoly, altPatch, altIdx, altIdx2, altLeaveNorthOrEast);
	}
	else{
		altDoU = false;
		vGetEdgePoint(ctrlPolyPoint, centerPoly, altPatch, altIdx, altIdx2, altLeaveNorthOrEast);
	}
}

void getEntryPatchEdgePoints(int& idx, int& idx2, bool doU, bool leaveNorthOrEast, int uDim){
	if(doU){
		if(leaveNorthOrEast){
			idx += uDim;
			idx2 += uDim;
		}
		else{
			idx -= uDim;
			idx2 -= uDim;
		}
	}
	else{
		if(leaveNorthOrEast){
			idx += 1;
			idx2 += 1;
		}
		else{
			idx -= 1;
			idx2 -= 1;
		}
	}
}

NurbsPatch* anotherPointContainedInMidPoint(const M_CtrlPolyPoint& ctrlPolyPoint, const NurbsPatch* checkPatch, const int checkIdx, const int checkIdx2, int centerPoly, int midPoly, int& otherIdx, int& otherIdx2);


/*
int getEntryPolyEdge(const M_CtrlPolyPoint& ctrlPolyPoint, const NurbsPatch* checkPatch, int poly, int checkIdx, int checkIdx2){
	int len = ctrlPolyPoint[poly].len;
	for(int edge = 0; edge < len; edge++){
		int edgeB = edge + 1;
		if(edgeB == len){
			edgeB = 0;
		}

		NurbsPatch* patchA = ctrlPolyPoint[poly + edge].p;
		NurbsPatch* patchB = ctrlPolyPoint[poly + edgeB].p;
		int idxA = ctrlPolyPoint[poly + edge].i;
		int idxB = ctrlPolyPoint[poly + edgeB].i;

		int absoluteIdx = getMidPointAbsoluteIdx(poly + len, edge, ctrlPolyPoint);

		if(checkPatch == patchA){
			if(isContainedInMidPoint(ctrlPolyPoint, poly, patchA, idxA)){
				return edge;
			}
		}
		if(checkPatch == patchB){
			if(isContainedInMidPoint(ctrlPolyPoint, poly, patchB, idxB)){
				return edge;
			}
		}
	}
	return -1;
}
*/

bool linePassesThroughClickedStitchPoly(const M_CtrlPolyPoint& ctrlPolyPoint, NurbsPatch* checkPatch, const int originIdx, const int checkIdx, const int checkIdx2, int& entryIdx, int& entryIdx2, const bool doU, const bool leaveNorthOrEast, NurbsPatch*& otherPatch, int& otherIdx, int& otherIdx2, bool& otherDoU, bool& otherLeaveNorthOrEast, NurbsPatch*& altPatch, int& altIdx, int& altIdx2, bool& altDoU, bool& altLeaveNorthOrEast, int& poly, int& edge, int& otherPolyEdge, int& altPolyEdge, int& midPoly, bool& aSide, int& crossOver, const int uOrV, const int uOrV2, const int uDim, bool& forbid){
	for(int polyIdx = g_insertPatch.clickedPolyEdge.length; --polyIdx>=0;){
		poly = g_insertPatch.clickedPolyEdge[polyIdx].x;
		const int len = ctrlPolyPoint[poly].len;
		const int midBegin = poly + len;
		midPoly = midBegin;
		//if(originIdx == 14 && checkIdx == 0){
		//	int a = 0;
		//}

		edge = 0;
		if(linePassesThroughClickedStitchPoly3(ctrlPolyPoint, checkPatch, midBegin, originIdx, entryIdx, entryIdx2, poly, edge, otherPolyEdge, midPoly, crossOver, doU, leaveNorthOrEast, aSide, uOrV, uOrV2, uDim, len)){
			int _edge = edge + 1;
			int _crossOver = 0;
			bool _aSide = false;
	//for(int j = 0; j < len; j++){
	//	MultiSelect::pointList += ControlPoint(ctrlPolyPoint[poly + j].p, ctrlPolyPoint[poly + j].i);
	//}
			//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[poly + edge].p, ctrlPolyPoint[poly + edge].i);
			//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[poly + _edge].p, ctrlPolyPoint[poly + _edge].i);
			//houseKeepVisualClue();

			int midLen = ctrlPolyPoint[midPoly].len;
			if(midLen == 0){
				midLen++;
			}
			int _midPoly = midPoly + midLen;

			int dummy = 0;

			if(linePassesThroughClickedStitchPoly3(ctrlPolyPoint, checkPatch, midBegin, originIdx, otherIdx, otherIdx2, poly, _edge, dummy, _midPoly, _crossOver, doU, leaveNorthOrEast, _aSide, uOrV, uOrV2, uDim, len)){
				if((leaveNorthOrEast == true && entryIdx > otherIdx) || (leaveNorthOrEast == false && entryIdx < otherIdx)){
					swap(entryIdx, otherIdx);
					swap(entryIdx2, otherIdx2);
					edge = _edge;
					crossOver = _crossOver;
					aSide = _aSide;
					midPoly = _midPoly;
				}
			}
			else{
				int centerPoly = getCenterBegin(ctrlPolyPoint, poly);
				//if(otherPatch = anotherPatchContainedInMidPoint(ctrlPolyPoint, checkPatch, centerPoly, crossOver, otherIdx, otherIdx2)){
				if(otherPatch = anotherPointContainedInMidPoint(ctrlPolyPoint, checkPatch, checkIdx, checkIdx2, centerPoly, crossOver, otherIdx, otherIdx2)){
				}
			}

//				MultiSelect::pointList += ControlPoint(otherPatch, entryIdx);
//				MultiSelect::pointList += ControlPoint(otherPatch, entryIdx2);
//				houseKeepVisualClue();
			//if(otherPatch == 0){
			//	return false;
			//}
			getEntryPatchEdgePoints(entryIdx, entryIdx2, doU, leaveNorthOrEast, uDim);
			if(otherPatch == 0){
				otherPatch = checkPatch;
				otherIdx = entryIdx;
				otherIdx2 = entryIdx2;
					
			}


			//getAltPatchEdgePoints(ctrlPolyPoint, otherPatch, otherIdx, otherIdx2, otherDoU, otherLeaveNorthOrEast, aSide, poly, polyIdx, midBegin, len);


			getAltPatchEdgePoints(ctrlPolyPoint, altPatch, altIdx, altIdx2, altDoU, altLeaveNorthOrEast, aSide, poly, polyIdx, midBegin, len, altPolyEdge);

			if(edge == altPolyEdge){
				return false;
			}

			//if(len > 4 && (abs(edge - altPolyEdge) == 1 || (edge == 0 && altPolyEdge == len - 1) || (altPolyEdge == 0 && edge == len - 1))){
			//	forbid = true;
			//	return false; //prevent creation of a control triangle which also results in a poly with more sides than the original
			//}
			if(otherPatch == altPatch && otherIdx == altIdx && otherIdx2 == altIdx2){
				forbid = true;
				return false;
			}

			return true;
		}
	}
	return false;
}


bool patchCrossesOverLeftSide(NurbsPatch* patch, int idx, int checkIdx, int uDim, int vDim, bool leaveNorthOrEast, int insertIdx){
	M_Vec3f& dpoint = patch->dpoint.refValue();
	int firstIdx = idx;
	if(leaveNorthOrEast){
		while(idx <= checkIdx){
			float x = patch->dpoint[idx].x;
			if(x > CROSS_OVER_AMOUNT){
				//if(false == (insertIdx == 0 && idx == firstIdx && (idx % uDim) == (uDim / 2))){
					return true;
				//}
			}
			idx++;
		}
	}
	else{
		while(idx >= checkIdx){
			float x = patch->dpoint[idx].x;
			if(x > CROSS_OVER_AMOUNT){
				//if(false == (insertIdx == 0 && idx == firstIdx && (idx % uDim) == (uDim / 2 + 1))){
					return true;
				//}
			}
			idx--;
		}
	}
	return false;
}

bool patchCrossesOverRightSide(NurbsPatch* patch, int idx, int checkIdx, int uDim, int vDim, bool leaveNorthOrEast, int insertIdx){
	M_Vec3f& dpoint = patch->dpoint.refValue();
	int firstIdx = idx;
	if(leaveNorthOrEast){
		while(idx <= checkIdx){
			float x = patch->dpoint[idx].x;
			if(x < -CROSS_OVER_AMOUNT){
				//if(false == (insertIdx == 0 && idx == firstIdx && (idx % uDim) == (uDim / 2))){
					return true;
				//}
			}
			idx++;
		}
	}
	else{
		while(idx >= checkIdx){
			float x = patch->dpoint[idx].x;
			if(x < -CROSS_OVER_AMOUNT){
				//if(false == (insertIdx == 0 && idx == firstIdx && (idx % uDim) == (uDim / 2 + 1))){
					return true;
				//}
			}
			idx--;
		}
	}
	return false;
}

bool endsCenterOfPatch(const Insert& insert){
	NurbsPatch* patch = insert.p;
	int uDim = patch->dim.x;
	if(patch->mirror && patch->mirrorPatch == 0 && insert.doU == false && insert.cutLine.y == (uDim / 2)){
		return true;
	}
	return false;
}



bool Seamless::insertLineFindClickedPoly(NurbsPatch* checkPatch, const int originIdx, int checkIdx, int checkIdx2, bool doU, bool leaveNorthOrEast, const int firstStitchPoly, bool& foundFirstPatch, const bool leftSide, const float ratio, const bool invertRatio, M_Insert& m_insert, int insertIdx, int& beginInsertIdx, bool& forbid){
	//if(checkPatch->name == "strip" && markClicked){
	//	leaveNorthOrEast = true;
	//	m_insert[0].leaveNorthOrEast = true;
	//}

	int uCheckPatchDim = checkPatch->dim.x;
	int vCheckPatchDim = checkPatch->dim.y;


	int uOrV = 0;
	int uOrV2 = 0;
	bool go = false;
	if(doU){
		uOrV = checkIdx % uCheckPatchDim;
		uOrV2 = checkIdx2 % uCheckPatchDim;
		go = uOrV > 0 && uOrV < uCheckPatchDim - 1;
	}
	else{
		uOrV = checkIdx / uCheckPatchDim;
		uOrV2 = checkIdx2 / uCheckPatchDim;
		go = uOrV > 0 && uOrV < vCheckPatchDim - 1;
	}
	if(go){
		if(checkMarkRightangle(checkPatch, checkIdx, checkIdx2, uOrV, doU, leaveNorthOrEast, firstStitchPoly, foundFirstPatch, leftSide, ratio, invertRatio, m_insert, insertIdx, beginInsertIdx)){
			return true;
		}
	}

	int entryIdx = 0;
	int entryIdx2 = 0;
	int poly = 0;
	int edge = 0;
	int altPolyEdge = 0;
	int midPoly = 0;
	int crossOver = 0;
	NurbsPatch* altPatch = 0;
	int altIdx = 0;
	int altIdx2 = 0;
	bool altDoU = false;
	bool altLeaveNorthOrEast = false;
	NurbsPatch* otherPatch = 0;
	int otherIdx = 0;
	int otherIdx2 = 0;
	bool otherDoU = false;
	bool otherLeaveNorthOrEast = false;
	bool aSide = false;


	Vec2i cutLine;
	
	int	otherPolyEdge = 0;

	if(g_insertPatch.clickedPolyEdge.length && linePassesThroughClickedStitchPoly(ctrlPolyPoint, checkPatch, originIdx, checkIdx, checkIdx2, entryIdx, entryIdx2, doU, leaveNorthOrEast, otherPatch, otherIdx, otherIdx2, otherDoU, otherLeaveNorthOrEast, altPatch, altIdx, altIdx2, altDoU, altLeaveNorthOrEast, poly, edge, otherPolyEdge, altPolyEdge, midPoly, aSide, crossOver, uOrV, uOrV2, uCheckPatchDim, forbid)){
		g_insertPatch.clickedPolyEdge.length = 0;

		beginInsertIdx = insertIdx + 1;

		int len = ctrlPolyPoint[poly].len;

		//int altPolyEdge2 = altPolyEdge + 1;
		//if(altPolyEdge2 == len){
		//	altPolyEdge2 = 0;
		//}

		//MultiSelect::pointList.length = 0;
		//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[poly + altPolyEdge].p, ctrlPolyPoint[poly + altPolyEdge].i);
		//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[poly + altPolyEdge2].p, ctrlPolyPoint[poly + altPolyEdge2].i);
		//houseKeepVisualClue();


		//int otherPolyEdge2 = otherPolyEdge + 1;
		//if(otherPolyEdge2 == len){
		//	otherPolyEdge2 = 0;
		//}

		//MultiSelect::pointList.length = 0;
		//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[poly + otherPolyEdge].p, ctrlPolyPoint[poly + otherPolyEdge].i);
		//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[poly + otherPolyEdge2].p, ctrlPolyPoint[poly + otherPolyEdge2].i);
		//houseKeepVisualClue();


		////
		//int edge2 = edge + 1;
		//if(edge2 == len){
		//	edge2 = 0;
		//}
		//MultiSelect::pointList.length = 0;
		//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[poly + edge].p, ctrlPolyPoint[poly + edge].i);
		//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[poly + edge2].p, ctrlPolyPoint[poly + edge2].i);
		//houseKeepVisualClue();

		//MultiSelect::pointList.length = 0;
		//MultiSelect::pointList += ControlPoint(checkPatch, entryIdx);
		//MultiSelect::pointList += ControlPoint(checkPatch, entryIdx2);
		//houseKeepVisualClue();

		//MultiSelect::pointList.length = 0;
		//MultiSelect::pointList += ControlPoint(otherPatch, otherIdx);
		//MultiSelect::pointList += ControlPoint(otherPatch, otherIdx2);
		//houseKeepVisualClue();

		//MultiSelect::pointList.length = 0;
		//MultiSelect::pointList += ControlPoint(altPatch, altIdx);
		//MultiSelect::pointList += ControlPoint(altPatch, altIdx2);
		//houseKeepVisualClue();
		//return;
		int prevUorVCenterIdx = 0;
		int other_uDim = otherPatch->dim.x;
		int other_vDim = otherPatch->dim.y;

		Vec2i otherUV = Vec2i(otherIdx % other_uDim, otherIdx / other_uDim);

		int cutY = 0;
		if(doU){
			cutY = entryIdx / uCheckPatchDim;;

			if(leaveNorthOrEast == false){
				otherUV.y--;
			}
		}
		else{
			cutY = entryIdx % uCheckPatchDim;

			if(leaveNorthOrEast == false){
				otherUV.x--;
			}
			if(otherIdx2 < otherIdx){
				otherUV.y--;
			}
		}
		if(leaveNorthOrEast){
			cutY--;
		}
		m_insert[insertIdx].cutLine.y = cutY;

		if(otherUV.x == other_uDim - 1){
			otherUV.x--;
		}
		if(otherUV.y == other_vDim - 1){
			otherUV.y--;
		}


		int altUDim = altPatch->dim.x;
		if(altDoU){
			if(altLeaveNorthOrEast){
				cutLine = Vec2i(altIdx / altUDim, altPatch->dim.y - 2);
			}
			else{
				cutLine = Vec2i(altIdx / altUDim - 1, 0);
			}
			if(cutLine.x < 0 || cutLine.y < 0){
				return false;
			}

			altPatch->uInsertLineDivert(firstStitchPoly, foundFirstPatch, leftSide, altIdx, altIdx2, ratio, invertRatio, m_insert, insertIdx, cutLine, poly, edge, otherPolyEdge, altPolyEdge, aSide, altIdx, otherPatch, otherUV, checkPatch, beginInsertIdx);
		}
		else{
			if(altLeaveNorthOrEast){
				cutLine = Vec2i(altIdx % altUDim, altPatch->dim.x - 2);
			}
			else{
				cutLine = Vec2i(altIdx % altUDim - 1, 0);
			}
			if(cutLine.x < 0 || cutLine.y < 0){
				return false;
			}
			
			altPatch->vInsertLineDivert(firstStitchPoly, foundFirstPatch, leftSide, altIdx, altIdx2, ratio, invertRatio, m_insert, insertIdx, cutLine, poly, edge, otherPolyEdge, altPolyEdge, aSide, altIdx, otherPatch, otherUV, checkPatch, beginInsertIdx);
		}


		return true;
	}
	return false;
}

void Seamless::insertLineFindMatchingEdge(NurbsPatch* checkPatch, int checkIdx, int checkIdx2, bool doU, bool leaveNorthOrEast, const int firstStitchPoly, bool& foundFirstPatch, const bool leftSide, const float ratio, const bool invertRatio, M_Insert& m_insert){


	int poly = 0;
	while(poly < ctrlPolyPoint.length){
		const int beginPoly = poly;
		const int len = ctrlPolyPoint[poly].len;
		const int centerBegin = getCenterBegin(ctrlPolyPoint, poly);
		poly += len;
		const int midBegin = poly;
		int centerLen = ctrlPolyPoint[centerBegin].len;
		if(centerLen == 0){
		
			for(int k = len; --k >= 0;){
				int midLen = ctrlPolyPoint[poly].len;
				if(midLen){
					if(isContainedInMidPoint(ctrlPolyPoint, poly, checkPatch, checkIdx)){
						const int holdPoly = poly;
						poly = midBegin;
						for(int k = len; --k >= 0;){
							if(isContainedInMidPoint(ctrlPolyPoint, poly, checkPatch, checkIdx2)){
								int otherMidIdx = 0;
								int otherCenterIdx = 0;
								//if(NurbsPatch* otherPatch = anotherPatchContainedInMidPoint(ctrlPolyPoint, poly, midBegin, checkPatch, otherMidIdx, otherCenterIdx)){
								if(NurbsPatch* otherPatch = anotherPatchContainedInMidPoint(ctrlPolyPoint, poly, holdPoly, checkPatch, otherMidIdx, otherCenterIdx)){
									if(beginPoly == firstStitchPoly){
										foundFirstPatch = true;
										return;
									}

									int uDim = otherPatch->dim.x;
									if(abs(otherMidIdx - otherCenterIdx) == 1){
										otherPatch->uInsertLine(firstStitchPoly, foundFirstPatch, leftSide, otherCenterIdx, otherMidIdx, ratio, invertRatio, m_insert);
									}
									else{
										otherPatch->vInsertLine(firstStitchPoly, foundFirstPatch, leftSide, otherCenterIdx, otherMidIdx, ratio, invertRatio, m_insert);
									}
									return;
								}
								else{
									//if(isOnPatchEdge(checkIdx, checkIdx2, uCheckPatchDim, vCheckPatchDim, doU, leaveNorthOrEast)){
									//	insertLineFindMatchingEdge(checkPatch, originIdx, checkIdx, checkIdx2, doU, leaveNorthOrEast, firstStitchPoly, foundFirstPatch, leftSide, ratio, invertRatio, m_insert, insertIdx, markClicked);
									//	return;
									//}

									foundFirstPatch = true;
									return;
								}
							}
							int midLen = ctrlPolyPoint[poly].len;
							if(midLen == 0){
								midLen++;
							}
							poly += midLen;
						}
						poly = holdPoly;
						//return;
					}
				}
				else{
					midLen++;
				}
				poly += midLen;
			}
		}
		poly = midBegin;
		if(isContainedInMidPoint(ctrlPolyPoint, centerBegin, checkPatch, checkIdx)){
			poly = midBegin;
			for(int k = len; --k >= 0;){
				if(isContainedInMidPoint(ctrlPolyPoint, poly, checkPatch, checkIdx2)){
					int otherIdx = 0;
					int otherIdx2 = 0;
					
					NurbsPatch* otherPatch = anotherPatchContainedInMidPoint(ctrlPolyPoint, checkPatch, centerBegin, poly, otherIdx, otherIdx2);
					
					if(otherPatch == 0){
						otherPatch = anotherPointContainedInMidPoint(ctrlPolyPoint, checkPatch, checkIdx, checkIdx2, centerBegin, poly, otherIdx, otherIdx2);
					}
					if(otherPatch){
						if(beginPoly == firstStitchPoly){
							foundFirstPatch = true;
							return;
						}
						if(abs(otherIdx - otherIdx2) == 1){
							//int g = 0;
							otherPatch->uInsertLine(firstStitchPoly, foundFirstPatch, leftSide, otherIdx, otherIdx2, ratio, invertRatio, m_insert);
						}
						else{
							otherPatch->vInsertLine(firstStitchPoly, foundFirstPatch, leftSide, otherIdx, otherIdx2, ratio, invertRatio, m_insert);
						}
						return;
					}
				}
				int midLen = ctrlPolyPoint[poly].len;
				if(midLen == 0){
					midLen++;
				}
				poly += midLen;
			}
			return;
		}

		if(centerLen == 0){
			centerLen++;
		}
		poly = centerBegin + centerLen;
	}
}


bool InsertPatch::monitorDirection(){
	if(m_insert.length == 0){
		return false;
	}
	int uDim = init_patch->dim.x;
	int vDim = init_patch->dim.y;
	int u = init_idx % uDim;
	int v = init_idx / uDim;

	if(init_doU){
		if(v == 0 || v == vDim - 1){
			return true;
		}
	}
	else{
		if(u == 0 || u == uDim - 1){
			return true;
		}
	}
	if(m_insert[0].p == init_patch){
		if(init_doU){
			if((init_leaveNorthOrEast && v == 1) || (init_leaveNorthOrEast == false && v == vDim - 2)){
				return true;
			}
		}
		else{
			if((init_leaveNorthOrEast && u == 1) || (init_leaveNorthOrEast == false && u == uDim - 2)){
				return true;
			}
		}
	}
	if(beginsFromCenterLineFunc(0, m_insert[0])){
		return true;
	}

	Vec3f basePoint = init_patch->dpoint[init_idx];

	Vec3f basePointB;
	Vec3f directionPoint;
	Vec3f directionPointB;
			
	if(init_doU){
		directionPoint = init_patch->dpoint[init_idx + uDim];
		directionPointB = init_patch->dpoint[init_idx + uDim + 1];
	}
	else{
		directionPoint = init_patch->dpoint[init_idx + 1];
		directionPointB = init_patch->dpoint[init_idx + 1 + uDim];
	}



	if(init_doU){
		basePointB = init_patch->dpoint[init_idx + 1];
	}
	else{
		basePointB = init_patch->dpoint[init_idx + uDim];
	}
			
	if((init_doU == true && u & 1) || (init_doU == false && v & 1)){
		basePoint = basePoint * init_ratio + basePointB * (1.0 - init_ratio);
		directionPoint = directionPoint * init_ratio + directionPointB * (1.0 - init_ratio);
	}
	else{
		basePoint = basePoint * (1.0 - init_ratio) + basePointB * init_ratio;
		directionPoint = directionPoint * (1.0 - init_ratio) + directionPointB * init_ratio;
	}

//////

	g_viewMatrix = g_scene->getViewMatrix();

	basePoint = g_viewMatrix * basePoint;
	directionPoint = g_viewMatrix * directionPoint;

	Vec2f diff = Vec2f(directionPoint.x - basePoint.x, directionPoint.y - basePoint.y);

	Vec2f mousePoint = Vec2f(g_pointer.translation.x - basePoint.x, g_pointer.translation.y - basePoint.y);

	if(abs(diff.x) > .0001 || abs(diff.y) > .0001){
		float angle = diff.getAngle2();
		diff = diff.rotate(-angle);
		mousePoint = mousePoint.rotate(-angle);

		if(mousePoint.x > 0){
			if(init_leaveNorthOrEast != true){
				verifyFirst();
				reverseDirection(g_insertPatch.m_insert);
				changeInitElement();
				init_leaveNorthOrEast = true;
				init_patch->houseKeep();
			}
		}
		else{
			if(init_leaveNorthOrEast != false){
				reverseDirection(g_insertPatch.m_insert);
				changeInitElement();
				init_leaveNorthOrEast = false;
				init_patch->houseKeep();
			}
		}
		
//		const float inverseSize = (1 / g_scene->getZoom());
//		float senseLen = .1 * inverseSize;
		float senseLen = g_mouseSense * 5;

		if(mousePoint.length() > senseLen){
			return true;
		}
	}
	else{
		return true;
	}
	return false;
}
