#include <windows.h>
#include <algorithm>    // std::max
using namespace std;

#include "seamless3d.h"
#include "gui.h"
#include "techuelife_triangle_editor.h"

void createPNG24(const char* filename, int width, int height, Pix24* pixels);
bool backFace(Vec2f* v);
float getDistanceFromVertex(const Vec2f& p0, const Vec2f& p1);
void setUpInternalEffectPtrs2(Seamless* smls); //this (2) wont be needed when setUpInternalEffectPtrs is fixed
void setUpInternalEffectPtrs(Seamless* smls);
void resetLooseEnds();
void initBnspePolies(Build* build, Art* part);

extern SelNode g_selNode;
M_Float g_shade;


RayView g_rayView;
RayView g_rayLight;

float g_contrast = 0.5;

#define NORMAL_SHADOW 1

void HitElement::getColorFromPointInTriangle(const int fromCoord, const S_Triangle& triangle, const Vec3f hitPoint, const ColorEffect* colorEffect, M_Float& shade, Matrix& matrix){

	for(int i0A = 0; i0A < 3; i0A++){
		int i0B = i0A + 1;
		if(i0B == 3){
			i0B = 0;
		}
		int ic0A = triangle.refIndex(i0A);
		int it0A = triangle.refTexIndex(i0A);
		Vec3f v0A = matrix * triangle.refPart(i0A)->coord[ic0A];

		int ic0B = triangle.refIndex(i0B);
		int it0B = triangle.refTexIndex(i0B);
		Vec3f v0B = matrix * triangle.refPart(i0B)->coord[ic0B];

		
		if(hitPoint.y >= v0A.y && hitPoint.y <= v0B.y && v0A.y != v0B.y){
			int i1A = i0A;
			for(int k = 0; k < 2; k++){
				i1A++;
				if(i1A == 3){
					i1A = 0;
				}
				int i1B = i1A + 1;
				if(i1B == 3){
					i1B = 0;
				}
				int ic1A = triangle.refIndex(i1A);
				int it1A = triangle.refTexIndex(i1A);
				Vec3f v1A = matrix * triangle.refPart(i1A)->coord[ic1A];
				int ic1B = triangle.refIndex(i1B);
				int it1B = triangle.refTexIndex(i1B);
				Vec3f v1B = matrix * triangle.refPart(i1B)->coord[ic1B];

				if(hitPoint.y <= v1A.y && hitPoint.y >= v1B.y && v1A.y != v1B.y){

					float diff0 = v0B.y - v0A.y;
					float r0 = (hitPoint.y - v0A.y) / diff0;
					float diff1 = v1A.y - v1B.y;
					float r1 = (hitPoint.y - v1B.y) / diff1;

					Vec3f v2A = v0A * (1.0 - r0) + v0B * r0;
					Vec3f v2B = v1B * (1.0 - r1) + v1A * r1;

					{
						Color3f c2A = colorEffect->color[it0A] * (1.0 - r0) + colorEffect->color[it0B] * r0;
						Color3f c2B = colorEffect->color[it1B] * (1.0 - r1) + colorEffect->color[it1A] * r1;
						float diff2 = v2B.x - v2A.x;
						float r2 = (hitPoint.x - v2A.x) / diff2;
						this->rawColor = c2A * (1.0 - r2) + c2B * r2;
					}


					Color3f c2A = colorEffect->color[it0A] * shade[fromCoord + ic0A] * (1.0 - r0) + colorEffect->color[it0B] * shade[fromCoord + ic0B] * r0;
					Color3f c2B = colorEffect->color[it1B] * shade[fromCoord + ic1B] * (1.0 - r1) + colorEffect->color[it1A] * shade[fromCoord + ic1A] * r1;
					float diff2 = v2B.x - v2A.x;
					float r2 = (hitPoint.x - v2A.x) / diff2;
					this->shadedColor = c2A * (1.0 - r2) + c2B * r2;;
				}
			}
		}
	}
}


bool Art::checkPointInTriangleRayTrace(RayView& rayView){
	int triIdx = rayView.testTriIdx;
	Vec3f pointer = rayView.pointer;

	Art* part[3] = {
		triangle.refPart(triIdx),
		triangle.refPart(triIdx+1),
		triangle.refPart(triIdx+2)
	};

	int crdIdx[3] = {
		triangle.refIndex(triIdx),
		triangle.refIndex(triIdx+1),
		triangle.refIndex(triIdx+2)
	};

	Vec3f sr[3] = {
		rayView.matrix * part[0]->coord.refValue()[crdIdx[0]],
		rayView.matrix * part[1]->coord.refValue()[crdIdx[1]],
		rayView.matrix * part[2]->coord.refValue()[crdIdx[2]]
	};

	Vec2f v[3];
	Vec2f mv;

	if(rayView.perspective){
		v[0].x = sr[0].x / (-sr[0].z + g_viewDist);
		v[0].y = sr[0].y / (-sr[0].z + g_viewDist);

		v[1].x = sr[1].x / (-sr[1].z + g_viewDist);
		v[1].y = sr[1].y / (-sr[1].z + g_viewDist);

		v[2].x = sr[2].x / (-sr[2].z + g_viewDist);
		v[2].y = sr[2].y / (-sr[2].z + g_viewDist);

		mv.x = pointer.x / (-pointer.z + g_viewDist);
		mv.y = pointer.y / (-pointer.z + g_viewDist);
	}
	else{
		v[0].x = sr[0].x;
		v[0].y = sr[0].y;

		v[1].x = sr[1].x;
		v[1].y = sr[1].y;

		v[2].x = sr[2].x;
		v[2].y = sr[2].y;

		mv.x = pointer.x;
		mv.y = pointer.y;
	}
	
	if(isTriangleHit(Vec2f(mv.x, mv.y), v)){
		if(g_settings->backfaceIgnore){
			if(backFace(v) == false)	return false; //if success exit loop 
		}

		Vec3f v3[3]={
			Vec3f(v[0].x, v[0].y, sr[0].z),
			Vec3f(v[1].x, v[1].y, sr[1].z),
			Vec3f(v[2].x, v[2].y, sr[2].z)
		};
		Vec3f clickPos;

		getTriangleHitPos(clickPos, Vec2f(pointer.x, pointer.y), v3);

		int hitTri = triIdx / 3;

		if(&g_rayView == &rayView){
			if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(workEffect)){
				HitElement hit;
				hit.hitZ = clickPos.z;
				hit.getColorFromPointInTriangle(fromCoord, triangle[hitTri], pointer, colorEffect, g_shade, g_rayView.matrix);
				//float transparency = colorEffect->transparency;
				if(this->plidx == 0){
					hit.transparency = .5;
				}
				else{
					hit.transparency = 0;
				}
				rayView.m_hit += hit;
			}
		}


		if(rayView.part == 0 || clickPos.z > rayView.hitZ){

			rayView.hitZ = clickPos.z;
			rayView.clickPos = clickPos;
			rayView.part = this;
			rayView.hitTri = hitTri;

			return true;
		}
	}
	return false;
}

bool Seamless::senseAllPartsTriangleRayTrace(RayView& rayView){
	const bool perspective = g_perspective;
	bool hit = false;

	for(int k = 0; k < partList.length; k++){

		Art* part = partList[k];
		M_Triangle& triangle = part->triangle;

		if(part->show){
			for(int i = triangle.getTriangleLength(); (i -= 3) >= 0; ){
				rayView.testTriIdx = i;
				if(part->checkPointInTriangleRayTrace(rayView)){
					hit = true;
				}
			}
		}
	}
	return hit;
}


void bubbleSortHitZ(M_<HitElement>& m_hit){
	bool flag;
	do{
		flag = false;
		for(int i = 0; i < m_hit.length - 1; i++){

			if (m_hit[i].hitZ > m_hit[i + 1].hitZ){
				HitElement swapHit = m_hit[i + 1];
				m_hit[i + 1] = m_hit[i];
				m_hit[i] = swapHit;

				flag = true;
			}
		}
	}while(flag);
}

Color3f mixColors(Color3f& canvas, Color3f& topColor, float transparency){
	float opacity = 1.0 - transparency;
	canvas = canvas * transparency;
	topColor = topColor * opacity;

	return canvas + topColor;
}

bool RayView::rayTraceView(Seamless* smls, const Vec3f pointer){
	m_hit.length = 0;

	this->pointer = pointer;
	this->part = 0;
	
	if(smls->senseAllPartsTriangleRayTrace(*this)){
		bubbleSortHitZ(m_hit);
		return true;

		//if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(part->workEffect)){
		//		//color = getColorFromPointInTriangle(part->triangle[hitTri], pointer, colorEffect, g_shade, g_rayView.matrix);
		//		return true;
		//	}
		//}
		//else if(TextureEffect* textureEffect = dynamic_cast<TextureEffect*>(part->workEffect)){
		//	unsigned int pixel = textureEffect->readPixelFromTexture();
		//	color = hexToColor(pixel);
		//	return true;
		//}
	}
	return false;
}

bool RayView::rayTraceLightNormalShading(Seamless* smls, const Vec3f pointer, Art* blockPart, int blockCoord){
	this->pointer = pointer;
	if(smls->senseAllPartsTriangleRayTrace(*this)){

		S_Triangle& tri = part->triangle[hitTri];

		if(	(tri.pa == blockPart && tri.ia == blockCoord) ||
			(tri.pb == blockPart && tri.ib == blockCoord) ||
			(tri.pc == blockPart && tri.ic == blockCoord)
			){
			return false;
		}

		return true;
	}
	return false;
}
bool RayView::rayTraceLight(Seamless* smls, const Vec3f pointer, Art* blockPart, int blockTri){
	this->pointer = pointer;
	if(smls->senseAllPartsTriangleRayTrace(*this)){

		if(part == blockPart){
			if(hitTri == blockTri){
				return false;
			}
		}

		return true;
	}
	return false;
}

void Seamless::convertNormalsToShade(M_Float& shade){
	Vec3f ln = -g_rayLight.rotation * Vec3f(0, 0, 1);
	ln = ln.normalize();
	float iContrast = 1.0 - g_contrast;
	int linear = 0;
	int fromCoord = 0;
	for(int k = 0; k < partList.length; k++){

		Art* part = partList[k];
		part->fromCoord = fromCoord;
		M_Vec3f& coord = part->coord;
		M_Vec3f& normal = part->normal;
		shade.length += normal.length;
		fromCoord += normal.length;

		for(int i = 0; i < normal.length; i++){

			Vec3f crd = coord[i];
			Vec3f pointer = g_rayLight.matrix * coord[i];
			g_rayLight.hitZ = pointer.z;

			{
				Vec3f n = normal[i];
				Rotation r = Rotation(ln, n);
				float s = 1.0 - r.angle / (PI * .5);
				if(s < 0.0){
					s = 0;
				}
				s = s * iContrast + g_contrast;

				if(s > g_contrast){
					if(g_rayLight.rayTraceLightNormalShading(this, pointer, part, i)){
						s = s * .92;
						if(s < g_contrast){
							s = g_contrast;
						}
					}
				}


				shade[linear++] = s;
			}
		}
	}
}

void renderRayTrace(){
	return;

	g_viewMatrix = g_scene->getViewMatrix();
	g_rayView.matrix = g_viewMatrix;

	//g_rayLight.translation = Vec3f(0.43006712, -0.41693449, 0.26251721);
	//g_rayLight.rotation = Rotation(0.95463735, -0.28077132, 0.099172562, 1.0798147);

	g_rayLight.translation = Vec3f(0.071130812, -0.41061890, -0.17637753);
	g_rayLight.rotation = Rotation(0.88067788, -0.41901696, -0.22097798, 0.78860098);

	//g_rayLight.rotation = Rotation(1, 0, 0, 0.78539801);
	//g_rayLight.translation = Vec3f(0, -0.87867999, -0.87867928);
	
	g_rayLight.matrix = translationMatrix(g_rayLight.translation) * rotationMatrix(g_rayLight.rotation) * scaleMatrix(g_rayLight.scale);
	//lightTran = g_scene->translation;
	//lightRot = g_scene->rotation;

	M_<Pix24> pixels;
	int screenScale = 8;
	float pixSize = 0.025 / screenScale;
	int width = 160 * screenScale;
	int height = 90 * screenScale;
	pixels.length = width * height;
	
	if(Art* throwAway = dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls = throwAway->smls;
		smls->rasterTrianglesToVoxels();
		return;

		//setUpInternalEffectPtrs2(smls);

		g_shade.length = 0;
		smls->convertNormalsToShade(g_shade);

		for(int yi = 0; yi < height; yi++){
			for(int xi = 0; xi < width; xi++){
				int i = yi * width + xi;
				Color3f color = Color3f(0,0,0);
				float x = -0.4 + xi * pixSize - 1;
				float y = 0.2 - (yi * pixSize - 1);
				Vec3f pointer = Vec3f(x, y, 0);
				if(g_rayView.rayTraceView(smls, pointer)){
					M_<HitElement>& m_hit = g_rayView.m_hit;

					//pointer = g_scene->getIViewMatrix() * g_rayView.clickPos;
					//pointer = g_rayLight.matrix * pointer;
					//g_rayLight.hitZ = pointer.z;

					//if(g_rayView.part->plidx == 1 && g_rayLight.rayTraceLight(smls, pointer, g_rayView.part, g_rayView.hitTri)){
					//	for(int i = 0; i < m_hit.length; i++){
					//		color = mixColors(color, m_hit[i].rawColor, m_hit[i].transparency);
					//	}
					//	color = color * g_contrast;
					//}
					//else
					{
						for(int i = 0; i < m_hit.length; i++){
							color = mixColors(color, m_hit[i].shadedColor, m_hit[i].transparency);
						}
					}
					pixels[i] = Pix24(color.r * 255, color.g * 255, color.b * 255);
				}
				else{
					pixels[i] = Pix24(0,0,0);
				}
			}
		}
		//setUpInternalEffectPtrs(smls);
		createPNG24("ray-trace.png", width, height, pixels.getArray());
	}
}

inline int getLowPoint(const Vec3i* tri, const float pixelSize){
	float most = tri[0].y;
	int a = 0;
	for(int i = 1; i < 3; i++){
		if(tri[i].y < most){
			most = tri[i].y;
			a = i;
		}
	}
	return a;
}

inline Vec3i convertTo3i(Vec3f& vec){
	float x = vec.x;
	if(x < 0){
		x -= .5;
	}
	else{
		x += .5;
	}

	float y = vec.y;
	if(y < 0){
		y -= .5;
	}
	else{
		y += .5;
	}

	float z = vec.z;
	if(z < 0){
		z -= .5;
	}
	else{
		z += .5;
	}
	return Vec3i(x, y, z);
}

inline Vec3f convertTo3f(Vec3i& vec){
	return Vec3f(vec.x, vec.y, vec.z);
}

void convertToFastTriangle(S_Triangle& triangle, Vec3i* tri, Color3f* col, const M_Color3f& m_color, const float pixelSize){
	Art* partA = triangle.refPart(0);
	Art* partB = triangle.refPart(1);
	Art* partC = triangle.refPart(2);

	tri[0] =  convertTo3i(partA->coord[triangle.refIndex(0)] / pixelSize);
	tri[1] =  convertTo3i(partB->coord[triangle.refIndex(1)] / pixelSize);
	tri[2] =  convertTo3i(partC->coord[triangle.refIndex(2)] / pixelSize);

	col[0] = m_color[triangle.refTexIndex(0)];
	col[1] = m_color[triangle.refTexIndex(1)];
	col[2] = m_color[triangle.refTexIndex(2)];
}


void addVoxel(M_Build& m_build, Art* part, Vec3f& position, const float pixelSize){
	BoxBuild* cube = new BoxBuild;
	cube->part.refValue() = part;
	cube->scale.refValue() = Vec3f(pixelSize * .5, pixelSize * .5, pixelSize * .5);
	cube->yCentred.refValue() = true;
	cube->translation.refValue() = position;

	cube->processFields();
	m_build += cube;
	
}

void calcEdgePositions(float xDiff, int yDiff, float zDiff, Vec3i& lowPoint, M_<Vec3i>& m_edge){
	Vec3f position;
	for(int y = 0; y < yDiff; y++){
		float fy = static_cast<float>(y);
		position.x = xDiff / yDiff * fy;
		position.y = fy;
		position.z = zDiff / yDiff * fy;
		Vec3i position3i = convertTo3i(position);
		position3i = position3i + lowPoint;

		m_edge += position3i;
	}
}

void plotEdgeVoxels(M_<Vec3i>& m_edge, const float pixelSize, Art* part, M_Build& m_build){
	
	for(int i = 0; i < m_edge.length; i++){
		Vec3f position = convertTo3f(m_edge[i]) * pixelSize;

		addVoxel(m_build, part, position, pixelSize);
	}
}




void Seamless::rasterTrianglesToVoxels(){
	Vec3i v;
	Vec3i tri[3];
	Color3f col[3];
	const float pixelSize = .05;
	M_<Vec3i> m_westEdge;
	M_<Vec3i> m_eastEdge;


	for(int k = 0; k < partList.length; k++){

		Art* part = partList[k];
		M_Triangle& m_triangle = part->triangle;
		const M_Color3f& m_color = dynamic_cast<ColorEffect*>(part->workEffect)->color;

		//for(int t = 0; t < m_triangle.length; t++){
		for(int t = 1; t < 2; t++){
			convertToFastTriangle(m_triangle[t], tri, col, m_color, pixelSize);

			int lowIdx = getLowPoint(tri, pixelSize);
				
			int topIdx0 = lowIdx + 1;
			if(topIdx0 >= 3){
				topIdx0 = 0;
			}
			int topIdx1 = topIdx0 + 1;
			if(topIdx1 >= 3){
				topIdx1 = 0;
			}
			if(tri[topIdx1].y > tri[topIdx0].y){
				swap(topIdx0, topIdx1);
			}



			//Vec3f position = convertTo3f(tri[lowIdx]) * pixelSize;
			//addVoxel(build.refValue(), part, position, pixelSize);


			Vec3f position = convertTo3f(tri[topIdx0]) * pixelSize;
			addVoxel(build.refValue(), part, position, pixelSize);


			int yDiff0 = tri[topIdx0].y - tri[lowIdx].y;
			float xDiff0 = tri[topIdx0].x - tri[lowIdx].x;
			float zDiff0 = tri[topIdx0].z - tri[lowIdx].z;

			int yDiff1 = tri[topIdx1].y - tri[lowIdx].y;
			float xDiff1 = tri[topIdx1].x - tri[lowIdx].x;
			float zDiff1 = tri[topIdx1].z - tri[lowIdx].z;

			//float fy1 = static_cast<float>(yDiff0);
			//xDiff1 = xDiff1 / yDiff1 * fy1;
			//zDiff1 = zDiff1 / yDiff1 * fy1;

			//position.x = xDiff1;
			//position.y = fy1;
			//position.z = zDiff1;
			//Vec3i position3i = convertTo3i(position);
			//position3i = position3i + tri[lowIdx];
			//position = convertTo3f(position3i) * pixelSize;
			//addVoxel(build.refValue(), part, position, pixelSize);

			Vec3i& lowPoint = tri[lowIdx];
			int y = 1;
			calcEdgePositions(xDiff0, yDiff0, zDiff0, lowPoint, m_westEdge);
			plotEdgeVoxels(m_westEdge, pixelSize, part, build.refValue());

			calcEdgePositions(xDiff1, yDiff1, zDiff1, lowPoint, m_eastEdge);

			int yDiff2 = tri[topIdx0].y - tri[topIdx1].y;
			float xDiff2 = tri[topIdx0].x - tri[topIdx1].x;
			float zDiff2 = tri[topIdx0].z - tri[topIdx1].z;

			calcEdgePositions(xDiff2, yDiff2, zDiff2, tri[topIdx1], m_eastEdge);
			plotEdgeVoxels(m_eastEdge, pixelSize, part, build.refValue());

			//int xDiff1b = tri[topIdx0].x -  tri[topIdx1].x;
			//yPlotVoxelWedgeIn(y, xDiff0, xDiff1, yDiff0, yDiff1, zDiff0, zDiff1, lowPoint, pixelSize, part, build.refValue());


		}

	}
	g_stw->genNodeList();
	//setModellingMode();

	g_stw->paint();
	g_stw->commonPostInsertNodeStuff();
	setModellingMode();

	for(int k = 1; k < build.length; k++){
		BoxBuild* box = dynamic_cast<BoxBuild*>(build[k].p);
		for(int j = 0; j < 6; j++){
			NurbsPatch* patch = box->m_patch[j];
			Stem* stem = dynamic_cast<Stem*>(patch->stem.refValue()[0].p);
			stem->bands.refValue() = 1;
			stem->staves.refValue() = 1;
		}
		initBnspePolies(box, box->part.refValue());
	}


	populateMirrorNodes();

	resetLooseEnds();
	g_stw->paint();
}

