/* Seamless3d

Copyright (C) 2001-2018 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include <windows.h>
#include <GL/glew.h>
#include <initializer_list>
#include <iostream>
#include <fstream>
using namespace std;



#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "render-ogl.h"
#include "robot_aid.h"

int g_wireFrameMode = 0;
int g_waitMiliseconds = 0;
extern Ptr<PartCageClue> g_partCageClue;
extern Ptr<PartClue> g_vcArtTriBase;
extern SeamlessPtr g_vcTransferBCyl;
extern ShapePtr g_gridShape;

void setUpInternalEffectPtrsTmpForRenderingTransparancy(Seamless* smls);

extern bool g_fogExist;
extern bool g_minimized;
extern bool g_runAnim;
extern bool g_windowSizeWindowBool;
extern bool g_browseMode;

bool g_backfaceRender = false;

extern double g_windowSizeWindowTime;
bool windowSizeHasChanged();
extern Vec2i g_mousePos;

extern bool g_winAPI3DMode;
void moveMouse3d();
void processPackets();

extern ShapePtr g_vcEllipticalPolyline2D;
extern ShapePtr g_nurbsIlsShape;
extern ShapePtr g_transformLeverClue;
extern ShapePtr g_translateDimensionClue;
extern ShapePtr g_vcFuseSurfaceFlatStageAidShape;

extern SeamlessPtr g_cameraObject;
extern bool g_showCamerasInAnimMode;

extern TransformPtr g_redConeBase;
extern TransformPtr g_greenConeBase;
extern TransformPtr g_blueConeBase;

extern TransformPtr g_nurbsIlsBase;
extern ShapePtr g_vcSelectRectShape;

extern IndexedLineSet* g_triClue;
extern Coordinate* g_triClueCoord;
extern Color* g_triClueColor;
extern Shape* g_triClueShape;
extern bool g_curvedCageMode;

extern M_Matrix m_matrix;
extern M_Int m_smi;
extern M_Vec3f g_m_coord;

struct RenderLast {
	Seamless* smls;
	Matrix matrix;
	Matrix catRotMatrix;
	RenderLast(Seamless* smls, Matrix matrix, Matrix catRotMatrix) :
		smls(smls),
		matrix(matrix),
		catRotMatrix(catRotMatrix)
	{}
	RenderLast() :
		smls(0)
	{}

};

struct TransparencyList {
	Node* node;
	Matrix matrix;
	Matrix catRotMatrix;
	TransparencyList(Node* node, Matrix parentMatrix, Matrix catRotMatrix) :
		node(node),
		matrix(parentMatrix),
		catRotMatrix(catRotMatrix)
	{}
	TransparencyList() :
		node(0)
	{}

};

struct AlphaChanList {
	Node* node;
	Matrix matrix;
	Matrix catRotMatrix;
	AlphaChanList(Node* node, Matrix parentMatrix, Matrix catRotMatrix):
		node(node),
		matrix(parentMatrix),
		catRotMatrix(catRotMatrix)
	{}
	AlphaChanList() :
		node(0)
	{}

};

extern ScenePtr g_scene;

bool g_renderLater = false;

M_<RenderLast> g_skelList;
M_<AlphaChanList> g_alphaChanList;
M_<TransparencyList> g_transparentList;
M_Node renderProcessedQue;


Matrix projMatrix;


void CNTShader::initTexture(Vec2i textureSize, unsigned char* pixels) {

	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, textureSize.x, textureSize.y, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
}

void CNT0WeightBoneShader::initTexture(Vec2i textureSize, unsigned char* pixels) {

	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, textureSize.x, textureSize.y, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
}

void WinShader::initTexture(Vec2i textureSize, unsigned char* pixels) {

	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, textureSize.x, textureSize.y, 0, GL_BGRA, GL_UNSIGNED_BYTE, pixels);
}

void WinShader::render() {

	glUseProgram(program);
	glBindTexture(GL_TEXTURE_2D, texture);

	glBindVertexArray(vertexArrayObject[0]);

	glDrawElementsBaseVertex(GL_TRIANGLES, coordIndexLen, RENDER_INT, 0, 0);

	glBindVertexArray(0);
}


void printProgramInfoLog(GLuint obj)
{
	int infologLength = 0;
	int charsWritten = 0;
	char *infoLog;

	glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

	if (infologLength > 1)
	{
		infoLog = (char *)malloc(infologLength);
		glGetProgramInfoLog(obj, infologLength, &charsWritten, infoLog);
		dLog(infoLog);
		free(infoLog);
	}
}

void printShaderInfoLog(GLuint obj)
{
	int infologLength = 0;
	int charsWritten = 0;
	char *infoLog;

	glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

	if (infologLength > 1)
	{
		infoLog = (char *)malloc(infologLength);
		glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);
		dLog(infoLog);
		free(infoLog);
	}
}

const char* vCMonoCStr = "\
#version 150\n\
\
uniform mat4 projMatrix;\n\
uniform mat4 viewMatrix;\n\
uniform mat4 rotationMatrix;\n\
uniform mat4 translationMatrix;\n\
\
in vec4 a_coord;\n\
\
void main() {\n\
	gl_Position = projMatrix * viewMatrix * a_coord;\n\
\
}";

const char* fCMonoCStr = "\
#version 150\n\
\
out vec4 fragOut;\n\
uniform vec4 f_color;\n\
\
void main()\n\
{\n\
	fragOut = vec4(f_color.xyz, 1);\n\
}";

const char* vCNMonoCStr = "\
#version 150\n\
\
uniform mat4 projMatrix;\n\
uniform mat4 viewMatrix;\n\
uniform mat4 rotationMatrix;\n\
uniform mat4 translationMatrix;\n\
\
in vec4 a_coord;\n\
in vec3 a_normal;\n\
\
out vec3 v_normal;\n\
out float v_Dot;\n\
\
void main() {\n\
	gl_Position = projMatrix * viewMatrix * a_coord;\n\
\
	vec4 transNormal = rotationMatrix * vec4(a_normal, 1);\n\
	v_Dot = max(dot(transNormal.xyz, vec3(0.0, 0.0, 1.0)), 0.0);\n\
\
	v_normal = a_normal;\n\
}";

const char* fCNMonoCStr = "\
#version 150\n\
\
in vec3 v_normal;\n\
in vec3 v_color;\n\
\
in float v_Dot;\n\
\
in vec3 v_Color;\n\
\
out vec4 fragOut;\n\
uniform vec4 f_color;\n\
\
void main()\n\
{\n\
	fragOut = vec4(f_color.xyz * v_Dot, 1);\n\
}";

const char* vCCStr = "\
#version 150\n\
\
uniform mat4 projMatrix;\n\
uniform mat4 viewMatrix;\n\
uniform mat4 rotationMatrix;\n\
uniform mat4 translationMatrix;\n\
\
in vec4 a_coord;\n\
in vec3 a_color;\n\
\
out vec3 v_color;\n\
out float v_Dot;\n\
\
void main() {\n\
	gl_Position = projMatrix * viewMatrix * a_coord;\n\
\
	v_color = a_color;\n\
\
}";

const char* fCCStr = "\
#version 150\n\
\
in vec3 v_color;\n\
\
in float v_Dot;\n\
\
in vec3 v_Color;\n\
\
out vec4 fragOut;\n\
\
void main()\n\
{\n\
	fragOut = vec4(v_color.xyz, 1);\n\
}";


const char* vCNCStr = "\
#version 150\n\
\
uniform mat4 projMatrix;\n\
uniform mat4 viewMatrix;\n\
uniform mat4 rotationMatrix;\n\
uniform mat4 translationMatrix;\n\
\
in vec4 a_coord;\n\
in vec3 a_normal;\n\
in vec3 a_color;\n\
\
out vec3 v_normal;\n\
out vec3 v_color;\n\
out float v_Dot;\n\
\
void main() {\n\
	gl_Position = projMatrix * viewMatrix * a_coord;\n\
\
	v_color = a_color;\n\
\
	vec4 transNormal = rotationMatrix * vec4(a_normal, 1);\n\
	v_Dot = max(dot(transNormal.xyz, vec3(0.0, 0.0, 1.0)), 0.0);\n\
\
	v_normal = a_normal;\n\
}";

const char* fCNCStr = "\
#version 150\n\
\
in vec3 v_normal;\n\
in vec3 v_color;\n\
\
in float v_Dot;\n\
\
in vec3 v_Color;\n\
uniform float a_transparency;\n\
\
out vec4 fragOut;\n\
\
void main()\n\
{\n\
	fragOut = vec4(v_color.xyz * v_Dot, a_transparency);\n\
}";

const char* vCNTStr = "\
#version 330\n\
\
uniform mat4 rotationMatrix;\n\
uniform mat4 projMatrix;\n\
uniform mat4 viewMatrix;\n\
\
in vec4 a_coord;\n\
in vec4 a_normal;\n\
in vec2 a_texCoord;\n\
\
out float v_Dot;\n\
\
out vec3 v_normal;\
out vec2 v_texCoord;\
\
void main() {\
\
	v_normal = normalize(rotationMatrix * a_normal).xyz;\
	v_texCoord = a_texCoord;\
\
	gl_Position = projMatrix * viewMatrix * a_coord;\
}";


const char* fCNTStr = "\
#version 330\n\
\
vec3 l_dir = vec3(0,0,1);\
\
in vec3 v_normal;\
in vec2 v_texCoord;\
\
uniform sampler2D diffuseMap;\n\
\
out vec4 fragOut;\
\
void main() {\
\
	vec3 n = normalize(v_normal);\
\
	float intensity = max(dot(n, l_dir), 0.0);\
\
	vec4 diffuseColour = texture(diffuseMap, v_texCoord);\n\
	fragOut = vec4(diffuseColour.xyz * intensity, diffuseColour.a);\
}";

const char* vCNC0WeightBoneStr = "\
#version 150\n\
\
uniform mat4 projMatrix;\n\
uniform mat4 boneMatrix[100];\n\
uniform mat4 rotBoneMatrix[100];\n\
uniform mat4 rotationMatrix;\n\
uniform mat4 translationMatrix;\n\
\
uniform vec3 dlook = vec3(0.0, 0.0, 1.0);\n\
\
in vec4 a_coord;\n\
in vec3 a_normal;\n\
in vec3 a_color;\n\
in int a_boneIndex;\n\
\
out vec3 v_normal;\n\
out vec3 v_color;\n\
out float v_Dot;\n\
\
void main() {\n\
	vec4 newVertex;\n\
	int index;\n\
	index = a_boneIndex;\n\
	newVertex = boneMatrix[index] * a_coord;\n\
	gl_Position = projMatrix * newVertex;\n\
\
	v_color = a_color;\n\
\
	vec4 transNormal = rotationMatrix * rotBoneMatrix[index] * vec4(a_normal, 1);\n\
	v_Dot = max(dot(transNormal.xyz, dlook), 0.0);\n\
\
	v_normal = transNormal.xyz;\n\
}";

const char* fCNC0WeightBoneStr = "\
#version 150\n\
\
in vec3 v_normal;\n\
in vec3 v_color;\n\
\
in float v_Dot;\n\
\
in vec3 v_Color;\n\
\
out vec4 fragOut;\n\
\
void main()\n\
{\n\
	fragOut = vec4(v_color.xyz * v_Dot, 1);\n\
}";

const char* vCNT0WeightBoneStr = "\
#version 150\n\
\
uniform mat4 projMatrix;\n\
uniform mat4 boneMatrix[100];\n\
uniform mat4 rotBoneMatrix[100];\n\
uniform mat4 rotationMatrix;\n\
uniform mat4 translationMatrix;\n\
\
uniform vec3 dlook = vec3(0.0, 0.0, 1.0);\n\
\
in vec4 a_coord;\n\
in vec3 a_normal;\n\
in vec2 a_texCoord;\n\
in int a_boneIndex;\n\
\
out vec3 v_normal;\n\
out vec2 v_texCoord;\n\
out float v_Dot;\n\
\
void main() {\n\
	vec4 newVertex;\n\
	int index;\n\
	index = a_boneIndex;\n\
	newVertex = boneMatrix[index] * a_coord;\n\
	gl_Position = projMatrix * newVertex;\n\
\
	v_texCoord = a_texCoord;\n\
\
	vec4 transNormal = rotationMatrix * rotBoneMatrix[index] * vec4(a_normal, 1);\n\
	v_Dot = max(dot(transNormal.xyz, dlook), 0.0);\n\
\
	v_normal = transNormal.xyz;\n\
}";

const char* fCNT0WeightBoneStr = "\
#version 330\n\
\
in vec3 v_normal;\
in vec2 v_texCoord;\
in float v_Dot;\n\
\
uniform sampler2D diffuseMap;\n\
\
out vec4 fragOut;\
\
void main() {\
\
	//float intensity = max(v_Dot, 0.0);\n\
\
	vec4 diffuseColour = texture(diffuseMap, v_texCoord);\n\
	//fragOut = vec4(diffuseColour.xyz * intensity, diffuseColour.a);\n\
	fragOut = vec4(diffuseColour.xyz * v_Dot, diffuseColour.a);\n\
}";


const char* vWinStr = "\
#version 330\n\
\
in vec2 a_coord;\n\
in vec2 a_texCoord;\n\
\
out vec2 v_texCoord;\n\
out float v_Dot;\n\
\
void main() {\n\
	gl_Position = vec4(a_coord.x, a_coord.y, 0, 1);\n\
	v_texCoord = a_texCoord;\n\
}";

const char* fWinStr = "\
#version 330\n\
\
in vec2 v_texCoord;\n\
\
uniform sampler2D sampler;\n\
\
out vec4 fragOut;\n\
\
void main()\n\
{\n\
	fragOut = texture2D(sampler, v_texCoord);\n\
}";


void CMonoCLineShader::setupShaders() {

	v = glCreateShader(GL_VERTEX_SHADER);
	f = glCreateShader(GL_FRAGMENT_SHADER);

	const char * vv = vCMonoCStr;
	const char * ff = fCMonoCStr;

	glShaderSource(v, 1, &vv, NULL);
	glShaderSource(f, 1, &ff, NULL);

	glCompileShader(v);
	glCompileShader(f);

	printShaderInfoLog(v);
	printShaderInfoLog(f);

	program = glCreateProgram();
	glAttachShader(program, v);
	glAttachShader(program, f);
	glBindFragDataLocation(program, 0, "fragOut");

	glLinkProgram(program);
	printProgramInfoLog(program);

	glGenVertexArrays(1, vertexArrayObject);
	glBindVertexArray(vertexArrayObject[0]);

	coordLoc = glGetAttribLocation(program, "a_coord");

	// Uniform variable Locations

	projMatrixLoc = glGetUniformLocation(program, "projMatrix");
	viewMatrixLoc = glGetUniformLocation(program, "viewMatrix");
	translationMatrixLoc = glGetUniformLocation(program, "translationMatrix");
	rotationMatrixLoc = glGetUniformLocation(program, "rotationMatrix");
	f_colorLoc = glGetUniformLocation(program, "f_color");
}

void CNMonoCShader::setupShaders() {

	v = glCreateShader(GL_VERTEX_SHADER);
	f = glCreateShader(GL_FRAGMENT_SHADER);

	const char * vv = vCNMonoCStr;
	const char * ff = fCNMonoCStr;

	glShaderSource(v, 1, &vv, NULL);
	glShaderSource(f, 1, &ff, NULL);

	glCompileShader(v);
	glCompileShader(f);

	printShaderInfoLog(v);
	printShaderInfoLog(f);

	program = glCreateProgram();
	glAttachShader(program, v);
	glAttachShader(program, f);
	glBindFragDataLocation(program, 0, "fragOut");

	glLinkProgram(program);
	printProgramInfoLog(program);

	glGenVertexArrays(1, vertexArrayObject);
	glBindVertexArray(vertexArrayObject[0]);

	coordLoc = glGetAttribLocation(program, "a_coord");
	normalLoc = glGetAttribLocation(program, "a_normal");

	// Uniform variable Locations

	projMatrixLoc = glGetUniformLocation(program, "projMatrix");
	viewMatrixLoc = glGetUniformLocation(program, "viewMatrix");
	translationMatrixLoc = glGetUniformLocation(program, "translationMatrix");
	rotationMatrixLoc = glGetUniformLocation(program, "rotationMatrix");
	f_colorLoc = glGetUniformLocation(program, "f_color");
}

void CCLineShader::setupShaders() {

	v = glCreateShader(GL_VERTEX_SHADER);
	f = glCreateShader(GL_FRAGMENT_SHADER);

	const char * vv = vCCStr;
	const char * ff = fCCStr;

	glShaderSource(v, 1, &vv, NULL);
	glShaderSource(f, 1, &ff, NULL);

	glCompileShader(v);
	glCompileShader(f);

	printShaderInfoLog(v);
	printShaderInfoLog(f);

	program = glCreateProgram();
	glAttachShader(program, v);
	glAttachShader(program, f);
	glBindFragDataLocation(program, 0, "fragOut");

	glLinkProgram(program);
	printProgramInfoLog(program);

	glGenVertexArrays(1, vertexArrayObject);
	glBindVertexArray(vertexArrayObject[0]);

	coordLoc = glGetAttribLocation(program, "a_coord");
	colorLoc = glGetAttribLocation(program, "a_color");

	// Uniform variable Locations

	projMatrixLoc = glGetUniformLocation(program, "projMatrix");
	viewMatrixLoc = glGetUniformLocation(program, "viewMatrix");
	translationMatrixLoc = glGetUniformLocation(program, "translationMatrix");
	rotationMatrixLoc = glGetUniformLocation(program, "rotationMatrix");
}


void CNCShader::setupShaders() {

	v = glCreateShader(GL_VERTEX_SHADER);
	f = glCreateShader(GL_FRAGMENT_SHADER);

	const char * vv = vCNCStr;
	const char * ff = fCNCStr;

	glShaderSource(v, 1, &vv, NULL);
	glShaderSource(f, 1, &ff, NULL);

	glCompileShader(v);
	glCompileShader(f);

	printShaderInfoLog(v);
	printShaderInfoLog(f);

	program = glCreateProgram();
	glAttachShader(program, v);
	glAttachShader(program, f);
	glBindFragDataLocation(program, 0, "fragOut");

	glLinkProgram(program);
	printProgramInfoLog(program);

	glGenVertexArrays(1, vertexArrayObject);
	glBindVertexArray(vertexArrayObject[0]);

	coordLoc = glGetAttribLocation(program, "a_coord");
	normalLoc = glGetAttribLocation(program, "a_normal");
	colorLoc = glGetAttribLocation(program, "a_color");
	transparencyLoc = glGetUniformLocation(program, "a_transparency");

	// Uniform variable Locations

	projMatrixLoc = glGetUniformLocation(program, "projMatrix");
	viewMatrixLoc = glGetUniformLocation(program, "viewMatrix");
	scaleMatrixLoc = glGetUniformLocation(program, "scaleMatrix");
	translationMatrixLoc = glGetUniformLocation(program, "translationMatrix");
	rotationMatrixLoc = glGetUniformLocation(program, "rotationMatrix");
}

void CNTShader::setupShaders() {

	v = glCreateShader(GL_VERTEX_SHADER);
	f = glCreateShader(GL_FRAGMENT_SHADER);

	const char * vv = vCNTStr;
	const char * ff = fCNTStr;

	glShaderSource(v, 1, &vv, NULL);
	glShaderSource(f, 1, &ff, NULL);

	glCompileShader(v);
	glCompileShader(f);

	printShaderInfoLog(v);
	printShaderInfoLog(f);

	program = glCreateProgram();
	glAttachShader(program, v);
	glAttachShader(program, f);
	glBindFragDataLocation(program, 0, "fragOut");

	glLinkProgram(program);
	printProgramInfoLog(program);

	glGenVertexArrays(1, vertexArrayObject);
	glBindVertexArray(vertexArrayObject[0]);

	coordLoc = glGetAttribLocation(program, "a_coord");
	normalLoc = glGetAttribLocation(program, "a_normal");
	texCoordLoc = glGetAttribLocation(program, "a_texCoord");

	// Uniform variable Locations

	projMatrixLoc = glGetUniformLocation(program, "projMatrix");
	viewMatrixLoc = glGetUniformLocation(program, "viewMatrix");
	scaleMatrixLoc = glGetUniformLocation(program, "scaleMatrix");
	translationMatrixLoc = glGetUniformLocation(program, "translationMatrix");
	rotationMatrixLoc = glGetUniformLocation(program, "rotationMatrix");
}

void CNC0WeightBoneShader::setupShaders() {

	v = glCreateShader(GL_VERTEX_SHADER);
	f = glCreateShader(GL_FRAGMENT_SHADER);

	const char * vv = vCNC0WeightBoneStr;
	const char * ff = fCNC0WeightBoneStr;

	glShaderSource(v, 1, &vv, NULL);
	glShaderSource(f, 1, &ff, NULL);

	glCompileShader(v);
	glCompileShader(f);

	printShaderInfoLog(v);
	printShaderInfoLog(f);

	program = glCreateProgram();
	glAttachShader(program, v);
	glAttachShader(program, f);
	glBindFragDataLocation(program, 0, "fragOut");

	glLinkProgram(program);
	printProgramInfoLog(program);

	glGenVertexArrays(1, vertexArrayObject);
	glBindVertexArray(vertexArrayObject[0]);

	coordLoc = glGetAttribLocation(program, "a_coord");
	normalLoc = glGetAttribLocation(program, "a_normal");
	colorLoc = glGetAttribLocation(program, "a_color");
	boneIndexLoc = glGetAttribLocation(program, "a_boneIndex");

	lookDirLoc = glGetUniformLocation(program, "dlook");

	// Uniform variable Locations

	projMatrixLoc = glGetUniformLocation(program, "projMatrix");
	boneMatrixLoc = glGetUniformLocation(program, "boneMatrix");
	rotBoneMatrixLoc = glGetUniformLocation(program, "rotBoneMatrix");
	translationMatrixLoc = glGetUniformLocation(program, "translationMatrix");
	rotationMatrixLoc = glGetUniformLocation(program, "rotationMatrix");

}

void CNT0WeightBoneShader::setupShaders() {

	v = glCreateShader(GL_VERTEX_SHADER);
	f = glCreateShader(GL_FRAGMENT_SHADER);

	const char * vv = vCNT0WeightBoneStr;
	const char * ff = fCNT0WeightBoneStr;

	glShaderSource(v, 1, &vv, NULL);
	glShaderSource(f, 1, &ff, NULL);

	glCompileShader(v);
	glCompileShader(f);

	printShaderInfoLog(v);
	printShaderInfoLog(f);

	program = glCreateProgram();
	glAttachShader(program, v);
	glAttachShader(program, f);
	glBindFragDataLocation(program, 0, "fragOut");

	glLinkProgram(program);
	printProgramInfoLog(program);

	glGenVertexArrays(1, vertexArrayObject);
	glBindVertexArray(vertexArrayObject[0]);

	coordLoc = glGetAttribLocation(program, "a_coord");
	normalLoc = glGetAttribLocation(program, "a_normal");
	texCoordLoc = glGetAttribLocation(program, "a_texCoord");
	boneIndexLoc = glGetAttribLocation(program, "a_boneIndex");

	lookDirLoc = glGetUniformLocation(program, "dlook");

	// Uniform variable Locations

	projMatrixLoc = glGetUniformLocation(program, "projMatrix");
	rotBoneMatrixLoc = glGetUniformLocation(program, "rotBoneMatrix");
	translationMatrixLoc = glGetUniformLocation(program, "translationMatrix");
	rotationMatrixLoc = glGetUniformLocation(program, "rotationMatrix");
}
void WinShader::setupShaders() {

	v = glCreateShader(GL_VERTEX_SHADER);
	f = glCreateShader(GL_FRAGMENT_SHADER);

	const char * vv = vWinStr;
	const char * ff = fWinStr;

	glShaderSource(v, 1, &vv, NULL);
	glShaderSource(f, 1, &ff, NULL);

	glCompileShader(v);
	glCompileShader(f);

	printShaderInfoLog(v);
	printShaderInfoLog(f);

	program = glCreateProgram();
	glAttachShader(program, v);
	glAttachShader(program, f);
	glBindFragDataLocation(program, 0, "fragOut");

	glLinkProgram(program);
	printProgramInfoLog(program);

	glGenVertexArrays(1, vertexArrayObject);
	glBindVertexArray(vertexArrayObject[0]);

	coordLoc = glGetAttribLocation(program, "a_coord");
	texCoordLoc = glGetAttribLocation(program, "a_texCoord");
}

void CMonoCLineShader::deleteProgram() {
	glDeleteBuffers(arrayBufferLen, arrayBuffer);
	glDeleteVertexArrays(1, vertexArrayObject);


	glDetachShader(program, v);
	glDeleteShader(v);

	glDetachShader(program, f);
	glDeleteShader(f);

	glDeleteProgram(program);
}

void CNMonoCShader::deleteProgram() {
	glDeleteBuffers(arrayBufferLen, arrayBuffer);
	glDeleteVertexArrays(1, vertexArrayObject);


	glDetachShader(program, v);
	glDeleteShader(v);

	glDetachShader(program, f);
	glDeleteShader(f);

	glDeleteProgram(program);
}

void CCLineShader::deleteProgram() {
	glDeleteBuffers(arrayBufferLen, arrayBuffer);
	glDeleteVertexArrays(1, vertexArrayObject);


	glDetachShader(program, v);
	glDeleteShader(v);

	glDetachShader(program, f);
	glDeleteShader(f);

	glDeleteProgram(program);
}


void CNCShader::deleteProgram() {
	glDeleteBuffers(arrayBufferLen, arrayBuffer);
	glDeleteVertexArrays(1, vertexArrayObject);


	glDetachShader(program, v);
	glDeleteShader(v);

	glDetachShader(program, f);
	glDeleteShader(f);

	glDeleteProgram(program);
}

void CNTShader::deleteProgram() {
	glDeleteBuffers(arrayBufferLen, arrayBuffer);
	glDeleteVertexArrays(1, vertexArrayObject);


	glDetachShader(program, v);
	glDeleteShader(v);

	glDetachShader(program, f);
	glDeleteShader(f);

	glDeleteProgram(program);
}

void CNC0WeightBoneShader::deleteProgram() {
	glDeleteBuffers(arrayBufferLen, arrayBuffer);
	glDeleteVertexArrays(1, vertexArrayObject);


	glDetachShader(program, v);
	glDeleteShader(v);

	glDetachShader(program, f);
	glDeleteShader(f);

	glDeleteProgram(program);
}

void CNT0WeightBoneShader::deleteProgram() {
	glDeleteBuffers(arrayBufferLen, arrayBuffer);
	glDeleteVertexArrays(1, vertexArrayObject);


	glDetachShader(program, v);
	glDeleteShader(v);

	glDetachShader(program, f);
	glDeleteShader(f);

	glDeleteProgram(program);
}

void WinShader::deleteProgram() {
	glDeleteBuffers(arrayBufferLen, arrayBuffer);
	glDeleteVertexArrays(1, vertexArrayObject);


	glDetachShader(program, v);
	glDeleteShader(v);

	glDetachShader(program, f);
	glDeleteShader(f);

	glDeleteProgram(program);
}


void CMonoCLineShader::initMesh() {

	glGenBuffers(arrayBufferLen, arrayBuffer);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, coord, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(coordLoc);
	glVertexAttribPointer(coordLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[1]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);

	glBindVertexArray(0);
}

void CNMonoCShader::initMesh() {

	glGenBuffers(arrayBufferLen, arrayBuffer);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, coord, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(coordLoc);
	glVertexAttribPointer(coordLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[1]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, normal, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(normalLoc);
	glVertexAttribPointer(normalLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[2]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);

	glBindVertexArray(0);
}

void CCLineShader::initMesh() {

	glGenBuffers(arrayBufferLen, arrayBuffer);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, coord, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(coordLoc);
	glVertexAttribPointer(coordLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[1]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, color, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(colorLoc);
	glVertexAttribPointer(colorLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[2]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);

	glBindVertexArray(0);
}


void CNCShader::initMesh() {

	glUseProgram(program);
	glGenBuffers(arrayBufferLen, arrayBuffer);
	glBindVertexArray(vertexArrayObject[0]);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, coord, GL_DYNAMIC_DRAW);
	//glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(Vec3f) * vertexLen, coord);
	glEnableVertexAttribArray(coordLoc);
	glVertexAttribPointer(coordLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[1]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, normal, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(normalLoc);
	glVertexAttribPointer(normalLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[2]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, color, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(colorLoc);
	glVertexAttribPointer(colorLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[3]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);
	//glBufferSubData(GL_ELEMENT_ARRAY_BUFFER, 0, sizeof(RenderInt) * coordIndexLen, coordIndex);

	glBindVertexArray(0);

}

void CNTShader::initMesh() {

	glUseProgram(program);
	glGenBuffers(arrayBufferLen, arrayBuffer);
	glBindVertexArray(vertexArrayObject[0]);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, coord, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(coordLoc);
	glVertexAttribPointer(coordLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[1]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, normal, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(normalLoc);
	glVertexAttribPointer(normalLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[2]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec2f) * vertexLen, texCoord, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(texCoordLoc);
	glVertexAttribPointer(texCoordLoc, 2, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[3]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);

	glBindVertexArray(0);

}


void CNC0WeightBoneShader::initMesh() {

	glGenBuffers(arrayBufferLen, arrayBuffer);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, coord, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(coordLoc);
	glVertexAttribPointer(coordLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[1]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, normal, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(normalLoc);
	glVertexAttribPointer(normalLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[2]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, color, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(colorLoc);
	glVertexAttribPointer(colorLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[3]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(int) * vertexLen, boneIndex.getArray(), GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(boneIndexLoc);
	glVertexAttribPointer(boneIndexLoc, 1, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[4]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);

	glBindVertexArray(0);
}

void CNT0WeightBoneShader::initMesh() {

	glGenBuffers(arrayBufferLen, arrayBuffer);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, coord, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(coordLoc);
	glVertexAttribPointer(coordLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[1]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, normal, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(normalLoc);
	glVertexAttribPointer(normalLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[2]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec2f) * vertexLen, texCoord, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(texCoordLoc);
	glVertexAttribPointer(texCoordLoc, 2, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[3]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(int) * vertexLen, boneIndex.getArray(), GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(boneIndexLoc);
	glVertexAttribPointer(boneIndexLoc, 1, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[4]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);

	glBindVertexArray(0);
}

void WinShader::initMesh() {

	glGenBuffers(arrayBufferLen, arrayBuffer);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec2f) * vertexLen, coord, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(coordLoc);
	glVertexAttribPointer(coordLoc, 2, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[1]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vec2f) * vertexLen, texCoord, GL_DYNAMIC_DRAW);
	glEnableVertexAttribArray(texCoordLoc);
	glVertexAttribPointer(texCoordLoc, 2, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[2]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);

	glBindVertexArray(0);
}

void CMonoCLineShader::render()
{

	glUseProgram(program);


	static float za = 0;

	za += .01f;


	Matrix  scaleMat = scaleMatrix(Vec3f(1, 1, 1));
	Matrix zRotationMatrix = rotationMatrix(g_scene->rotation);

	//Vec3f translation = Vec3f(0.0, -1.65, 0.0);
	Vec3f translation = Vec3f(0.0, 0.0, 0.0);

	Matrix translationMatrix = {
		1, 0, 0, 0,
		0, 1, 0, 0,
		0, 0, 1, 0,
		translation.x, translation.y, translation.z, 1
	};

	glUniformMatrix4fv(scaleMatrixLoc, 1, GL_FALSE, &scaleMat.f11);
	glUniformMatrix4fv(translationMatrixLoc, 1, GL_FALSE, &translationMatrix.f11);
	glUniformMatrix4fv(rotationMatrixLoc, 1, GL_FALSE, &zRotationMatrix.f11);
	glUniformMatrix4fv(projMatrixLoc, 1, GL_FALSE, &projMatrix.f11);
	glUniformMatrix4fv(viewMatrixLoc, 1, GL_FALSE, &viewMatrix.f11);

	glUniform4f(f_colorLoc, f_color.r, f_color.g, f_color.b, 1.0f);

	glBindVertexArray(vertexArrayObject[0]);

	glDrawElementsBaseVertex(GL_LINES, coordIndexLen, RENDER_INT, 0, 0);

	glBindVertexArray(0);
}

void CNMonoCShader::render()
{

	glUseProgram(program);


	static float za = 0;

	za += .01f;


	Matrix  scaleMat = scaleMatrix(Vec3f(1, 1, 1));
	Matrix zRotationMatrix = rotationMatrix(g_scene->rotation);

	//Vec3f translation = Vec3f(0.0, -1.65, 0.0);
	Vec3f translation = Vec3f(0.0, 0.0, 0.0);

	Matrix translationMatrix = {
		1, 0, 0, 0,
		0, 1, 0, 0,
		0, 0, 1, 0,
		translation.x, translation.y, translation.z, 1
	};

	glUniformMatrix4fv(scaleMatrixLoc, 1, GL_FALSE, &scaleMat.f11);
	glUniformMatrix4fv(translationMatrixLoc, 1, GL_FALSE, &translationMatrix.f11);
	glUniformMatrix4fv(rotationMatrixLoc, 1, GL_FALSE, &zRotationMatrix.f11);
	glUniformMatrix4fv(projMatrixLoc, 1, GL_FALSE, &projMatrix.f11);
	glUniformMatrix4fv(viewMatrixLoc, 1, GL_FALSE, &viewMatrix.f11);

	glUniform4f(f_colorLoc, f_color.r, f_color.g, f_color.b, 1.0f);

	glBindVertexArray(vertexArrayObject[0]);

	glDrawElementsBaseVertex(GL_TRIANGLES, coordIndexLen, RENDER_INT, 0, 0);

	glBindVertexArray(0);
}

void CCLineShader::render()
{

	glUseProgram(program);


	static float za = 0;

	za += .01f;


	Matrix  scaleMat = scaleMatrix(Vec3f(1, 1, 1));
	Matrix zRotationMatrix = rotationMatrix(g_scene->rotation);

	//Vec3f translation = Vec3f(0.0, -1.65, 0.0);
	Vec3f translation = Vec3f(0.0, 0, 0.0);

	Matrix translationMatrix = {
		1, 0, 0, 0,
		0, 1, 0, 0,
		0, 0, 1, 0,
		translation.x, translation.y, translation.z, 1
	};

	glUniformMatrix4fv(scaleMatrixLoc, 1, GL_FALSE, &scaleMat.f11);
	glUniformMatrix4fv(translationMatrixLoc, 1, GL_FALSE, &translationMatrix.f11);
	glUniformMatrix4fv(rotationMatrixLoc, 1, GL_FALSE, &zRotationMatrix.f11);
	glUniformMatrix4fv(projMatrixLoc, 1, GL_FALSE, &projMatrix.f11);
	glUniformMatrix4fv(viewMatrixLoc, 1, GL_FALSE, &viewMatrix.f11);

	glBindVertexArray(vertexArrayObject[0]);

	glDrawElementsBaseVertex(GL_LINES, coordIndexLen, RENDER_INT, 0, 0);

	glBindVertexArray(0);
}


void CNCShader::render()
{

	glUseProgram(program);

	glUniformMatrix4fv(rotationMatrixLoc, 1, GL_FALSE, &rotationMat.f11);
	glUniformMatrix4fv(projMatrixLoc, 1, GL_FALSE, &projMatrix.f11);
	glUniformMatrix4fv(viewMatrixLoc, 1, GL_FALSE, &viewMatrix.f11);

	float t = 1.0 - transparency;
	glUniform1f(transparencyLoc, t);

	glBindVertexArray(vertexArrayObject[0]);

	if(transparency != 0) {
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDrawElementsBaseVertex(GL_TRIANGLES, coordIndexLen, RENDER_INT, 0, 0);
		glDisable(GL_BLEND);
	}
	else {
		glDrawElementsBaseVertex(GL_TRIANGLES, coordIndexLen, RENDER_INT, 0, 0);
	}

	glBindVertexArray(0);
}

void CNTShader::render()
{
	glUseProgram(program);
	glBindTexture(GL_TEXTURE_2D, texture);

	glUniformMatrix4fv(rotationMatrixLoc, 1, GL_FALSE, &rotationMat.f11);
	glUniformMatrix4fv(projMatrixLoc, 1, GL_FALSE, &projMatrix.f11);
	glUniformMatrix4fv(viewMatrixLoc, 1, GL_FALSE, &viewMatrix.f11);

	glBindVertexArray(vertexArrayObject[0]);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDrawElementsBaseVertex(GL_TRIANGLES, coordIndexLen, RENDER_INT, 0, 0);

	glBindVertexArray(0);
	glDisable(GL_BLEND);
}

void CNC0WeightBoneShader::render()
{

	glUseProgram(program);

	Vec3f lookDir = g_scene->rotation * Vec3f(0, 0, 1);
	glUniform3f(lookDirLoc, lookDir.x, lookDir.y, lookDir.z);

	glUniformMatrix4fv(rotationMatrixLoc, 1, GL_FALSE, &rotationMat.f11);
	glUniformMatrix4fv(projMatrixLoc, 1, GL_FALSE, &projMatrix.f11);
	glUniformMatrix4fv(viewMatrixLoc, 1, GL_FALSE, &viewMatrix.f11);
	glUniformMatrix4fv(boneMatrixLoc, boneMatrix.length, GL_FALSE, &boneMatrix.getArray()->f11);
	glUniformMatrix4fv(rotBoneMatrixLoc, rotBoneMatrix.length, GL_FALSE, &rotBoneMatrix.getArray()->f11);

	glBindVertexArray(vertexArrayObject[0]);

	glDrawElementsBaseVertex(GL_TRIANGLES, coordIndexLen, RENDER_INT, 0, 0);

	glBindVertexArray(0);
}

void CNT0WeightBoneShader::render()
{

	glUseProgram(program);

	glBindTexture(GL_TEXTURE_2D, texture);

	Vec3f lookDir = g_scene->rotation * Vec3f(0, 0, 1);
	//Vec3f lookDir = Vec3f(0, 0, 1);
	glUniform3f(lookDirLoc, lookDir.x, lookDir.y, lookDir.z);
	//glUniform3fv(lookDirLoc, 3, &lookDir.x);

	Vec3f empty;
	glGetUniformfv(program, lookDirLoc, &empty.x);

	glUniformMatrix4fv(rotationMatrixLoc, 1, GL_FALSE, &rotationMat.f11);
	glUniformMatrix4fv(projMatrixLoc, 1, GL_FALSE, &projMatrix.f11);
	glUniformMatrix4fv(viewMatrixLoc, 1, GL_FALSE, &viewMatrix.f11);
	glUniformMatrix4fv(boneMatrixLoc, boneMatrix.length, GL_FALSE, &boneMatrix.getArray()->f11);
	glUniformMatrix4fv(rotBoneMatrixLoc, rotBoneMatrix.length, GL_FALSE, &rotBoneMatrix.getArray()->f11);

	glBindVertexArray(vertexArrayObject[0]);

	glDrawElementsBaseVertex(GL_TRIANGLES, coordIndexLen, RENDER_INT, 0, 0);

	glBindVertexArray(0);
}

void CMonoCLineShader::convertToRenderFormat(Mesh& mesh) {
	vertexLen = mesh.coord.length;
	coordIndexLen = mesh.coordIndex.length;

	coord = reinterpret_cast<const float*>(mesh.coord.getArray());
	coordIndex = reinterpret_cast<RenderInt*>(mesh.coordIndex.getArray());

	/*

	coordIndex = new RenderInt[i_coordIndex.length];

	for (int j = 0; j < i_coordIndex.length; j++) {
		int g = i_coordIndex[j];
		coordIndex[j] = i_coordIndex[j];
	}
	*/

	if (firstTime) {
		setupShaders();

		initMesh();
	}

	else {
		glUseProgram(program);

		glBindVertexArray(vertexArrayObject[0]);

		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, coord, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[1]);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);

		//glBindVertexArray(0);
	}


	firstTime = false;
}

void CNMonoCShader::convertToRenderFormat(Mesh& mesh) {
	vertexLen = mesh.coord.length;
	coordIndexLen = mesh.coordIndex.length;

	coord = reinterpret_cast<const float*>(mesh.coord.getArray());
	normal = reinterpret_cast<const float*>(mesh.normal.getArray());

	coordIndex = reinterpret_cast<RenderInt*>(mesh.coordIndex.getArray());


	if (firstTime) {
		setupShaders();

		initMesh();
	}

	else {
		glUseProgram(program);

		glBindVertexArray(vertexArrayObject[0]);

		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, coord, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[1]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, normal, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[2]);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);

		//glBindVertexArray(0);
	}


	firstTime = false;
}

void CCLineShader::convertToRenderFormat(Mesh& mesh) {
	
	int meshCoordLen = mesh.coord.length;
	vertexLen = mesh.color.length;
	coordIndexLen = mesh.coordIndex.length;

	coord = reinterpret_cast<const float*>(mesh.coord.getArray());
	color = reinterpret_cast<const float*>(mesh.color.getArray());

	//coordIndex = new RenderInt[i_coordIndex.length];
	coordIndex = reinterpret_cast<RenderInt*>(mesh.coordIndex.getArray());
/*
	for (int j = 0; j < i_coordIndex.length; j++) {
		int g = i_coordIndex[j];
		coordIndex[j] = i_coordIndex[j];
	}
*/

	if (firstTime) {
		setupShaders();

		initMesh();
	}

	else {
		glUseProgram(program);

		glBindVertexArray(vertexArrayObject[0]);

		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, coord, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[1]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, color, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[2]);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);

		//glBindVertexArray(0);
	}


	firstTime = false;
}

void CCLineShader::convertToRenderFormat(Line& mesh) {

	int meshCoordLen = mesh.coord.length;
	vertexLen = mesh.color.length;
	coordIndexLen = mesh.coordIndex.length;

	coord = reinterpret_cast<const float*>(mesh.coord.getArray());
	color = reinterpret_cast<const float*>(mesh.color.getArray());

	//coordIndex = new RenderInt[i_coordIndex.length];
	coordIndex = reinterpret_cast<RenderInt*>(mesh.coordIndex.getArray());
	/*
		for (int j = 0; j < i_coordIndex.length; j++) {
			int g = i_coordIndex[j];
			coordIndex[j] = i_coordIndex[j];
		}
	*/

	if (firstTime) {
		setupShaders();

		initMesh();
	}

	else {
		glUseProgram(program);

		glBindVertexArray(vertexArrayObject[0]);

		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, coord, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[1]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, color, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[2]);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);

		//glBindVertexArray(0);
	}


	firstTime = false;
}

void Mesh::matchCoordColorLengths() {
	M_Vec3f& coord = this->coord;
	M_Vec3f& normal = this->normal;
	M_Color3f& color = this->color;
	M_Int& coordIndex = this->coordIndex;
	M_Int& rc = this->rc;

	int oldlen = coord.length;

	coord.length = color.length;
	normal.length = color.length;


	for (int i = 0; i < rc.length; i++) {
		int ci = getIndex(rc[i]);
		coord[oldlen + i] = coord[ci];
		normal[oldlen + i] = normal[ci];
	}


	rc.length = 0;
}

void CNCShader::convertToRenderFormat(Mesh& mesh) {
	const M_Vec3f& meshCoord = mesh.coord;
	const M_Vec3f& meshNormal = mesh.normal;
	const M_Color3f& meshColor = mesh.color;
	const M_Int& meshCoordIndex = mesh.coordIndex;
	const M_Int& meshRc = mesh.rc;

	if (mesh.rc.length) {
		mesh.matchCoordColorLengths();
	}

	int meshCoordLen = meshCoord.length;
	vertexLen = meshColor.length;
	coordIndexLen = meshCoordIndex.length;



	coord = reinterpret_cast<const float*>(meshCoord.getArray());
	normal = reinterpret_cast<const float*>(meshNormal.getArray());
	color = reinterpret_cast<const float*>(meshColor.getArray());

	coordIndex = reinterpret_cast<const RenderInt*>(meshCoordIndex.getArray());



	if (firstTime) {
		setupShaders();

		initMesh();
	}

	else {
		glUseProgram(program);


		glBindVertexArray(vertexArrayObject[0]);
		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, coord, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[1]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, normal, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[2]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, color, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[3]);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);

		//glEnableVertexAttribArray(coordLoc);
		//glVertexAttribPointer(coordLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);
		//glBindVertexArray(0);
	}
	

	firstTime = false;
}

void CNTShader::convertToRenderFormat(Mesh& mesh) {
	const M_Vec3f& meshCoord = mesh.coord;
	const M_Vec3f& meshNormal = mesh.normal;
	const M_Vec2f& meshTexCoord = mesh.texCoord;
	const M_Int& meshCoordIndex = mesh.coordIndex;
	const M_Int& meshRc = mesh.rc;

	int meshCoordLen = meshCoord.length;
	vertexLen = meshTexCoord.length;
	coordIndexLen = meshCoordIndex.length;

	if (mesh.rc.length) {
		mesh.matchCoordColorLengths();
	}


	coord = reinterpret_cast<const float*>(meshCoord.getArray());
	normal = reinterpret_cast<const float*>(meshNormal.getArray());
	texCoord = reinterpret_cast<const float*>(meshTexCoord.getArray());

	coordIndex = reinterpret_cast<const RenderInt*>(meshCoordIndex.getArray());



	if (firstTime) {
		setupShaders();

		initMesh();
	}

	else {
		glUseProgram(program);


		glBindVertexArray(vertexArrayObject[0]);
		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, coord, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[1]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, normal, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[2]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec2f) * vertexLen, texCoord, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[3]);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);

		//glEnableVertexAttribArray(coordLoc);
		//glVertexAttribPointer(coordLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);
		//glBindVertexArray(0);
	}


	firstTime = false;
}

void CNC0WeightBoneShader::convertToRenderFormat(Mesh& mesh) {
	const M_Vec3f& meshCoord = mesh.coord;
	const M_Vec3f& meshNormal = mesh.normal;
	const M_Color3f& meshColor = mesh.color;
	const M_Int& meshCoordIndex = mesh.coordIndex;
	const M_Int& meshRc = mesh.rc;

	int meshCoordLen = meshCoord.length;
	vertexLen = meshColor.length;
	coordIndexLen = meshCoordIndex.length;

	if (mesh.rc.length) {
		mesh.matchCoordColorLengths();
	}


	coord = reinterpret_cast<const float*>(meshCoord.getArray());
	normal = reinterpret_cast<const float*>(meshNormal.getArray());
	color = reinterpret_cast<const float*>(meshColor.getArray());

	coordIndex = reinterpret_cast<const RenderInt*>(meshCoordIndex.getArray());



	if (firstTime) {
		setupShaders();

		initMesh();
	}

	else {
		glUseProgram(program);


		glBindVertexArray(vertexArrayObject[0]);
		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, coord, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[1]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, normal, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[2]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, color, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[3]);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);

		//glEnableVertexAttribArray(coordLoc);
		//glVertexAttribPointer(coordLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);
		//glBindVertexArray(0);
	}


	firstTime = false;
}

void CNT0WeightBoneShader::convertToRenderFormat(Mesh& mesh) {
	const M_Vec3f& meshCoord = mesh.coord;
	const M_Vec3f& meshNormal = mesh.normal;
	const M_Vec2f& meshTexCoord = mesh.texCoord;
	const M_Int& meshCoordIndex = mesh.coordIndex;
	const M_Int& meshRc = mesh.rc;

	int meshCoordLen = meshCoord.length;
	vertexLen = meshTexCoord.length;
	coordIndexLen = meshCoordIndex.length;

	if (mesh.rc.length) {
		mesh.matchCoordColorLengths();
	}


	coord = reinterpret_cast<const float*>(meshCoord.getArray());
	normal = reinterpret_cast<const float*>(meshNormal.getArray());
	texCoord = reinterpret_cast<const float*>(meshTexCoord.getArray());

	coordIndex = reinterpret_cast<const RenderInt*>(meshCoordIndex.getArray());



	if (firstTime) {
		setupShaders();

		initMesh();
	}

	else {
		glUseProgram(program);


		glBindVertexArray(vertexArrayObject[0]);
		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, coord, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[1]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * vertexLen, normal, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[2]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vec2f) * vertexLen, texCoord, GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[3]);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(RenderInt) * coordIndexLen, coordIndex, GL_DYNAMIC_DRAW);

		//glEnableVertexAttribArray(coordLoc);
		//glVertexAttribPointer(coordLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);
		//glBindVertexArray(0);
	}


	firstTime = false;
}

void Mesh::renderNode(Shader* shader, Matrix parentMatrix, Matrix catRotMatrix) {

	if(dynamic_cast<CNC0WeightBoneShader*>(shader)){
		int a = 0;
	}

	if (shader == 0) {

		if (normal.length == coord.length) {
			if (color.length == coord.length) {
				CNCShader* cnc = new CNCShader;
				shader = cnc;
				cnc->convertToRenderFormat(*this);
			}
			else if (texCoord.length == coord.length) {
				CNTShader* cnt = new CNTShader;
				shader = cnt;
				cnt->convertToRenderFormat(*this);
			}
			else throw;
		}
		else if (color.length == 1) {
			CMonoCLineShader* cmc = new CMonoCLineShader;
			shader = cmc;
			cmc->f_color = color[0];
			cmc->convertToRenderFormat(*this);
		}
		else{
			CCLineShader* cc = new CCLineShader;
			shader = cc;
			cc->convertToRenderFormat(*this);

		}

	}
	shader->viewMatrix = parentMatrix;
	shader->rotationMat = catRotMatrix;
	//renderNode(parentMatrix, parentColScale, render, 0);
	shader->render();
}

void GeomRender::renderNode(Matrix parentMatrix, Matrix catRotMatrix){
	Mesh::renderNode(shader, parentMatrix, catRotMatrix, appearance);
}

void Line::renderNode(Matrix parentMatrix, Matrix catRotMatrix) {
	if (shader == 0) {

		CCLineShader* cc = new CCLineShader;
		shader = cc;

		cc->convertToRenderFormat(*this);

	}
	//renderNode(parentMatrix, parentColScale, render, 0);
	shader->render();
}

void IndexedFaceSet::renderNode(Matrix parentMatrix, Matrix catRotMatrix, GeomRender& mesh) {
	Appearance* appearance = mesh.appearance;
	if (coordIndex.refValue().length == 0)return;
	Coordinate* crd = dynamic_cast<Coordinate*>(coord.p);
	Normal* nrm = dynamic_cast<Normal*>(normal.p);
	Color* color = dynamic_cast<Color*>(this->color.p);
	TextureCoordinate* txc = dynamic_cast<TextureCoordinate*>(texCoord.p);
	Material* mat = 0;
	if (appearance) {
		mat = dynamic_cast<Material*>(appearance->material.p);
	}

	if (crd == 0) {
		return;
	}
	if (crd->renderProcessed == false) {
		renderProcessedQue += crd;
		goto here;
	}
	else if (nrm && nrm->renderProcessed == false) {
		renderProcessedQue += nrm;
		goto here;
	}
	else if (color && color->renderProcessed == false) {
		renderProcessedQue += color;
		goto here;
	}
	else if (color && mat && mat->renderProcessed == false) {
		renderProcessedQue += mat;
		goto here;
	}
	else if (txc && txc->renderProcessed == false) {
		renderProcessedQue += txc;
		goto here;
	}
	else if (renderProcessed == false) {
	here:
		renderProcessedQue += this;
		try {
			convertToGeomSource(mesh, g_tmpIndex);
		}
		catch (...) {
			return;
		}
		crd->renderProcessed = true;
		nrm->renderProcessed = true;
		color->renderProcessed = true;
		mat->renderProcessed = true;
		renderProcessed = true;
	}
	if (appearance) {
		TextureTransform* textureTransform = dynamic_cast<TextureTransform*>(appearance->textureTransform.p);
		if (textureTransform && textureTransform->renderProcessed == false) {
			throw;
			//mesh.transformRenderTexture(render, textureTransform);
			renderProcessedQue += textureTransform;
		}
	}
	mesh.renderNode(parentMatrix, catRotMatrix);
}


void IndexedTriangleSet::renderNode(Matrix parentMatrix, Matrix catRotMatrix, GeomRender& mesh) {
	if (index.refValue().length == 0)return;
	Coordinate* crd = dynamic_cast<Coordinate*>(coord.p);
	Normal* nrm = dynamic_cast<Normal*>(normal.p);
	Color* color = dynamic_cast<Color*>(this->color.p);
	TextureCoordinate* txc = dynamic_cast<TextureCoordinate*>(texCoord.p);
	Material* mat = 0;
	Appearance* appearance = mesh.appearance;
	if (appearance) {
		mat = dynamic_cast<Material*>(appearance->material.p);
	}

	if (crd == 0) {
		return;
	}
	if (crd->renderProcessed == false) {
		renderProcessedQue += crd;
		goto here;
	}
	else if (nrm && nrm->renderProcessed == false) {
		renderProcessedQue += nrm;
		goto here;
	}
	else if (color && color->renderProcessed == false) {
		renderProcessedQue += color;
		goto here;
	}
	else if (color && mat && mat->renderProcessed == false) {
		renderProcessedQue += mat;
		goto here;
	}
	else if (txc && txc->renderProcessed == false) {
		renderProcessedQue += txc;
		goto here;
	}
	else if (renderProcessed == false) {
	here:
		renderProcessedQue += this;
		try {
			convertToGeomSource(mesh, g_tmpIndex);
		}
		catch (...) {
			return;
		}
		crd->renderProcessed = true;
		nrm->renderProcessed = true;
		color->renderProcessed = true;
		mat->renderProcessed = true;
		renderProcessed = true;
	}
	if (appearance) {
		TextureTransform* textureTransform = dynamic_cast<TextureTransform*>(appearance->textureTransform.p);
		if (textureTransform && textureTransform->renderProcessed == false) {
			throw;
			//mesh.transformRenderTexture(render, textureTransform);
			renderProcessedQue += textureTransform;
		}
	}
	mesh.renderNode(parentMatrix, catRotMatrix);
}


void Shape::renderNode(Matrix parentMatrix, Matrix catRotMatrix) {

	if (Appearance* appearance = dynamic_cast<Appearance*>(this->appearance.p)) {
		Material* material = dynamic_cast<Material*>(appearance->material.p);
		if (ImageTexture* texture = dynamic_cast<ImageTexture*>(appearance->texture.p)) {

			if (g_renderLater == false) {
				if (texture->alphaTexture) {
					g_alphaChanList += AlphaChanList(this, parentMatrix, catRotMatrix);
					return;
				}
				if (material && material->transparency > 0) {
					g_transparentList += TransparencyList(this, parentMatrix, catRotMatrix);
					return;
				}
			}

			if (texture->update) {
				const M_String& url = texture->url;
				if (url.length > 0) {
					if(1){
						throw;
						//(FAILED(D3DXCreateTextureFromFile(g_pd3dDevice, url[0].c_str(), &texture->dxTexture))) {
						//dLog (String("failed to set Texture ")+texture->url.refValue()[0]);
					}
					else {
						dLog("set Texture from local file");
						throw;
						/*
						texture->dxTexture->GetSurfaceLevel(0, &g_textureSurface);
						g_textureSurface->GetDesc(&g_surfaceBuffer);
						texture->alphaTexture = false;
						if (g_surfaceBuffer.Format == D3DFMT_A8R8G8B8) {
							texture->alphaTexture = true;
						}
						texture->update = false;
						*/
					}
				}
			}
		}
		else {
			if (g_renderLater == false) {
				if (material && material->transparency > 0) {
					g_transparentList += TransparencyList(this, parentMatrix, catRotMatrix);
					return;
				}
			}
		}
	}
	else {
		return;
	}
	if (Geometry* geometry = dynamic_cast<Geometry*>(this->geometry.getRawPtr())) {
		mesh.appearance = appearance;
		geometry->renderNode(parentMatrix, catRotMatrix, mesh);
	}
}

void CNCShader::dumpMesh() {
	String str;
	str = "vertexLen = ";
	str += vertexLen;
	dLog(str);
	str = "coord: ";
	for (int i = 0; i < vertexLen * 3;) {
		str += '{';
		str += coord[i++];
		str += ' ';
		str += coord[i++];
		str += ' ';
		str += coord[i++];
		str += "} ";
	}
	dLog(str);

	str = "normal: ";
	for (int i = 0; i < vertexLen * 3;) {
		str += '{';
		str += normal[i++];
		str += ' ';
		str += normal[i++];
		str += ' ';
		str += normal[i++];
		str += "} ";
	}
	dLog(str);

	str = "color: ";
	for (int i = 0; i < vertexLen * 3;) {
		str += '{';
		str += color[i++];
		str += ' ';
		str += color[i++];
		str += ' ';
		str += color[i++];
		str += "} ";
	}
	dLog(str);


	str = "coordIndexLen = ";
	str += coordIndexLen;
	dLog(str);
	str = "coordIndexLen: ";
	for (int i = 0; i < coordIndexLen;) {
		str += coordIndex[i++];
		str += ' ';
	}
	dLog(str);
}

void Seamless::convertToRenderFormat() {
	if (g_bReady == false)return;
	if (partList.length == 0)return;
	if (partList[0]->workEffect == 0)return;

	if (partList.length == 0 || effect == 0) {
		if (mesh.shader != 0) {
			mesh.shader = 0;
		}
		return;
	}
	if (mesh.shader == 0) {
		if (effect.refValue()->getType() == COLOREFFECT) {
			if (animCP) {
				mesh.shader = new CNCShader;
			}
			else {
				mesh.shader = new CNC0WeightBoneShader;
			}
		}
		else if (effect.refValue()->getType() == TEXTUREEFFECT) {
			if (animCP) {
				mesh.shader = new CNTShader;
			}
			else {
				mesh.shader = new CNT0WeightBoneShader;
			}
		}
		else {
			throw;
		}
	}



	setUpInternalEffectPtrs2();
	setUpInternalEffectPtrs();

	if (workAnimCP == false && g_colorOwnership) {
		if (this != g_vcTransferBCyl) {
			convertToOwnershipAidRenderFormat();
			return;
		}
	}

	if (geometry == 0) {
		geometry = new IndexedFaceSet;
	}
	IndexedFaceSet* ifs = geometry.getRawPtr();
	ifs->coord = new Coordinate;
	ifs->normal = new Normal;
	if (effect.refValue()->getType() == COLOREFFECT || effect.refValue()->getType() == MONOCOLOREFFECT) {
		ifs->color = new Color;
	}
	else {
		ifs->texCoord = new TextureCoordinate;
	}
	int coordLen = 0;

	M_Int& coordIndex = ifs->coordIndex.refValue();

	catTranslationMatrix();
	ifs->genTransformedNormals(partList);
	catMatrix();


	addExtraCoordsRender(coordIndex);

	int indexCount = coordIndex.length / 4 * 3;

	if (!indexCount) {
		return;
	}

	{
		M_Int& meshCoordIndex = mesh.coordIndex.refValue();
		meshCoordIndex.length = indexCount;

		for (int k = 0, fromPoly = 0; k < partList.length; k++) {
			Art* part = partList[k];
			part->fromPoly = fromPoly;
			fromPoly += part->triangle.length;
			part->toPoly = fromPoly;
		}
		int e = 0;
		int j = 0;
		if (g_hidePolysInHiddenPatches) {
			for (; e < effectList.length; e++) {
				Effect* effect = effectList[e];
				M_Art& partList = effect->partList;
				int newTriLen = 0;
				for (int k = 0; k < partList.length; k++) {
					Art* part = partList[k];
					coordLen += part->coord.length;
					int triLen = part->triangle.length;
					if (part->show) {
						for (int i = triLen, t = part->toPoly * 4; --i >= 0; ) {
							S_Triangle& tri = part->triangle[i];
							int ia = tri.ia;
							int ib = tri.ib;
							int ic = tri.ic;

							Art* pa = tri.pa;
							Art* pb = tri.pb;
							Art* pc = tri.pc;

							NurbsPatch* patchA = pa->patchOwner[ia];
							NurbsPatch* patchB = pa->patchOwner[ib];
							NurbsPatch* patchC = pc->patchOwner[ic];

							if (patchA->cage || patchB->cage || patchC->cage) {
								meshCoordIndex[j++] = coordIndex[t - 4];
								meshCoordIndex[j++] = coordIndex[t - 3];
								meshCoordIndex[j++] = coordIndex[t - 2];
								newTriLen++;
							}
							t -= 4;
						}
					}
				}
				effect->triLen = newTriLen;
			}
			meshCoordIndex.length = j;
		}
		else {
			for (int t = 0, h = 0, k = 0; k < partList.length; k++) {
				Art* part = partList[k];
				coordLen += part->coord.length;
				int triLen = part->triangle.length;
				if (part->show) {
					for (int i = triLen; --i >= 0; ) {

						meshCoordIndex[j++] = coordIndex[t++];
						meshCoordIndex[j++] = coordIndex[t++];
						meshCoordIndex[j++] = coordIndex[t++];
						t++;
					}
				}
				else {
					t += triLen * 4;
				}
			}
		}

	}
	if (effect.refValue()->getType() == COLOREFFECT) {
		// Create the vertex buffer.

		M_Vec3f& meshCoord = mesh.coord.refValue();
		M_Vec3f& meshNormal = mesh.normal.refValue();
		M_Color3f& meshColor = mesh.color.refValue();
		meshCoord.length = coordLen;
		meshNormal.length = coordLen;
		meshColor.length = coordLen;

		{
			setUpInternalEffectPtrsTmpForRenderingTransparancy(this);

			for (int j = 0, smi = 0, k = 0; k < partList.length; k++) {

				Art* part = partList[k];

				part->renderFromCoord = j;
				float alpha = 1.0f - reinterpret_cast<ColorEffect*>(part->workEffect)->transparency;

				const M_Vec3f& coord = part->coord;
				const M_Vec3f& normal = part->normal;
				const M_Color3f& color = part->color;
				const M_Int& coordGroup = part->coordGroup;
				const M_Int& matrixIndex = part->matrixIndex;
				Vec3f absolutePivotPoint;
				if (workAnimCP == false) {
					absolutePivotPoint = part->absolutePivotPoint;
				}

				for (int i = 0; i < part->coord.length; i++, j++) {
					meshCoord[j] = coord[i];
					meshNormal[j] = normal[i];
					meshColor[j] = color[i];
				}
				smi++;
			}
			setUpInternalEffectPtrs();
		}
	}
	else if (effect.refValue()->getType() == TEXTUREEFFECT) {
		// Create the vertex buffer.

		M_Vec3f& meshCoord = mesh.coord.refValue();
		M_Vec3f& meshNormal = mesh.normal.refValue();
		M_Vec2f& meshTexCoord = mesh.texCoord.refValue();
		meshCoord.length = coordLen;
		meshNormal.length = coordLen;
		meshTexCoord.length = coordLen;

		{
			setUpInternalEffectPtrsTmpForRenderingTransparancy(this);

			for (int j = 0, smi = 0, k = 0; k < partList.length; k++) {

				Art* part = partList[k];

				part->renderFromCoord = j;
				float alpha = 1.0f - reinterpret_cast<ColorEffect*>(part->workEffect)->transparency;

				const M_Vec3f& coord = part->coord;
				const M_Vec3f& normal = part->normal;
				const M_Vec2f& texCoord = part->texCoord;
				const M_Int& coordGroup = part->coordGroup;
				const M_Int& matrixIndex = part->matrixIndex;
				Vec3f absolutePivotPoint;
				if (workAnimCP == false) {
					absolutePivotPoint = part->absolutePivotPoint;
				}

				for (int i = 0; i < part->coord.length; i++, j++) {
					meshCoord[j] = coord[i];
					meshNormal[j] = normal[i];
					meshTexCoord[j] = texCoord[i];
				}
				smi++;
			}
			setUpInternalEffectPtrs();
		}
	}
	if (CNC0WeightBoneShader* shader = dynamic_cast<CNC0WeightBoneShader*>(mesh.shader)) {
		M_Vec3f& meshCoord = mesh.coord;
		const M_Vec3f& meshNormal = mesh.normal;
		const M_Color3f& meshColor = mesh.color;
		const M_Int& meshCoordIndex = mesh.coordIndex;
		const M_Int& meshRc = mesh.rc;

		int meshCoordLen = meshCoord.length;
		shader->vertexLen = meshColor.length;
		shader->coordIndexLen = meshCoordIndex.length;

		if (mesh.rc.length) {
			mesh.matchCoordColorLengths();
		}


		shader->coord = reinterpret_cast<const float*>(meshCoord.getArray());
		shader->normal = reinterpret_cast<const float*>(meshNormal.getArray());
		shader->color = reinterpret_cast<const float*>(meshColor.getArray());

		shader->coordIndex = reinterpret_cast<const RenderInt*>(meshCoordIndex.getArray());

		M_Int& boneIndex = shader->boneIndex;
		boneIndex.length = meshCoord.length;


		setUpInternalEffectPtrsTmpForRenderingTransparancy(this);

		int smi = 0;
		for (int j = 0, k = 0; k < partList.length; k++) {

			Art* part = partList[k];

			part->renderFromCoord = j;
			float alpha = 1.0f - reinterpret_cast<ColorEffect*>(part->workEffect)->transparency;

			const M_Vec3f& coord = part->coord;
			const M_Vec3f& normal = part->normal;
			const M_Color3f& color = part->color;
			const M_Int& coordGroup = part->coordGroup;
			const M_Int& matrixIndex = part->matrixIndex;
			Vec3f absolutePivotPoint;
			if (workAnimCP == false) {
				absolutePivotPoint = part->absolutePivotPoint;
			}

			int i = 0;

			for (int cg = 0; cg < coordGroup.length; cg++) {
				for (int c = coordGroup[cg]; --c >= 0 && i < part->coord.length; i++, j++) {
					meshCoord[j] = (coord[i] - absolutePivotPoint);
					boneIndex[j] = matrixIndex[i];
				}
				smi++;
			}
			for (; i < part->coord.length; i++, j++) {
				meshCoord[j] = (coord[i] - absolutePivotPoint);
				boneIndex[j] = matrixIndex[i];
			}
			smi++;
		}
		M_Matrix& boneMatrix = shader->boneMatrix;
		M_Matrix& rotBoneMatrix = shader->rotBoneMatrix;
		boneMatrix.length = smi;
		rotBoneMatrix.length = smi;


		setUpInternalEffectPtrs();


		shader->setupShaders();
		shader->initMesh();

	}
	else if (CNT0WeightBoneShader* shader = dynamic_cast<CNT0WeightBoneShader*>(mesh.shader)) {
		mesh.color.refValue().length = mesh.coord.length;
		M_Vec3f& meshCoord = mesh.coord;
		const M_Vec3f& meshNormal = mesh.normal;
		const M_Vec2f& meshTexCoord = mesh.texCoord;
		const M_Int& meshCoordIndex = mesh.coordIndex;
		const M_Int& meshRc = mesh.rc;

		int meshCoordLen = meshCoord.length;
		shader->vertexLen = meshTexCoord.length;
		shader->coordIndexLen = meshCoordIndex.length;

		if (mesh.rc.length) {
			mesh.matchCoordColorLengths();
		}


		shader->coord = reinterpret_cast<const float*>(meshCoord.getArray());
		shader->normal = reinterpret_cast<const float*>(meshNormal.getArray());
		shader->texCoord = reinterpret_cast<const float*>(meshTexCoord.getArray());

		shader->coordIndex = reinterpret_cast<const RenderInt*>(meshCoordIndex.getArray());

		M_Int& boneIndex = shader->boneIndex;
		boneIndex.length = meshCoord.length;


		setUpInternalEffectPtrsTmpForRenderingTransparancy(this);

		int smi = 0;
		for (int j = 0, k = 0; k < partList.length; k++) {

			Art* part = partList[k];

			part->renderFromCoord = j;
			float alpha = 1.0f - reinterpret_cast<ColorEffect*>(part->workEffect)->transparency;

			const M_Vec3f& coord = part->coord;
			const M_Vec3f& normal = part->normal;
			const M_Color3f& color = part->color;
			const M_Int& coordGroup = part->coordGroup;
			const M_Int& matrixIndex = part->matrixIndex;
			Vec3f absolutePivotPoint;
			if (workAnimCP == false) {
				absolutePivotPoint = part->absolutePivotPoint;
			}

			int i = 0;

			for (int cg = 0; cg < coordGroup.length; cg++) {
				for (int c = coordGroup[cg]; --c >= 0 && i < part->coord.length; i++, j++) {
					meshCoord[j] = (coord[i] - absolutePivotPoint);
					boneIndex[j] = matrixIndex[i];
				}
				smi++;
			}
			for (; i < part->coord.length; i++, j++) {
				meshCoord[j] = (coord[i] - absolutePivotPoint);
				boneIndex[j] = matrixIndex[i];
			}
			smi++;
		}
		M_Matrix& boneMatrix = shader->boneMatrix;
		M_Matrix& rotBoneMatrix = shader->rotBoneMatrix;
		boneMatrix.length = smi;
		rotBoneMatrix.length = smi;


		setUpInternalEffectPtrs();


		shader->setupShaders();
		shader->initMesh();

	}
	else {
		mesh.shader->convertToRenderFormat(mesh);
	}

	resetCoordLen();
	ctrf = false;
	setCoordFroms();
}

Matrix dxWindMatrix(Matrix& parentMatrix, Vec3f& scale, Rotation& rotation, Vec3f& translation) {
	return scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation) * parentMatrix;
}

void ColorEffect::setUpVertexShaders(Matrix& parentMatrix, Matrix&  catRotMatrix, Seamless& smls) {

	if (CNC0WeightBoneShader* boneShader = dynamic_cast<CNC0WeightBoneShader*>(smls.mesh.shader)) {
		M_SArt& skeleton = smls.skeleton.refValue();
		for (int i = 0, smi = 0; i < skeleton.length; i++) {
			skeleton[i]->genDynamicTransforms(parentMatrix, catRotMatrix, smi, boneShader->boneMatrix, boneShader->rotBoneMatrix);
		}
	}
}


void TextureEffect::setUpVertexShaders(Matrix& parentMatrix, Matrix&  catRotMatrix, Seamless& smls) {

	if (CNT0WeightBoneShader* boneShader = dynamic_cast<CNT0WeightBoneShader*>(smls.mesh.shader)) {
		M_SArt& skeleton = smls.skeleton.refValue();
		for (int i = 0, smi = 0; i < skeleton.length; i++) {
			skeleton[i]->genDynamicTransforms(parentMatrix, catRotMatrix, smi, boneShader->boneMatrix, boneShader->rotBoneMatrix);
		}
	}
}
void MonoColorEffect::setUpVertexShaders(Matrix& parentMatrix, Matrix&  catRotMatrix, Seamless& smls) {
}


void Seamless::renderNode(Matrix parentMatrix, Matrix catRotMatrix) {
	if (ignoreWireFrame == true && g_wireFrameMode) {
		glPolygonMode(GL_FRONT, GL_FILL);
	}

	if (ikProcessFlag && (g_scene->mode & (HOME_MODE | ANIMATION_MODE))) {
		if ((g_visualClue == g_partCageClue && g_partCageClue->smls == this) || (g_visualClue == g_vcArtTriBase && g_vcArtTriBase->smls == this)) {
		}
		else {
			calcIK();
		}
		ikProcessFlag = false;
	}

	if (animCP) {
		if (animCPBuildFlag) {
			doBuild();
		}
	}
	else {
		if (morphFlag) {
			int j = 0;
			if (cpMorpherPartLastLen != cpMorpherPart.length) {
				for (int i = cpMorpherPartLastLen; i < cpMorpherPart.length;) {
					cpMorpherPart[j++] = cpMorpherPart[i++];
				}

				cpMorpherPart.length = cpMorpherPartLastLen = j;
			}

			for (int k = cpMorpherPartLastLen; --k >= 0;) {
				CPMorpherPart& cpmp = cpMorpherPart[k];
				cpmp.morpher->do_polyFrame(cpmp.part);
			}
		}
	}

	if (partList.length == 0 || this->effect == 0)return;

	parentMatrix = dxWindMatrix(parentMatrix, scale, rotation, translation);
	catRotMatrix = rotationMatrix(rotation) * catRotMatrix;

	//TextureEffect* effect=dynamic_cast<TextureEffect*>(this->effect.getRawPtr());
	if (mesh.shader == 0 || ctrf) {
		convertToRenderFormat();
		ctrf = false;
		if (mesh.shader == 0) {
			return;
		}
	}

	if (workAnimCP) {
		M_SArt& skeleton = this->skeleton.refValue();
		for (int i = 0, smi = 0; i < skeleton.length; i++) {
			skeleton[i]->genDynamicTransformsCP(parentMatrix, parentMatrix, smi);
		}
	}
	else if (effect != 0) {
		dynamic_cast<Effect*>(effect.p)->setUpVertexShaders(parentMatrix, catRotMatrix, *this);
	}
	//g_pd3dDevice->SetRenderState(D3DRS_VERTEXBLEND, D3DVBF_0WEIGHTS);
	//g_pd3dDevice->SetRenderState(D3DRS_INDEXEDVERTEXBLENDENABLE, TRUE);

	if (false == (ignoreWireFrame == false && g_scene->mode & HIDE_MESH_MODE)) {
		for (int k = 0, triFrom = 0; k < effectList.length; k++) {
			Effect* effect = effectList[k];
			effect->renderMaterial(triFrom, *this, parentMatrix, catRotMatrix);
			triFrom += effect->triLen;
		}
	}
	if (ignoreWireFrame == true && g_wireFrameMode) {
		glPolygonMode(GL_FRONT, GL_LINE);
	}

	//g_pd3dDevice->SetRenderState(D3DRS_VERTEXBLEND, D3DVBF_DISABLE);
	//g_pd3dDevice->SetRenderState(D3DRS_INDEXEDVERTEXBLENDENABLE, FALSE);


	morphFlag = false;


//	if (mesh.coord.length) {
//		mesh.renderNode(parentMatrix, catRotMatrix);
//	}
/*
	if (workAnimCP == false || stickMan) {
		if (CNC0WeightBoneShader* boneShader = dynamic_cast<CNC0WeightBoneShader*>(mesh.shader)) {
			M_SArt& skeleton = this->skeleton.refValue();
			for (int i = 0, smi = 0; i < skeleton.length; i++) {
				skeleton[i]->genDynamicTransforms(parentMatrix, catRotMatrix, smi, boneShader->boneMatrix, boneShader->rotBoneMatrix);
			}
		}
		else if (CNT0WeightBoneShader* boneShader = dynamic_cast<CNT0WeightBoneShader*>(mesh.shader)) {
			M_SArt& skeleton = this->skeleton.refValue();
			for (int i = 0, smi = 0; i < skeleton.length; i++) {
				skeleton[i]->genDynamicTransforms(parentMatrix, catRotMatrix, smi, boneShader->boneMatrix, boneShader->rotBoneMatrix);
			}
		}
	}
*/
	if (stickMan) {
		g_skelList += RenderLast(this, matrix, catRotMatrix);
	}

	if (nurbsLineShape) {
		nurbsLineShape->renderNode(parentMatrix, catRotMatrix);
	}
}

void InitPerspectiveProj(Matrix& m){

	const float ar = float(g_3d->size.x) / float(g_3d->size.y);
	const float zNear = .5;
	const float zFar = 400;
	const float zRange = zNear - zFar;
	const float tanHalfFOV = tanf(g_scene->fieldOfView * .5);

	m.m[0][0] = 1.0f / (tanHalfFOV * ar);
	m.m[0][1] = 0.0f;
	m.m[0][2] = 0.0f;
	m.m[0][3] = 0.0f;

	m.m[1][0] = 0.0f;
	m.m[1][1] = 1.0f / tanHalfFOV;
	m.m[1][2] = 0.0f;
	m.m[1][3] = 0.0f;

	m.m[2][0] = 0.0f;
	m.m[2][1] = 0.0f;
	m.m[2][2] = (-zNear - zFar) / zRange;
	m.m[2][3] = 2.0f * zFar * zNear / zRange;

	m.m[3][0] = 0.0f;
	m.m[3][1] = 0.0f;
	m.m[3][2] = 1.0f;
	m.m[3][3] = 0.0f;
}

void buildProjectionMatrix(Matrix& projMatrix, float fov, float ratio, float nearP, float farP) {

	float f = 1.0f / tan(fov * .5);

	projMatrix.m[0][0] = f / ratio;
	projMatrix.m[1][1] = f;
	projMatrix.m[2][2] = (farP + nearP) / (nearP - farP);
	projMatrix.m[3][2] = (2.0f * farP * nearP) / (nearP - farP);
	projMatrix.m[2][3] = -1.0f;
	projMatrix.m[3][3] = 0.0f;
}

// res = a cross b;
void crossProduct(Vec3f& a, Vec3f& b, Vec3f& res) {

	res.x = a.y * b.z - b.y * a.z;
	res.y = a.z * b.x - b.z * a.x;
	res.z = a.x * b.y - b.x * a.y;
}

// Normalize a vec3
void normalize(Vec3f& a) {

	float mag = sqrt(a.x * a.x + a.y * a.y + a.z * a.z);

	a.x /= mag;
	a.y /= mag;
	a.z /= mag;
}

void setCamera(Matrix& viewMatrix, Vec3f pos, Vec3f lookAt) {

	Vec3f dir;

	Vec3f right;
	Vec3f up;

	//up = Vec3f(.12331, .805093, -.58019);
	up = Vec3f(0, 1, 0);

	dir = lookAt - pos;
	normalize(dir);

	crossProduct(dir, up, right);
	normalize(right);

	crossProduct(right, dir, up);
	normalize(up);


	viewMatrix.m[0][0] = right.x;
	viewMatrix.m[1][0] = right.y;
	viewMatrix.m[2][0] = right.z;
	viewMatrix.m[3][0] = 0.0f;

	viewMatrix.m[0][1] = up.x;
	viewMatrix.m[1][1] = up.y;
	viewMatrix.m[2][1] = up.z;
	viewMatrix.m[3][1] = 0.0f;

	viewMatrix.m[0][2] = -dir.x;
	viewMatrix.m[1][2] = -dir.y;
	viewMatrix.m[2][2] = -dir.z;
	viewMatrix.m[3][2] = 0.0f;

	viewMatrix.m[0][3] = 0.0f;
	viewMatrix.m[1][3] = 0.0f;
	viewMatrix.m[2][3] = 0.0f;
	viewMatrix.m[3][3] = 1.0f;

	Matrix aux = translationMatrix(Vec3f(-pos.x, -pos.y, -pos.z));

	viewMatrix = aux * viewMatrix;
}

void render() {
	bool fogExist = g_fogExist;
	g_fogExist = false;

	if (g_minimized) {
		return;
	}
	if (g_windowSizeWindowBool && g_windowSizeWindowTime + 1 < g_time) {
		windowSizeHasChanged();
		g_windowSizeWindowBool = false;
	}

	renderProcessedQue.length = 0;


#ifdef ROBOT_SYNC
	InputEvent::runRobotSequence();
#endif

	renderProcessedQue.length = 0;

#ifdef FRAME_TIME_SYNC
	g_time += FRAME_TIME_SYNC;
#else
	g_time = timeGetTime() * MILISECONDS_TO_SECONDS;
#endif
	processPackets();

	if (g_winAPI3DMode) {
		//POINT ptCursor;
		//GetCursorPos(&ptCursor);
		//ScreenToClient( hWndMain, &ptCursor);
		//g_mousePos.x = ptCursor.x;
		//g_mousePos.y = ptCursor.y;
		//dLog("moveMouse3d");
		//g_mousePos.x++;
		//dLog("g_mousePos.x");
		moveMouse3d();
}

	if (g_runAnim) {
		for (int i = 0; i < g_animBarList.length; i++) {
			g_animBarList[i]->renderNodeFirst();
		}
	}

	Matrix viewMatrix = g_scene->getViewMatrix();
	if (g_perspective) {
		buildProjectionMatrix(projMatrix, g_scene->fieldOfView, float(g_3d->size.x) / float(g_3d->size.y), 0.5f, 400.0f);
		viewMatrix = viewMatrix * translationMatrix(Vec3f(0, 0, -g_viewDist));
	}
	else {
		projMatrix = Matrix();
		float aspect = float(g_3d->size.x) / float(g_3d->size.y);
		float left = -aspect;
		float right = aspect;
		float bottom = -1;
		float top = 1;
		float nearPoint = -3;
		float farPoint = 100;

		projMatrix.m[0][0] = 2.0f / (right - left);
		projMatrix.m[1][1] = 2.0f / (top - bottom);
		projMatrix.m[2][2] = -2.0f / (farPoint - nearPoint);
		projMatrix.m[3][0] = -(right + left) / (right - left);
		projMatrix.m[3][1] = -(top + bottom) / (top - bottom);
		projMatrix.m[3][2] = -(farPoint + nearPoint) / (farPoint - nearPoint);

	}

/*

	Matrix viewMatrix =  g_scene->getViewMatrix() * translationMatrix(Vec3f(0, 0, -g_viewDist));
	Matrix viewMatrix = g_scene->getViewMatrix();
	//Matrix viewMatrix;
	//setCamera(viewMatrix, Vec3f(-1.165654, 1.731431, 2.154855), Vec3f(0, 0, 0));
	//setCamera(viewMatrix, Vec3f(0, 0, 3), Vec3f(0, 0, 0));

	//viewMatrix = g_scene->getViewMatrix() * viewMatrix;

	Matrix catRotMatrix = rotationMatrix(g_scene->rotation);
	*/

	//Matrix viewMatrix = g_scene->getViewMatrix();
	Matrix catRotMatrix = rotationMatrix(g_scene->rotation);


	if (Grid* grid = dynamic_cast<Grid*>(g_scene->grid.p)) {
		if (grid->show && g_gridShape) {
			g_gridShape->renderNode(viewMatrix, catRotMatrix);
		}
	}

	M_Node& children = g_scene->children;
	{
		for (int i = 0; i < children.length; i++) {
			Node* node = (children[i].p);

			node->renderNode(viewMatrix, catRotMatrix);
			
		}
	}
	if (g_visualClue) {
		{
			if (g_scene->mode.refValue() & ZBUFFER_CAGE_MODE) {
				if (g_curvedCageMode) {
					viewMatrix = viewMatrix * translationMatrix(Vec3f(0, 0, .005));
				}
				else {
					viewMatrix = viewMatrix * translationMatrix(Vec3f(0, 0, .025));
				}
				g_visualClue->renderNode(viewMatrix, catRotMatrix);
			}
			else {
				//viewMatrix = viewMatrix * translationMatrix(Vec3f(0,0,.005));
				glDisable(GL_DEPTH_TEST);
				g_visualClue->renderNode(viewMatrix, catRotMatrix);
				//g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE);
				//glDisable(GL_DEPTH_TEST);
			}
		}
	}
	if (MultiSelect::withinSights && g_dragActuallyHappened == false) {
		if (false == (g_scene->mode & SELECT_OBJECT_MODE && g_transformObj->mode == TransformObj::SELECT)) {
			g_cursorOverPointIlsBase->renderNode(Matrix(), catRotMatrix);
		}
	}
	if (g_transparentList.length) {
		glEnable(GL_DEPTH_TEST);
		g_renderLater = true;
		for (int k = g_transparentList.length; --k >= 0;) {
			TransparencyList& rl = g_transparentList[k];
			rl.node->renderNode(rl.matrix, rl.catRotMatrix);
		}
		g_renderLater = false;
		g_transparentList.length = 0;
		glDisable(GL_DEPTH_TEST);
	}
	glDisable(GL_DEPTH_TEST);
	if (g_alphaChanList.length) {
		//g_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
		//g_pd3dDevice->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTA_TEXTURE);
		g_renderLater = true;
		for (int k = g_alphaChanList.length; --k >= 0;) {
			AlphaChanList& rl = g_alphaChanList[k];
			rl.node->renderNode(rl.matrix, rl.catRotMatrix);
		}
		g_renderLater = false;
		g_alphaChanList.length = 0;
		//g_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
		//g_pd3dDevice->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_DISABLE);
	}

	if (g_skelList.length) {
		for (int k = g_skelList.length; --k >= 0;) {
			RenderLast& rl = g_skelList[k];
			rl.smls->stickMan->renderStickMan(rl.smls, rl.matrix, rl.catRotMatrix);
		}
		g_skelList.length = 0;
	}
	{
		for (int i = renderProcessedQue.length; --i >= 0;) {
			renderProcessedQue[i]->renderProcessed = true;
		}
	}

	glDisable(GL_DEPTH_TEST);

	if (g_wireFrameMode) {
		glPolygonMode(GL_FRONT, GL_FILL);
		Window::renderVisible();
		glPolygonMode(GL_FRONT, GL_LINE);
	}
	else {
		Window::renderVisible();
	}

	for (int i = 0; i < g_callBackList.length; i++) {
		g_callBackList[i]->doCallBack();
	}

	if (g_acp && g_acp->visible) {
		Video* video = dynamic_cast<Video*>(g_acp->anim->video.p);

		if (false == (video && video->audioTimeMode)) {
			g_acp->paint();
		}
	}

	glEnable(GL_DEPTH_TEST);


	InputEvent::renderCounter++;
	setMathHiPresision();
	InputEvent::runRobotSequence();
	if (g_waitMiliseconds > 0) {
		Sleep(g_waitMiliseconds);
	}




	//get2dMousePosFrom3d();
	//g_time+=1.0/12.0;
#ifdef BIND_DEVELOPER_SCRIPT
	scriptBindAnimation();
#endif
}


void Geometry::renderNode(Matrix parentMatrix, Matrix catRotMatrix, GeomRender& mesh) {//should not exist, should be virtual in future
}

void IndexedLineSet::convertToRenderFormat(GeomRender& mesh, const Appearance* appearance) {
	Coordinate*	coord = dynamic_cast<Coordinate*>(this->coord.p);
	mesh.coord = coord->point;

	if (mesh.shader == nullptr) {
		if (color) {
			mesh.shader = new CCLineShader;
		}
		else {
			mesh.shader = new CMonoCLineShader;
		}
	}


	if (Color*	color = dynamic_cast<Color*>(this->color.p)) {
		mesh.color = color->color;
	}
	else if (appearance) {
		if (Material* mat = dynamic_cast<Material*>(appearance->material.p)) {
			if (CMonoCLineShader * cmc = dynamic_cast<CMonoCLineShader*>(mesh.shader)) {
				cmc->f_color = mat->diffuseColor.refValue();
			}
		}
	}


	indexCount = 0;
	for (int i = 0, j = 1; j < coordIndex.length; ) {
		if (coordIndex[j] == -1 && ++j < coordIndex.length) {
			j++;
		}
		indexCount += 2;
		j++;
	}


	M_Int& meshCoordIndex = mesh.coordIndex.refValue();

	meshCoordIndex.length = indexCount;

	if(coordIndex.length){
		for (int i = 0, j = 1, q = coordIndex[0]; j < coordIndex.length; ) {
			if (coordIndex[j] == -1) {
				if (++j == coordIndex.length) {
					break;
				}
				q = coordIndex[j++];
			}
			meshCoordIndex[i++] = q;
			meshCoordIndex[i++] = q = coordIndex[j++];
		}
	}
	if (CMonoCLineShader * cmc = dynamic_cast<CMonoCLineShader*>(mesh.shader)) {
		cmc->convertToRenderFormat(mesh);
	}
	else if (CCLineShader* cc = dynamic_cast<CCLineShader*>(mesh.shader)) {
		cc->convertToRenderFormat(mesh);
	}
}

void IndexedLineSet::renderNode(Matrix parentMatrix, Matrix catRotMatrix, GeomRender& mesh) {
	Coordinate* coord = dynamic_cast<Coordinate*>(this->coord.p);
	if (coord == 0)return;
	M_Vec3f& point = coord->point;
	if (point.length == 0)return;
	Color* color = dynamic_cast<Color*>(this->color.p);

	if (coord == 0 || point.length <= 1) {
		return;
	}
	if (coord->renderProcessed == false) {
		renderProcessedQue += coord;
		goto here;
	}
	else if (color != 0 && color->renderProcessed == false) {
		renderProcessedQue += color;
		color->renderProcessed = true;
		goto here;
	}
	else if (renderProcessed == false) {
	here:
		renderProcessedQue += this;
		try {
			this->convertToRenderFormat(mesh, mesh.appearance);
		}
		catch (...) {
			return;
		}
		coord->renderProcessed = true;
		
		renderProcessed = true;
	}
	if (noZbuf && (g_scene->mode.refValue() & ZBUFFER_CAGE_MODE)) {
		glDisable(GL_DEPTH_TEST);
		mesh.renderNode(parentMatrix, catRotMatrix);
		glEnable(GL_DEPTH_TEST);
	}

	else {
		mesh.renderNode(parentMatrix, catRotMatrix);
	}
}

void Polyline2D::renderNode(Matrix parentMatrix, Matrix catRotMatrix, GeomRender& mesh) {
}


void Window::releaseWindow() {
	if (this) {
		if (shader) {
			shader->deleteProgram();
			shader = 0;
		}
	}
}


void WinShader::updatePosSize(Vec2i pos, Vec2i size) {
	setCoord(pos, size);

	glUseProgram(program);

	glBindVertexArray(vertexArrayObject[0]);
	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[0]);
	glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(Vec2f) * vertexLen, coord);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBuffer[1]);
	glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(Vec2f) * vertexLen, texCoord);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, arrayBuffer[2]);
	glBufferSubData(GL_ELEMENT_ARRAY_BUFFER, 0, sizeof(RenderInt) * coordIndexLen, coordIndex);
}


bool Window::loadSurfaceFromMemory() {
	if (dynamic_cast<Menu*>(this)) {
		int a = 0;
	}
	if (shader) {
		

		shader->setCoord(pos, size);

		shader->updatePosSize(pos, size);

		pixBuf.length = size.x * size.y;
		
		glBindTexture(GL_TEXTURE_2D, shader->texture);

		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, size.x, size.y, GL_BGRA, GL_UNSIGNED_BYTE, pixBuf.getArray());


	}
	return true;
}

float t_texCoord3[] = {
	0, 1,
	1, 1,
	0, 0,
	1, 0,
};
RenderInt t_coordIndex3[] = {
	1, 2, 0,
	1, 3, 2
};


void WinShader::setCoord(Vec2i pos, Vec2i size) {
	
	float top = -float(pos.y) / float(g_3d->size.y) * 2 + 1;
	float bottom = -float(size.y) / float(g_3d->size.y) * 2 + top;

	float left = float(pos.x) / float(g_3d->size.x) * 2 - 1;
	float right = float(size.x) / float(g_3d->size.x) * 2 + left;

	coord[0] = Vec2f(left, bottom);
	coord[1] = Vec2f(right, bottom);
	coord[2] = Vec2f(left, top);
	coord[3] = Vec2f(right, top);
}

bool Window::initWindow() {
	if (dynamic_cast<Menu*>(this)) {
		int a = 0;
	}
	if (g_bReady) {
		releaseWindow();
		//pos = Vec2i(1000, 690);

		shader = new WinShader;

		shader->coord = new Vec2f[4];

		shader->texCoord = t_texCoord3;
		shader->coordIndex = t_coordIndex3;

		shader->setupShaders();
		shader->setCoord(pos, size);
		shader->initMesh();
		shader->initTexture(size, reinterpret_cast<unsigned char*>(pixBuf.getArray()));
	}

	return true;
}


void Window::render() {
	if (g_bReady == false) {
		return;
	}
	if (dynamic_cast<Menu*>(this)) {
		int a = 0;
	}
	if (g_settings->showToolBar == false) {
		if (this == g_tb.p) {
			if (g_mousePos.x < g_tb->pos.x ||
				g_mousePos.x > g_tb->size.x + g_tb->pos.x ||
				g_mousePos.y < g_tb->pos.y ||
				g_mousePos.y > g_tb->size.y + g_tb->pos.y
				) {
				return;
			}
		}
		if (this == g_tb2.p) {
			if (g_mousePos.x < g_tb2->pos.x ||
				g_mousePos.x > g_tb2->size.x + g_tb2->pos.x ||
				g_mousePos.y < g_tb2->pos.y ||
				g_mousePos.y > g_tb2->size.y + g_tb2->pos.y
				) {
				return;
			}
		}
	}
	RECT srcRect;
	srcRect.top = 0;
	srcRect.left = 0;
	srcRect.right = size.x;
	srcRect.bottom = size.y;

	POINT dstPoint = { pos.x, pos.y };
	if (dstPoint.x < 0) {
		srcRect.left = -dstPoint.x;
		dstPoint.x = 0;
	}
	if (dstPoint.x + size.x >= g_3d->size.x) {
		srcRect.right = size.x - ((dstPoint.x + size.x) - g_3d->size.x);
	}
	if (dstPoint.y < 0) {
		srcRect.top = -dstPoint.y;
		dstPoint.y = 0;
	}
	if (dstPoint.y + size.y >= g_3d->size.y) {
		srcRect.bottom = size.y - ((dstPoint.y + size.y) - g_3d->size.y);
	}

	if (shader == 0) {
		initWindow();
	}
	if (updateShader) {
		shader->updatePosSize(pos, size);
		loadSurfaceFromMemory();
		updateShaderPosSize = false;
		updateShaderTexture = false;
		updateShader = false;
	}
	else {
		if (updateShaderPosSize) {
			shader->updatePosSize(pos, size);
			updateShaderPosSize = false;
		}
		if (updateShaderTexture) {
			loadSurfaceFromMemory();
			updateShaderTexture = false;
		}
	}

	shader->render();
}

void PartClue::renderNode(Matrix matrix, Matrix catRotMatrix) {
	return;
}

void TransformObj::renderNode(Matrix parentMatrix, Matrix catRotMatrix) {
	if (g_scene->mode & SINGLE_ACTIVE_MODE) {
		return;
	}
	//glDisable(GL_DEPTH_TEST);


	Vec3f pivotPointVec = parentMatrix * Vec3f();// * (pivotPoint + translation);
	Matrix tranMatrix = translationMatrix(pivotPointVec);
	g_lastPoint->shader->viewMatrix = tranMatrix;
	g_lastPoint->shader->render();


	for (int k = rotCtrlPnt.length; --k >= 0;) {
		Vec3f rcp = rotCtrlPnt[k];
		Matrix cpMatrix = translationMatrix(rcp);
		g_nextPoint->shader->viewMatrix = cpMatrix;
		g_nextPoint->shader->render();
	}

	IndexedLineSet* vcEllipticalPolyline2D = dynamic_cast<IndexedLineSet*>(g_transformLeverClue->geometry.p);
	M_Vec3f& coord = dynamic_cast<Coordinate*>(vcEllipticalPolyline2D->coord.p)->point;

	//coord[0] = pivotPointVec;
	//coord[1] = rotCtrlPnt;

	Matrix lineMatrix;
	g_transformLeverClue->mesh.shader->viewMatrix = lineMatrix;//do we need this?
	vcEllipticalPolyline2D->updateAllVertexInBuffers(g_transformLeverClue->mesh);

	g_transformLeverClue->renderNode(lineMatrix, catRotMatrix);


	//glEnable(GL_DEPTH_TEST);
}


void TransformCP::renderNode(Matrix parentMatrix, Matrix catRotMatrix) {
	if (MultiSelect::pointList.length == 0) {
		return;
	}

	Matrix identity;


	glDisable(GL_DEPTH_TEST);

	if (showDimFilterVisual && g_translateDimensionClue->mesh.shader && g_scene->mode & MODELLING_MODE) {

		IndexedLineSet* ils = dynamic_cast<IndexedLineSet*>(g_translateDimensionClue->geometry.p);

		Vec3f pivotPointVec = (pivotPoint + translation);

		Matrix lineMatrix0 = translationMatrix(pivotPointVec) * parentMatrix;

		g_translateDimensionClue->renderNode(lineMatrix0, catRotMatrix);

		/////////////
		if (xTug) {
			g_redConeBase->renderNode(parentMatrix, catRotMatrix);
		}

		if (yTug) {
			g_greenConeBase->renderNode(parentMatrix, catRotMatrix);
		}

		if (zTug) {
			g_blueConeBase->renderNode(parentMatrix, catRotMatrix);
		}
	}
	else if (false == (g_scene->mode & SINGLE_ACTIVE_MODE) && MultiSelect::pointList.length >= leverSelection) {


		if (CNCShader* lastShader = dynamic_cast<CNCShader*>(g_lastPoint->shader)) {
			Vec3f pivotPointVec = parentMatrix * (pivotPoint + translation);
			lastShader->viewMatrix = translationMatrix(pivotPointVec);
			lastShader->rotationMat = identity;
			lastShader->render();
		}


		if (CNCShader* shader = dynamic_cast<CNCShader*>(g_nextPoint->shader)) {

			for (int k = rotCtrlPnt.length; --k >= 0;) {
				Vec3f rcp = rotCtrlPnt[k];
				shader->viewMatrix = translationMatrix(rcp);
				shader->rotationMat = identity;
				shader->render();

			}
		}

		IndexedLineSet* vcEllipticalPolyline2D = dynamic_cast<IndexedLineSet*>(g_transformLeverClue->geometry.p);


		g_transformLeverClue->mesh.shader->viewMatrix = identity;

		vcEllipticalPolyline2D->updateAllVertexInBuffers(g_transformLeverClue->mesh);

		g_transformLeverClue->renderNode(identity, catRotMatrix);

	}

	glEnable(GL_DEPTH_TEST);
}


void GeomRender::convertToRenderFormat() {

	if (normal.length != coord.length) {
		genNormals();
	}

	if (CNMonoCShader* cmc = dynamic_cast<CNMonoCShader*>(shader)) {
		cmc->convertToRenderFormat(*this);
	}
	else if (normal.length) {

		CNCShader* cnc = new CNCShader;
		shader = cnc;
		cnc->convertToRenderFormat(*this);
	}
	else if (color.length == 1) {
		CMonoCLineShader* cmlc = new CMonoCLineShader;
		shader = cmlc;
		cmlc->f_color = color[0];
		cmlc->convertToRenderFormat(*this);
	}
	else {
		CCLineShader* cc = new CCLineShader;
		shader = cc;
		cc->convertToRenderFormat(*this);

	}
}

void convertVisualCluesToRenderFormat() {
	g_redPoint->convertToRenderFormat();
	g_nextPoint->convertToRenderFormat();
	g_lastPoint->convertToRenderFormat();
	g_yellowPoint->convertToRenderFormat();
	g_triClueShape->convertToRenderFormat();


	g_redConePoint->convertToRenderFormat();
	g_greenConePoint->convertToRenderFormat();
	g_blueConePoint->convertToRenderFormat();

}

void IndexedLineSet::updateAllVertexInBuffers(GeomRender& mesh) {
	Coordinate* coord = dynamic_cast<Coordinate*>(this->coord.p);
	Color* color = dynamic_cast<Color*>(this->color.p);
	if (coord == 0)return;
	M_Vec3f& point = coord->point;

	mesh.coord = point;

	if (color && color->color.length == point.length) {
		mesh.color = color->color;
	}

	mesh.shader->convertToRenderFormat(mesh);
}

void Shape::convertToRenderFormat() {
	if (IndexedLineSet* ils = dynamic_cast<IndexedLineSet*>(geometry.p)) {
		ils->convertToRenderFormat(mesh, appearance);
	}
}

void toggleWireFrame() {
#ifdef PROG_ROB
	dLog("new ClickButton(toggleWireFrame),");
#endif
	//g_scene->mode.refValue() = (g_scene->mode.refValue() & ~WIREFRAME_MODE) | (g_scene->mode.refValue() ^ WIREFRAME_MODE);
	g_scene->mode.refValue() = g_scene->mode.refValue() ^ WIREFRAME_MODE;
	g_wireFrameMode = g_scene->mode.refValue() & WIREFRAME_MODE;

	if (g_wireFrameMode) {
		glPolygonMode(GL_FRONT, GL_LINE);
	}
	else {
		glPolygonMode(GL_FRONT, GL_FILL);
	}
}

void PartCageClue::renderNode(Matrix parentMatrix, Matrix catRotMatrix) {
	g_nurbsIlsBase->renderNode(parentMatrix, catRotMatrix);


	for (int k = MultiSelect::pointList.length; --k >= 0;) {
		NurbsPatch* patch = MultiSelect::pointList[k].p;
		Vec3f point = patch->dpoint.refValue()[MultiSelect::pointList[k].i];
		Vec3f pointTran = parentMatrix * point;

		Matrix tranMatrix = translationMatrix(pointTran);
		g_redPoint->shader->viewMatrix = tranMatrix;
		g_redPoint->shader->render();
	}


	if (g_transformLeverClue->mesh.shader) {
		if (MultiSelect::pointList.length) {
			g_transformCP->renderNode(parentMatrix, catRotMatrix);
		}
	}
}

void AnimLever::renderNode(Matrix matrix, Matrix catRotMatrix) {
	if (Art* part = dynamic_cast<Art*>(g_sceneTreeList[g_scene->selLine].node.p)) {
		if (part->smls->workAnimCP) {
			g_partCageClue->renderNode(matrix, catRotMatrix);
		}
		if (g_vcArtAnimBase->showLever == false) {
			return;
		}
		//g_pd3dDevice->SetRenderState(D3DRS_ZENABLE, FALSE);
		//glDepthMask(false);
		Seamless* smls = part->smls;

		if (MultiSelect::pointList.length < g_transformCP->leverSelection) {
			Vec3f pivotPointVec;
			Matrix tranMatrix;




			pivotPointVec = part->pivotPointMatrix * Vec3f(0, 0, 0);
			//g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
			tranMatrix = translationMatrix(pivotPointVec);


			for (int k = rotCtrlPnt.length; --k >= 0;) {
				
				g_nextPoint->shader->viewMatrix = translationMatrix(rotCtrlPnt[k]);
				g_nextPoint->shader->render();


			}


			{


		////////////////



				if (smls->workAnimCP == false && MultiSelect::coordList.length) {
					g_viewMatrix = g_scene->getViewMatrix();
					if (g_dragActuallyHappened == false) {
						smls->genDynamicMatricesAndPopulate();
					}
					for (int i = MultiSelect::coordList.length; --i >= 0;) {
						Art* part = MultiSelect::coordList[i].p;
						int cdx = MultiSelect::coordList[i].i;
						Vec3f crd = m_matrix[m_smi[part->fromCoord + cdx]] * g_viewMatrix * (g_m_coord[part->fromCoord + cdx] - part->absolutePivotPoint);

						//Render* render = part->smls->render;
						//if(render==0)return;

						//switch(render->type){
						//	case COLOR_CUSTOMVERTEX_0WEIGHT:{
						//		setCrd<COLORVERTEX_0WEIGHT>(render, part, cdx, crd);
						//		break;
						//	}
						//	case TEXTURE_CUSTOMVERTEX_0WEIGHT:{
						//		setCrd<TEXTUREVERTEX_0WEIGHT>(render, part, cdx, crd);
						//		break;
						//	}
						//}

						tranMatrix = translationMatrix(crd);

					}
				}
			}
			if (MultiSelect::pointList.length < g_transformCP->leverSelection) {
				IndexedLineSet* vcEllipticalPolyline2D = dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);
				M_Vec3f& coord = dynamic_cast<Coordinate*>(vcEllipticalPolyline2D->coord.p)->point;

				Matrix lineMatrix;
				vcEllipticalPolyline2D->updateAllVertexInBuffers(g_vcEllipticalPolyline2D->mesh);

				g_vcEllipticalPolyline2D->mesh.renderNode(lineMatrix, catRotMatrix);
			}
		}
		if (MultiSelect::selectRectMode && smls->animCP == false) {
			g_vcSelectRectShape->renderNode(Matrix(), catRotMatrix);
		}
		//glDepthMask(true);
		//g_pd3dDevice->SetRenderState(D3DRS_ZENABLE, TRUE);
	}
}


Shader* createShader(Ptr<Effect> effect) {
	if (effect == nullPtr) {
		return nullPtr;
	}
	if (dynamic_cast<TextureEffect*>(effect.p)) {
		throw;
	}
	else {
		return new CNMonoCShader;
	}
}

void Sphere::renderNode(Matrix parentMatrix, Matrix catRotMatrix, GeomRender& mesh) {
	if (renderProcessed == false || mesh.shader == 0) {
		Ptr<Effect> effect = workOutEffectType(mesh.appearance.p);
		genSphere(radius, 12, 6, mesh, effect);
		mesh.shader = createShader(effect);
		mesh.convertToRenderFormat();
		renderProcessedQue += this;
	}
	mesh.renderNode(parentMatrix, catRotMatrix);
}

void Box::renderNode(Matrix parentMatrix, Matrix catRotMatrix, GeomRender& mesh) {
	if (renderProcessed == false || mesh.shader == 0) {
		Ptr<Effect> effect = workOutEffectType(mesh.appearance);
		genBox(size, 1, 1, mesh, effect);
		mesh.shader = createShader(effect);
		mesh.convertToRenderFormat();
		renderProcessedQue += this;
	}
	mesh.renderNode(parentMatrix, catRotMatrix);
}

void Cylinder::renderNode(Matrix parentMatrix, Matrix catRotMatrix, GeomRender& mesh) {
	if (renderProcessed == false || mesh.shader == 0) {
		Ptr<Effect> effect = workOutEffectType(mesh.appearance);
		genCylinder(height, radius, side, top, bottom, 12, 1, mesh, effect);
		mesh.shader = createShader(effect);
		mesh.convertToRenderFormat();
		renderProcessedQue += this;
	}
	mesh.renderNode(parentMatrix, catRotMatrix);
}

void Cone::renderNode(Matrix parentMatrix, Matrix catRotMatrix, GeomRender& mesh) {
	if (renderProcessed == false || mesh.shader == 0) {
		Ptr<Effect> effect = workOutEffectType(mesh.appearance);
		genCone(height, bottomRadius, side, bottom, 16, mesh, effect);
		mesh.shader = createShader(effect);
		mesh.convertToRenderFormat();
		renderProcessedQue += this;
	}
	mesh.renderNode(parentMatrix, catRotMatrix);
}


void Mesh::renderNode(Shader* shader, Matrix parentMatrix, Matrix catRotMatrix, const Appearance* appearance) {
	if (CNMonoCShader* cnmc = dynamic_cast<CNMonoCShader*>(shader)) {
		if (Material* material = dynamic_cast<Material*>(appearance->material.p)) {
			cnmc->f_color = material->diffuseColor;
		}
	}
	renderNode(shader, parentMatrix, catRotMatrix);
}


void Text::renderNode(Matrix parentMatrix, Matrix catRotMatrix, GeomRender& mesh) {}
void PointSet::renderNode(Matrix parentMatrix, Matrix catRotMatrix, GeomRender& mesh) {}
void ElevationGrid::renderNode(Matrix parentMatrix, Matrix catRotMatrix, GeomRender& mesh) {}

void Viewpoint::renderNode(Matrix parentMatrix, Matrix catRotMatrix) {
	if ((g_scene->mode & ANIMATION_MODE) && g_showCamerasInAnimMode == false) {
		return;
	}
	if (g_scene->mode & (HOME_MODE)) {
		return;
	}
	if ((g_scene->mode & (SELECT_OBJECT_MODE | ANIMATION_MODE)) && g_cameraObject) {

		g_cameraObject->translation.refValue() = position;
		g_cameraObject->rotation.refValue() = orientation;

		g_cameraObject->renderNode(parentMatrix, catRotMatrix);
	}

	//	if(g_visualClue != this){
	//		return;
	//	}
	if (this != dynamic_cast<Viewpoint*>(g_selNode.getRaw())) {
		return;
	}
	glDisable(GL_DEPTH_TEST);


	Vec3f pivotPointVec = parentMatrix * position;
	Matrix tranMatrix = translationMatrix(pivotPointVec);

	g_yellowPoint->shader->viewMatrix = translationMatrix(pivotPointVec);
	g_yellowPoint->shader->render();


	
	Vec3f rcp = parentMatrix * lookAt;
	Matrix cpMatrix = translationMatrix(rcp);

	g_yellowPoint->shader->viewMatrix = translationMatrix(pivotPointVec);
	g_yellowPoint->shader->render();

	Vec3f upp = parentMatrix * (up + position);
	Matrix upMatrix = translationMatrix(upp);

	g_nextPoint->shader->viewMatrix = translationMatrix(pivotPointVec);
	g_nextPoint->shader->render();
	////////////////

	IndexedLineSet* polyLine = dynamic_cast<IndexedLineSet*>(g_nurbsIlsShape->geometry.p);
	M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point.refValue();
	M_Color3f& color = dynamic_cast<Color*>(polyLine->color.p)->color.refValue();
	M_Int& coordIndex = polyLine->coordIndex.refValue();

	//coord[0] = pivotPointVec;
	//coord[1] = rotCtrlPnt;


	Matrix lineMatrix = translationMatrix(pivotPointVec);

	polyLine->updateAllVertexInBuffers(g_nurbsIlsShape->mesh);

	g_nurbsIlsBase->renderNode(parentMatrix, catRotMatrix);

	//g_nurbsIlsShape->renderNode(lineMatrix0, parentColScale);


	//if(g_wireFrameMode){
	//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
	//}
	glEnable(GL_DEPTH_TEST);



	{//dubug
		IndexedLineSet* vcEllipticalPolyline2D = dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);
		Matrix lineMatrix = parentMatrix;

		vcEllipticalPolyline2D->updateAllVertexInBuffers(g_vcEllipticalPolyline2D->mesh);

		//g_vcEllipticalPolyline2D->renderNode(lineMatrix0, parentColScale);
	}//end debug

}


void renderPivotPoint(Seamless* smls, Node* selNode) {

	M_Art& partList = smls->partList;


	for (int k = 0; k < partList.length; k++) {
		Art* part = partList[k];

		if (part->ik && part->niki == 0) {
			Vec3f ikPos = g_scene->getViewMatrix() * part->ikPos;
			Matrix tranMatrix = translationMatrix(ikPos);
			g_yellowPoint->shader->viewMatrix = tranMatrix;
			g_yellowPoint->shader->render();

		}
		else if (part != selNode) {
			Vec3f pivotPointVec = part->pivotPointMatrix * Vec3f(0, 0, 0);
			Matrix tranMatrix = translationMatrix(pivotPointVec);
			g_yellowPoint->shader->viewMatrix = tranMatrix;
			g_yellowPoint->shader->render();
		}

	}

	if (Art* part = dynamic_cast<Art*>(selNode)) {
		if (part->smls == smls) {
			Vec3f pivotPointVec = part->pivotPointMatrix * Vec3f(0, 0, 0);
			Matrix tranMatrix = translationMatrix(pivotPointVec);
			g_redPoint->shader->viewMatrix = tranMatrix;
			g_redPoint->shader->render();

		}
	}
}


void renderPivotPoint(Art* part, Matrix& tranMatrix, Matrix catRotMatrix) {
}


void StickMan::renderStickMan(Seamless* smls, Matrix parentMatrix, Matrix catRotMatrix) {

	if (partList.length > 1) {
		if (shape->mesh.shader == 0) {
			shape->mesh.shader = new CMonoCLineShader;
			shape->convertToRenderFormat();
		}

		M_Vec3f& point = coord->point;
		M_Vec3f& coord = shape->mesh.coord.refValue();

		for (int i = coord.length; --i >= 0; ) {
			Vec3f vec = partList[i]->pivotPointMatrix * Vec3f(0, 0, 0);
			coord[i] = vec;
			point[i] = vec;
		}

		if (g_scene->mode & (XRAY_EDIT_MODE | OWNERSHIP_MODE | NURBS_OWNERSHIP_MODE)) {
			renderPivotPoint(smls, g_selNode.getRaw());
		}
		else {
			renderPivotPoint(smls, 0);
		}


		shape->mesh.shader->convertToRenderFormat(shape->mesh);

		shape->renderNode(Matrix(), catRotMatrix);

	}
	if (g_scene->mode & (XRAY_EDIT_MODE | OWNERSHIP_MODE | NURBS_OWNERSHIP_MODE)) {
		renderPivotPoint(smls, g_selNode.getRaw());
	}
	else {
		renderPivotPoint(smls, 0);
	}
}
void Art::genDynamicTransforms(Matrix parentMatrix, Matrix catRotMatrix, int& smi, M_Matrix& boneMatrix, M_Matrix& rotBoneMatrix) {
	pivotPointMatrix = translationMatrix(translation + pivotPoint) * parentMatrix;

	for (int cg = 0; cg < coordGroup.length; cg++) {
		Matrix rotMatrix = rotationMatrix(Rotation(rotation.x, rotation.y, rotation.z, rotation.angle * weight[cg]));
		boneMatrix[smi] = scaleMatrix(scale) * rotMatrix * pivotPointMatrix;
		rotBoneMatrix[smi] = rotMatrix * catRotMatrix;
		smi++;
	}

	Matrix rotMatrix = rotationMatrix(rotation);
	pivotPointMatrix = scaleMatrix(scale) * rotMatrix * pivotPointMatrix;
	catRotMatrix = rotMatrix * catRotMatrix;

	boneMatrix[smi] = pivotPointMatrix;
	rotBoneMatrix[smi] = catRotMatrix;
	smi++;

	for (int k = 0; k < part.length; k++) {
		part.refValue()[k]->genDynamicTransforms(pivotPointMatrix, catRotMatrix, smi, boneMatrix, rotBoneMatrix);
	}
}

void Art::genDynamicTransformsCP(Matrix parentMatrix, Matrix parentMatrixCP, int& smi) {
	pivotPointMatrix = translationMatrix(translation + pivotPoint) * parentMatrix;

	for (int cg = 0; cg < coordGroup.length; cg++) {
		smi++;
	}

	pivotPointMatrix = scaleMatrix(scale) * rotationMatrix(rotation) * pivotPointMatrix;

	smi++;
	for (int k = 0; k < part.length; k++) {
		part.refValue()[k]->genDynamicTransformsCP(pivotPointMatrix, parentMatrixCP, smi);
	}
}

void Seamless::genDynamicTransforms() {
	Matrix parentMatrix = g_scene->getViewMatrix();
	Matrix catRotMatrix = rotationMatrix(g_scene->rotation);
	M_SArt& skeleton = this->skeleton.refValue();

	if (workAnimCP) {
		for (int i = 0, smi = 0; i < skeleton.length; i++) {
			skeleton[i]->genDynamicTransformsCP(parentMatrix, parentMatrix, smi);
		}
	}
	else {
		if (CNC0WeightBoneShader* boneShader = dynamic_cast<CNC0WeightBoneShader*>(mesh.shader)) {
			for (int i = 0, smi = 0; i < skeleton.length; i++) {
				skeleton[i]->genDynamicTransforms(parentMatrix, catRotMatrix, smi, boneShader->boneMatrix, boneShader->rotBoneMatrix);
			}
		}
	}
}

Window::~Window() {
	if (shader) {
		shader->deleteProgram();
	}
}

void setBackface(bool backfaceRender) {
	if (backfaceRender != g_backfaceRender) {
		g_backfaceRender = backfaceRender;
		if (backfaceRender) {
			//g_pd3dDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
		}
		else {
			//g_pd3dDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
		}
	}
}

bool read_png_file(const char* file_name, M_<Pix32>& character, Vec2i& size);
bool read_jpg_file(const char* file_name, M_<Pix32>& character, Vec2i& size);


/////////////////////////////////////
#include <stdio.h>
#include "jpeglib.h"
#include <setjmp.h>

int g_i = 0;

Pix24* g_pix24Buf = new Pix24[1024 * 1024];

/*
unsigned char myImage[] = {
	0, 0, 255,  0, 0, 255,  0, 0, 255,  0, 0, 255, 0, 0, 255,  0, 0, 255,  0, 0, 255,  0, 0, 255,
	0, 0, 255,  0, 0, 255,  0, 0, 255,  0, 0, 255, 0, 0, 255,  0, 0, 255,  0, 0, 255,  0, 0, 255,
	0, 0, 255,  0, 0, 255,  0, 0, 255,  0, 0, 255, 0, 0, 255,  0, 0, 255,  0, 0, 255,  0, 0, 255,
	0, 0, 255,  0, 0, 255,  0, 0, 255,  0, 0, 255, 0, 0, 255,  0, 0, 255,  0, 0, 255,  0, 0, 255,

	0, 0, 255,  0, 0, 255,  255, 0, 0,  255, 0, 0, 0, 0, 255,  0, 0, 255,  0, 0, 255,  0, 0, 255,
	0, 0, 255,  0, 0, 255,  255, 0, 0,  255, 0, 0, 0, 0, 255,  0, 0, 255,  0, 0, 255,  0, 0, 255,
	0, 0, 255,  0, 0, 255,  255, 0, 0,  255, 0, 0, 0, 0, 255,  0, 0, 255,  0, 0, 255,  0, 0, 255,
	0, 0, 255,  0, 0, 255,  255, 0, 0,  255, 0, 0, 0, 0, 255,  0, 0, 255,  0, 0, 255,  0, 0, 255,
};
*/
extern "C" JSAMPLE * image_buffer = (JSAMPLE *)&g_pix24Buf[0];	/* Points to large array of R,G,B-order data */
//extern "C" JSAMPLE * image_buffer = new JSAMPLE[0x100000];	/* Points to large array of R,G,B-order data */
extern "C" int image_height = 8;	/* Number of rows in image */
extern "C" int image_width = 8;		/* Number of columns in image */



extern "C" void put_scanline_someplace(char* buffer, int row_stride) {
	for (int i = 0; i < row_stride; i++) {
		image_buffer[g_i++] = buffer[i];
	}
}
//void jpeg_get_small() {}
//void jpeg_free_small() {}
//void jpeg_get_large(){}
//void jpeg_free_large() {}
//void jpeg_mem_available() {}
//void jpeg_open_backing_store() {}
//void jpeg_mem_init() {}
//void jpeg_mem_term() {}

extern "C" GLOBAL(void)write_JPEG_file(char * filename, int quality);

extern "C" GLOBAL(int)read_JPEG_file(const char * filename, int* width, int* height);

bool read_jpg_file(const char* file_name, M_<Pix32>& character, Vec2i& size) {

	image_buffer[0] = 0;
	read_JPEG_file(file_name, &size.x, &size.y);

	character.length = size.x * size.y;

	for (int r = 0; r < size.y; r++) {
		for (int c = 0; c < size.x; c++) {
			Pix24 pix = g_pix24Buf[(size.y - 1 - r) * size.x + c];
			character[r * size.x + c] = Pix32(pix.r, pix.g, pix.b, 0xff);
		}
	}

	return true;
}


bool readTextureLife(String filename, M_<Pix32>& pixels, Vec2i& size) {

	filename = filename.toUpperCase();

	int end = filename.length - 5;

	if (end >= 0 &&
		filename[end + 0] == '.' &&
		filename[end + 1] == 'J' &&
		filename[end + 2] == 'P' &&
		filename[end + 3] == 'E' &&
		filename[end + 4] == 'G') {

		return read_jpg_file(filename.c_str(), pixels, size);
	}
	end++;
	if (end >= 0 &&
		filename[end + 0] == '.' &&
		filename[end + 1] == 'J' &&
		filename[end + 2] == 'P' &&
		filename[end + 3] == 'G') {

		return read_jpg_file(filename.c_str(), pixels, size);
	}
	if (end >= 0 &&
		filename[end + 0] == '.' &&
		filename[end + 1] == 'P' &&
		filename[end + 2] == 'N' &&
		filename[end + 3] == 'G') {

		return read_png_file(filename.c_str(), pixels, size);
	}
	return false;
}

void ColorEffect::renderMaterial(int triFrom, Seamless& smls, Matrix& parentMatrix, Matrix& catRotMatrix) {
	if (g_renderLater == false) {
		if (transparency > 0) {
			g_transparentList += TransparencyList(&smls, parentMatrix, catRotMatrix);
			return;
		}
	}
	else {
		if (false == (transparency > 0)) {
			return;
		}
	}
	setBackface(backfaceRender);

	M_SArt& skeleton = smls.skeleton.refValue();

	GeomRender& mesh = smls.mesh;
	CNCShader* shader = dynamic_cast<CNCShader*>(mesh.shader);

	if (shader == nullptr) {
		return;
	}

	shader->viewMatrix = parentMatrix;
	shader->rotationMat = catRotMatrix;

	glUseProgram(shader->program);

	glUniformMatrix4fv(shader->rotationMatrixLoc, 1, GL_FALSE, &shader->rotationMat.f11);
	glUniformMatrix4fv(shader->projMatrixLoc, 1, GL_FALSE, &projMatrix.f11);
	glUniformMatrix4fv(shader->viewMatrixLoc, 1, GL_FALSE, &shader->viewMatrix.f11);


	glBindVertexArray(shader->vertexArrayObject[0]);

	glUniform1f(shader->transparencyLoc, 1.0 - transparency);

	RenderInt* triIndex = 0;
	triIndex += triFrom * 3;
	if (transparency != 0) {
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDrawElementsBaseVertex(GL_TRIANGLES, triLen * 3, RENDER_INT, triIndex, 0);
		glDisable(GL_BLEND);
	}
	else {
		//RenderInt* ptr = 0;
		//glDrawElementsBaseVertex(GL_TRIANGLES, 0, RENDER_INT, 0, 0);
			//glDrawElements(GL_TRIANGLES, shader->coordIndexLen / 2, GL_UNSIGNED_INT, (void*)(NULLptr + 3));
		glDrawElementsBaseVertex(GL_TRIANGLES, triLen * 3, RENDER_INT, triIndex, 0);

		//glDrawElementsBaseVertex(GL_TRIANGLES, shader->coordIndexLen, RENDER_INT, 0, 0);
		//glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_INT, coordIndex);
	}

	glBindVertexArray(0);
}

void TextureEffect::renderMaterial(int triFrom, Seamless& smls, Matrix& parentMatrix, Matrix& catRotMatrix) {
	if (g_renderLater == false) {
		if (alphaTexture) {
			g_alphaChanList += AlphaChanList(&smls, parentMatrix, catRotMatrix);
			return;
		}
		if (transparency > 0) {
			g_transparentList += TransparencyList(&smls, parentMatrix, catRotMatrix);
			return;
		}
	}
	else {
		if (false == (alphaTexture || transparency > 0)) {
			return;
		}
	}
	setBackface(backfaceRender);

	M_SArt& skeleton = smls.skeleton.refValue();

	if (g_browseMode == false && url.update && url.refValue().length) {
		dLog("set Texture from local file");
		url.update = false;


		if (CNTShader* shader = dynamic_cast<CNTShader*>(smls.mesh.shader)) {
			Vec2i size;
			readTextureLife(url, shader->pixels, size);
			shader->initTexture(size, &shader->pixels.getArray()[0].r);
		}
		else if (CNT0WeightBoneShader* shader = dynamic_cast<CNT0WeightBoneShader*>(smls.mesh.shader)) {
			Vec2i size;
			readTextureLife(url, shader->pixels, size);
			shader->initTexture(size, &shader->pixels.getArray()[0].r);
		}

	}
	GeomRender& mesh = smls.mesh;
	mesh.renderNode(parentMatrix, catRotMatrix);
}
void MonoColorEffect::renderMaterial(int triFrom, Seamless& smls, Matrix& parentMatrix, Matrix& catRotMatrix) {
}

unsigned onCursorImage[] = {
	0xff0000,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0xff0000,
};
unsigned onCursorImageInverted[] = {
	0x00ffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0x00ffff,
};
#define ON_CURSOR_IMAGE_LEN (sizeof(onCursorImage)/sizeof(unsigned))
unsigned offCursorImage[ON_CURSOR_IMAGE_LEN];
inline void CaretWindow::doWinkOff(const Vec2i& c, Rect& srcRect)const {
	if (static_cast<unsigned>(c.x) < static_cast<unsigned>(win.size.x) && static_cast<unsigned>(c.y + ON_CURSOR_IMAGE_LEN) < static_cast<unsigned>(win.size.y)) {

		const int* p = win.pixBuf.getArray();
		for (int i = ON_CURSOR_IMAGE_LEN; --i >= 0;) {
			offCursorImage[i] = p[(c.y + i)*win.size.x + c.x];
		}

		RECT dstRect;
		dstRect.left = c.x;
		dstRect.top = c.y;
		dstRect.right = c.x + 1;
		dstRect.bottom = c.y + ON_CURSOR_IMAGE_LEN;

		WinShader* shader = win.shader;
		glBindTexture(GL_TEXTURE_2D, shader->texture);
		glTexSubImage2D(GL_TEXTURE_2D, 0, c.x, c.y, 1, ON_CURSOR_IMAGE_LEN, GL_RGBA, GL_UNSIGNED_BYTE, win.pixBuf.getArray());
	}
}

inline void CaretWindow::doWinkOn(const Vec2i& c, Rect& srcRect)const {
	if (static_cast<unsigned>(c.x) < static_cast<unsigned>(win.size.x) && static_cast<unsigned>(c.y + ON_CURSOR_IMAGE_LEN) < static_cast<unsigned>(win.size.y)) {
		RECT dstRect;
		dstRect.left = c.x;
		dstRect.top = c.y;
		dstRect.right = c.x + 1;
		dstRect.bottom = c.y + ON_CURSOR_IMAGE_LEN;

		unsigned * oCI = onCursorImage;
		if (g_settings->invertWindowColor) {
			oCI = onCursorImageInverted;
		}
		WinShader* shader = win.shader;
		glBindTexture(GL_TEXTURE_2D, shader->texture);
		glTexSubImage2D(GL_TEXTURE_2D, 0, c.x, c.y, 1, ON_CURSOR_IMAGE_LEN, GL_RGBA, GL_UNSIGNED_BYTE, oCI);
	}
}
void CaretWindow::doCallBack() {//renderCaret
	if (win.shader == 0) {
		return;
	}

	Rect srcRect;
	srcRect.left = 0;
	srcRect.top = 0;
	srcRect.right = 1;
	srcRect.bottom = ON_CURSOR_IMAGE_LEN;

	Vec2i currentPos = getCaretPos();
	//if(remPreviousPos!=currentPos || firstVisChar!=remFirstChar){
	if ((remPreviousPos != currentPos || firstVisChar != remFirstChar) && remPreviousPos.y >= 0) {//remPreviousPos.y >= 0 seems to stop crashes when clicking on scene tree node
		winkOn = true;
		lastWinkTime = g_time;
		doWinkOff(remPreviousPos, srcRect);
		doWinkOn(currentPos, srcRect);

	}
	else if (g_time > lastWinkTime + .5) {
		lastWinkTime = g_time;
		winkOn = !winkOn;
		if (winkOn) {
			doWinkOn(currentPos, srcRect);
		}
		else {
			doWinkOff(currentPos, srcRect);
		}
	}
	remPreviousPos = currentPos;
	remFirstChar = firstVisChar;
}
void CaretWindow::doNoThinkWink() {
	Rect srcRect;
	srcRect.left = 0;
	srcRect.top = 0;
	srcRect.right = 1;
	srcRect.bottom = ON_CURSOR_IMAGE_LEN;

	Vec2i currentPos = getCaretPos();
	if (winkOn) {
		doWinkOn(currentPos, srcRect);
	}
	else {
		doWinkOff(currentPos, srcRect);
	}
}
void CaretWindow::doNoThinkWinkOff() {
	Rect srcRect;
	srcRect.left = 0;
	srcRect.top = 0;
	srcRect.right = 1;
	srcRect.bottom = ON_CURSOR_IMAGE_LEN;

	Vec2i currentPos = getCaretPos();
	doWinkOff(currentPos, srcRect);
}

