#include <string.h>
using namespace System;
using namespace System::Speech::Synthesis;
using namespace System::Collections::Generic;
using namespace System::Collections::ObjectModel;
using namespace System::Diagnostics;

public ref class Globals abstract sealed {
public:
	static SpeechSynthesizer^ speaker = gcnew SpeechSynthesizer;
	static InstalledVoice^ defaultMaleVoice = nullptr;
	static InstalledVoice^ defaultFemaleVoice = nullptr;
};

const char* maleVoices[]={
"Microsoft Sam",
"LH Michael",
"Microsoft Mike",
};
const char* femaleVoices[]={
"Microsoft Anne",
"LH Michelle",
"Microsoft Mary",
};
bool cmpstr(String^ strA, const char* strB){
	int lenB = strlen(strB);
	if(lenB!=strA->Length){
		return false;
	}
	for(int i = lenB; --i>=0;){
		if(strA[i] != strB[i]){
			return false;
		}
	}
	return true;
}
bool findDefaultVoiceByName(List<InstalledVoice^>% listOfVoices, InstalledVoice^% defaultVoice, const char** voices, int numOfVoices){
	for(int j = numOfVoices; --j>=0;){
		for(int i = listOfVoices.Count; --i>=0;){
			InstalledVoice^ voice = listOfVoices[i];
			if(cmpstr(voice->VoiceInfo->Name, voices[j])){
				defaultVoice = voice;
				return true;
			}
		}
	}
	return false;
}
void findDefaultVoiceByGender(List<InstalledVoice^>% listOfVoices, InstalledVoice^% defaultVoice, VoiceGender gender){
	for(int i = listOfVoices.Count; --i>=0;){
		InstalledVoice^ voice = listOfVoices[i];
		if(voice->VoiceInfo->Gender == gender){
			Globals::defaultMaleVoice = voice;
			return;
		}
	}
}
void initDefaultVoice(){
	List<InstalledVoice^> listOfVoices = Globals::speaker->GetInstalledVoices();

	if(listOfVoices.Count){
		Globals::defaultFemaleVoice = Globals::defaultMaleVoice = listOfVoices[0];
		if(false == findDefaultVoiceByName(listOfVoices, Globals::defaultFemaleVoice, femaleVoices, sizeof(femaleVoices)/sizeof(char*))){
			findDefaultVoiceByGender(listOfVoices, Globals::defaultFemaleVoice, VoiceGender::Female);
		}
		if(false == findDefaultVoiceByName(listOfVoices, Globals::defaultMaleVoice, maleVoices, sizeof(maleVoices)/sizeof(char*))){
			findDefaultVoiceByGender(listOfVoices, Globals::defaultMaleVoice, VoiceGender::Male);
		}
	}
}
bool checkVoiceExists(const char* voiceRequest){
	List<InstalledVoice^> listOfVoices = Globals::speaker->GetInstalledVoices();
	for(int i = listOfVoices.Count; --i>=0;){
		InstalledVoice^ voice = listOfVoices[i];
		if(cmpstr(voice->VoiceInfo->Name, voiceRequest)){
			return true;
		}
	}
	return false;
}
void textToSpeech(const char* str, bool male, const char* voice){
	if(Globals::defaultMaleVoice != nullptr){
		if(voice && voice[0] != 0){
			Globals::speaker->SelectVoice(gcnew String(voice));
		}
		else{
			if(male){
				Globals::speaker->SelectVoice(Globals::defaultMaleVoice->VoiceInfo->Name);
			}
			else{
				Globals::speaker->SelectVoice(Globals::defaultFemaleVoice->VoiceInfo->Name);
			}
		}
		Globals::speaker->SpeakAsync(gcnew String(str));
	}
}
