/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#define INITGUID
//#define OPENGL 1 //uncomment this line if using opengl if not defined else where

#include <windows.h>
#include <direct.h>
#include <process.h>
#include <shlobj.h>

#include <sys/stat.h>
#include <share.h>
#include <fcntl.h>
#include <io.h>

#include <vector>
#include <map>
#include <algorithm>




#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "resource.h"
#include "input.h"


#include "seamless3d_gui_render.h"
#include "robot_aid.h"
#include "chat_window.h"


#ifdef WACOM_FILES
#include "WacomMultiTouch.h"
#include "WintabUtils.h"

#define PACKETDATA	(PK_X | PK_Y | PK_BUTTONS | PK_NORMAL_PRESSURE)
#define PACKETMODE	PK_BUTTONS
#include "pktdef.h"
#endif



#ifdef RECORD_SOUNDMOVIE
extern HANDLE g_hNotificationEvent;
#endif

void outputScreenShotForDemo();
bool smlsIsSaved();
void saveBeforeClosing(void (*i_yesFunc)(FileCP& fileCP), void (*i_noFunc)(FileCP& fileCP), void (*i_cancelFunc)(FileCP& fileCP));
bool InitRender();
extern bool g_viewPosDragMode;
extern bool g_touchViewPosDragMode;
extern Vec3f g_touchBeginMousePos;
extern Vec3f g_touchBeginViewPos;
extern bool g_viewRotDragMode;
extern bool g_viewZRotDragMode;
extern Rotation g_beginViewRot;
extern Vec3f g_beginViewPos;
extern Vec2i g_beginMousePos;

void release3dDrag();
void release3dRightDrag();
void mouse2dSync();
void initPen();

HACCEL hAccel;
extern bool g_browseMode;
void renderScene();
void initializeDeveloperScript();
void insertRecentFile(const char* fullFileName);
void openDroppedFile();
bool checkAnimCtrlPoint(const bool perspective, Vec3f sr);
extern String droppedFilename;
extern bool g_recordSoundMovie;
extern bool g_playSoundMovie;
extern bool g_playing;
extern Ptr<PartCageClue> g_partCageClue;
extern BrushPtr g_brush;
extern M_<Ptr<Viewpoint> > g_viewpoints;

bool g_online = false;
bool g_viewer = false;
bool g_minimized = false;
bool g_processComnandLine = true;


double g_previousClickTime=0;
bool g_leftDoubleClicked = false;
double g_currentClickTime=0;

extern bool g_windowSizeWindowBool;
extern double g_windowSizeWindowTime;
extern Vec3f g_viewDragMousePos;

//void mouseLeaveCleanUp();

bool g_bActive  = true;//false; // Whether the app is active (not minimized)
bool g_bReady   = false; // Whether the app is ready to render frames

HINSTANCE  g_hinst;

unsigned int colorToHex( Color3f color);
bool setDirectInput( HWND hWnd );
void openSmlsLastPhase(String fullFilePath, const bool recover = false);
void setupRecentFiles();
bool checkForLockFiles();
void updatePlayBuffer();
void updatePlayBufferRealTimeSynth();
extern bool g_playing;
extern bool g_showCamerasInAnimMode;

HRESULT InitDirectInput( HWND hWnd );
HRESULT FreeDirectInput();
BackUp::Monostable BackUp::monostable;
Rect clientOffest;

long g_crossCursor = 0;
long g_currentType2 = 0;
long g_modellingCursor = 0;

long g_lockedOnPointCursor = 0;
long g_addPointCursor = 0;
long g_insertLineCursor = 0;
long g_deleteLineCursor = 0;
long g_mergePatchesCursor = 0;
long g_addGeometryCursor = 0;
long g_addGeometryMirrorCursor = 0;
long g_ringSelectCursor = 0;
long g_rectangleSelectPointCursor = 0;
long g_rectangleUnselectCursor = 0;
long g_rectangleSelectPolyCursor = 0;
long g_fillCursor = 0;
long g_brushCursor = 0;
long g_eyeDropperCursor = 0;
long g_knifeCursor = 0;
long g_addBoneCursor = 0;
long g_insertBoneCursor = 0;

long g_arrowCursor = 0;
long g_handCursor = 0;
long g_waitCursor = 0;

long g_sizeAllCursor = 0;
long g_sizeNWSECursor = 0;
long g_sizeNESWCursor = 0;
long g_sizeWECursor = 0;
long g_sizeNSCursor = 0;







extern bool g_dontCloseMenu;
bool g_createdINI = false;
bool outPutIndices = false;
extern int	g_mirrorVertex=0;
//BOOL g_localControl=false;
BackUp::SmlsSaved BackUp::smlsSaved;

double	g_timeBegan;
double g_time;
int g_viewIndex = 0;

extern float g_version;
extern bool initWinINet();
extern int g_numOfVisLines;

//void convertJoinPartToTEPart(IndexedFaceSet* part);
NodePtr g_before;
ArtPtr remPJ;
extern Ptr<Grid> g_grid;
extern Vec2i g_previousLeftClickPos;


//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------
//Vec2i g_3dwinPos;
//Vec2i g_3dwinSize=Vec2i(800,572-24*2-2);
//Vec2i g_sceneTreePos;
//Vec2i g_sceneTreeSize=Vec2i(400,0);
//Vec2i g_nodeWinPos=Vec2i(400, 570);


HWND hWndMain = NULL;
HWND g_hDlg = NULL;


//#define GET_WM_COMMAND_ID(wp, lp)               LOWORD(wp)
//#define GET_WM_COMMAND_HWND(wp, lp)             (HWND)(lp)
//#define GET_WM_COMMAND_CMD(wp, lp)              HIWORD(wp)
//BOOL CALLBACK selectedVertexProc(HWND hDlg,UINT iMessage, WPARAM wParam, LPARAM lParam);

bool windowSizeHasChanged();
void convertVisualCluesToRenderFormat();
void resetLooseEnds();
void addNodesForNewFile();
void initVisualClue();
void initRadialTug();

extern bool g_winAPI3DMode;

char defaultMsg[]={"message from thyme's code"};

tagTRACKMOUSEEVENT g_trackMouseEvent={sizeof(tagTRACKMOUSEEVENT), TME_LEAVE, 0, HOVER_DEFAULT};
//tagTRACKMOUSEEVENT g_trackMouseEvent={sizeof(tagTRACKMOUSEEVENT), 0, 0, HOVER_DEFAULT};

String g_seamlessDir;
const String& createSeamlessConstDirString(String& str){
	getFolderPath(str, CSIDL_MYDOCUMENTS);
	//getFolderPath(str, CSIDL_PROGRAM_FILES);
	int strPathLen=str.length;
	if(strPathLen>0&&str[strPathLen-1]=='\\'||strPathLen>0&&str[strPathLen-1]=='/'){
		str+="Seamless3d\\";
	}
	else{
		str+="\\Seamless3d\\";
	}
	return str;
}

const String g_appDataDir = createSeamlessConstDirString(g_seamlessDir) + "Application Data For Seamless3d\\";
//const String g_undoDir = g_appDataDir + "undoCache\\";
const String g_undoDir = g_appDataDir + "Recover\\";
const String g_demoDir = g_appDataDir + "For Robot Demos Only\\";
const String g_demoCacheDir = g_demoDir + "Cache\\";
const String g_setupPath = g_appDataDir + "seamless3dSetup.smls";
const String g_debugLogPath = g_appDataDir + "debugLog.txt";
const String g_screenshotDir = g_seamlessDir + "Screenshot\\";


void showCursor(bool show){
	static bool remVis = true;
	if(remVis != show){
		remVis = show;
		ShowCursor(show);
	}
	
}
void getFolderPath(String& str, unsigned id){
	TCHAR buf[MAX_PATH];
	SHGetSpecialFolderPath(
		NULL,
		buf,
		id,
		NULL
	);
//	str="";
	for(int i=0; buf[i]!=0; i++){
		str+=buf[i];
	}
}
void makeTitleStr(char* str){
	char tempStr[0x200];
	char drive[0x200];
	char dir[0x200];
	char ext[0x200];

	_splitpath(g_currentFilename.c_str(), drive, dir, str, ext );
	strcat(str, ext);

#ifdef OPENGL
	sprintf(tempStr," - Seamless3d %.3f OpenGL",g_version);
#elif TTS
	sprintf(tempStr," - Seamless3d TTS %.3f",g_version);
#else
	sprintf(tempStr," - Seamless3d %.3f",g_version);
//	sprintf(tempStr," - Seamless3d %.5f",g_version);
#endif


	if(!BackUp::smlsSaved){
		strcat(str," *");
	}
	strcat(str, tempStr);
	strcat(str," - ");
	strcat(str, g_selNode->name.c_str());
}
void setTitleName(){
	//selectPartFromVertex(g_seamless);
	char TempStr[0x200];
	makeTitleStr(TempStr);
	SetWindowText(hWndMain, TempStr);
	//SendMessage(g_hDlg, WM_USER, 0, 0);
}
void setSmlsSaved(){
	BackUp::smlsSaved.set();
	setTitleName();
}
void BackUp::Monostable::beginCycle(){
	if(beginBool==false){
		backedUpFileCntr++;
	}
	beginBool=true;
	setTitleName(); //this might as well be done here in case its been forgotten and cant do any significant harm slowing things down
	beginTime=g_time;
	g_selCount.reset();//this might as well be done here in case its been forgotten and cant do any significant harm slowing things down
}
bool BackUp::Monostable::state()const{
	return beginBool && (g_time>beginTime+g_settings->backupDelay);
}
inline void BackUp::checkForBackUpOperation(){
	if(mouse3dMode == false && DownButton::getDownButton() == 0){
		if(monostable.state()){
			backupToCache();
			monostable.clear();
		}
	}
}

int msG(String str1,String str2,String str3,unsigned dlgTypeAndIcon){

	String tempStr=String(str1);
	tempStr+=str2;
	tempStr+=str3;
	
	return MessageBox(hWndMain,tempStr.c_str(),defaultMsg,dlgTypeAndIcon);
}

int msG(String str,unsigned dlgTypeAndIcon){
	return MessageBox(hWndMain,str.c_str(),defaultMsg,dlgTypeAndIcon);
}

extern float g_remVersionLoaded;

void changeExtensionToSmls(String& str){
	for(int i=str.length; --i>=0;){
		if(str[i]=='.'){
			str = str.substr(0, i);
			break;
		}
		else if(str[i]=='\\' || str[i]=='/'){
			break;
		}
	}
	for(int i=2; ;i++){
		String newName;
		newName+=str;
		newName+=i;
		newName+=".smls";
		FILE* file;
		if((file=fopen(newName.c_str(), "rb"))==0){
			str=newName;
			return;
		}
		else{
			fclose(file);
		}
	}
}


void searchTillText(char*& p);

void openSceneTreeInterface(){

	g_stw->newNodeBar.dstPos=Vec2i(220, 25+16+8);
	g_numOfVisLines=(g_stw->size.y-SceneTree::beginYOffset)/pixSymbolCY;

	g_stw->size = Vec2i(g_settings->sceneTreeSize.x, g_settings->sceneTreeSize.y);
	if(g_stw->size.x < 10){
		g_stw->size = Vec2i(350, g_stw->size.y);
	}
	if(g_stw->size.y < 10){
		g_stw->size = Vec2i(g_stw->size.x, 400);
	}
	g_stw->pos=Vec2i(g_settings->sceneTreePos.x, g_settings->sceneTreePos.y);
	g_stw->pixBuf.length=g_stw->size.x*g_stw->size.y;


//	g_stw->paint();
}
/*
void roundToNearest(){
	ArtPtr part=g_selNode;
	SeamlessPtr smls=part->smls;
	M_Vec3f& point=smls->refPoint();
	Vec3f vec;
	vec=point[smls->ActCtrlPnt].multiply(100);
	vec.x=floor(vec.x+.5);
	vec.y=floor(vec.y+.5);
	vec.z=floor(vec.z+.5);
	point[smls->ActCtrlPnt]=vec.multiply(.01);
	
	smls->updateAllVertexInBuffers();
	makePointerEqualIndexLocation();
}
*/
void autoAddColor(ArtPtr part){
	if(part->color.length==0){
		M_Art& partList=part->smls->partList;
		for(int k=partList.length; --k>=0; ){
			Art* part=partList[k];
			part->color.length=part->coord.length;
		}
	}
}

void workOutVertexSameParts(int& newFrom, int& newTo, M_Int fromVertex, M_Int toVertex){
	bubbleSort(fromVertex, toVertex);

	for(int i=0; i<fromVertex.length; i++){
		if(fromVertex[i]<=newFrom){
			newFrom++;
		}
		if(fromVertex[i]<=newTo){
			newTo++;
		}
	}
}
void workOutVertexDiffParts(int& newFrom, M_Int fromVertex, M_Int toVertex){
	bubbleSort(fromVertex, toVertex);

	for(int i=0; i<fromVertex.length; i++){
		if(fromVertex[i]<=newFrom){
			newFrom++;
		}
	}
}

/*
void loopBackIFS(Seamless* smls){
	M_Build& build=smls->build.refValue();
	for(int k=build.length; --k>=0;){
		if(build[k]->getType()==STEM){
			Stem* stem=dynamic_cast<Stem*>(build[k].getRawPtr());
			M_Node& source=stem->source.refValue();
			M_SArt& part=stem->part.refValue();
			for(int i=source.length; --i>=0; ){
				if(i<part.length){
					Art* prt=dynamic_cast<Art*>(part[i].getRawPtr());
					IndexedFaceSet* ifs=dynamic_cast<IndexedFaceSet*>(source[i].getRawPtr());
					if(ifs!=0){
						int remPointSize=ifs->refPoint().length;
						int remColorSize=ifs->refColor().length;
						int remTriSize=ifs->refCoordIndex().length/4;
						//source[i]=dynamic_cast<Art*>(part[i].getRawPtr())->extractGeometry();
						ifs->copyNotClose(prt->extractGeometry());
						if(remPointSize!=ifs->refPoint().length){
							String str=prt->getName();
							str+=" coord diff ";
							str+=ifs->refPoint().length-remPointSize;
							dLog(str);
						}
						if(remColorSize!=ifs->refColor().length){
							String str=prt->getName();
							str+=" color diff ";
							str+=ifs->refColor().length-remColorSize;
							dLog(str);
						}
						if(remTriSize!=ifs->refCoordIndex().length/4){
							String str=prt->getName();
							str+=" triangle diff ";
							str+=ifs->refCoordIndex().length/4-remTriSize;
							dLog(str);
						}
					}
				}
			}
		}
	}
	smls->doBuild();
	smls->convertToRenderFormat();
	genNodeList();
	SendMessage(hSceneTreeWnd, WM_PAINT, 0, 0);
	makePointerEqualIndexLocation();
}
*/
bool checkVertexBelongsToTriangle(Art* vertexPart, Art* triPart, int vertexIndex, int &corner, int triangleIndex);
/*
void transferSingleVertex(){
	if(remPJ==NULL)return;
	Color3f scolor;
	ArtPtr part=g_selNode;
	SeamlessPtr &smls=part->smls;
	BackUp::monostable.beginCycle();
	int _actCtrlPnt=smls->ActCtrlPnt;
	M_Int &coordIndex=smls->geometry->coordIndex;

	Vec3f svec = smls->geometry->coord->point[_actCtrlPnt];
	svec=part->matrix*svec-remPJ->matrix*Vec3f();

	if(smls->geometry->color!=NULL){
		scolor=smls->geometry->color->color[_actCtrlPnt];
	}

	smls->genVertex(remPJ, svec, scolor);

	if(part->plidx>remPJ->plidx){
		_actCtrlPnt++;
	}

	//replace old vertex reference with the new vertex reference
	for(int i = -1; i<coordIndex.length-1 ; ){
		if(coordIndex[++i]==_actCtrlPnt){
			coordIndex[i]=remPJ->toCoord-1;
		}
		if(coordIndex[++i]==_actCtrlPnt){
			coordIndex[i]=remPJ->toCoord-1;
		}
		if(coordIndex[++i]==_actCtrlPnt){
			coordIndex[i]=remPJ->toCoord-1;
		}
		i++;
	}

	smls->deleteVertex(_actCtrlPnt); //delete old vertex

	smls->orderPolygonsFromVertices();

	smls->convertToRenderFormat();

	smls->ActCtrlPnt=remPJ->toCoord-1;
	makePointerEqualIndexLocation();
}
*/
/*
void joinCloseVertexImmediate(){
	ArtPtr part=g_selNode;
	SeamlessPtr &smls=part->smls;
	BackUp::monostable.beginCycle();

	JoinCloseVertex joinCloseVertex;
	joinCloseVertex.part=part;
	joinCloseVertex.joinRange=g_joinRange;
	joinCloseVertex.build();

	smls->convertToRenderFormat();

	makePointerEqualIndexLocation();
}
*/
/*
void reset3dCursor(){
	g_pointer->translation=Vec3f(0,0,0);
	g_pointer->rotation=Rotation();
	g_pointer->scale=Vec3f(1,1,1);
}
*/
bool _3DWin::click(const Vec2i ptCursor){
	return false;
}
void Window::checkIfAWindowHasBeenClickedAndUpdate(const Vec2i& mousePos){
	if(Window::activeWindow->hog){
		if(false == Window::activeWindow->click(mousePos)){
			PlaySound("SystemDefault", 0, SND_ASYNC);
		}
	}
	else{
		for(int i=visWinList.length; --i>=0;){
			if(visWinList[i]->click(mousePos)){
				return;
			}
		}
		activeWindow = g_3d.p;
		mouse2dSync();
		//enablePen();
		g_winAPI3DMode = true;
		setMouseCapture();
		g_viewDragMousePos = g_pointer.translation;
		SenseTugGlobals::initTug();
	}
}
inline void Window::checkIfAWindowHasBeenMidClickedAndUpdate(const Vec2i& mousePos){
	if(Window::activeWindow->hog){
		if(false == Window::activeWindow->midClick(mousePos)){
			PlaySound("SystemDefault", 0, SND_ASYNC);
		}
	}
	else{
		for(int i=visWinList.length; --i>=0;){
			if(visWinList[i]->midClick(mousePos)){
				return;
			}
		}
		activeWindow = g_3d.p;
		//enablePen();
		g_winAPI3DMode = true;
		setMouseCapture();
		g_viewPosDragMode = true;
		SenseTugGlobals::initMidTug();
	}
}

void setMouseCapture(){
	g_trackMouseEvent.hwndTrack = hWndMain;
	TrackMouseEvent(&g_trackMouseEvent);
	SetCapture(hWndMain);
}

inline void Window::checkIfAWindowHasBeenRightClickedAndUpdate(const Vec2i& mousePos){
	if(Window::activeWindow->hog){
		if(false == Window::activeWindow->rightClick(mousePos)){
			PlaySound("SystemDefault", 0, SND_ASYNC);
		}
	}
	else{
		for(int i=visWinList.length; --i>=0;){
			if(visWinList[i]->rightClick(mousePos)){
				return;
			}
		}
#ifdef PROG_ROB
	dLog("\nnew Move("+ String(g_mousePos.x) + ',' + String(g_mousePos.y) + "),");
	dLog("\nnew RightButtonDown(),");
#endif

		activeWindow = g_3d.p;
		//enablePen();
		g_winAPI3DMode = true;
		setMouseCapture();
		if((controlKeyState()) || InputEvent::l_ctrlDown){
			g_viewZRotDragMode = true;
		}
		else{
			g_viewRotDragMode = true;
		}
		g_beginMousePos = mousePos;
		g_beginViewRot = g_scene->rotation;
		mouse2dSync();
	}
}


inline void Window::checkIfMouseIsHoveringAWindowAndUpdate(const Vec2i& mousePos){
	if(g_remMenu == 0 && Window::activeWindow->hog){
		if(Window::activeWindow->hover(mousePos) == false){
			changeCursor(g_arrowCursor);
		}
	}
	else{
		for(int i = visWinList.length; --i>=0;){
			if(visWinList[i]->hover(mousePos)){
				return;
			}
		}
		if(false == g_panViewMode){
			//changeCursor(IDC_CROSS);
			//g_modellingCursor = MAKEINTRESOURCE(IDC_CROSS);
			//LPCTSTR lp = MAKEINTRESOURCE(LOCKED_ON_POINT_CURSOR);
			//LONG h = (LONG)LoadCursor(g_hinst, lp);
			//SetClassLong(hWndMain, GCL_HCURSOR, h);
			//(LONG)LoadCursor(0, requestType);
			changeCursor(g_modellingCursor);
		}
	}
}
/*
bool isDoubleLeftClick(){
	if(g_leftDoubleClicked == false && g_previousLeftClickPos == g_mousePos){

		double difference = g_currentClickTime - g_previousClickTime;
		double doubleClickTime = GetDoubleClickTime() * MILISECONDS_TO_SECONDS;

		//if(difference <= doubleClickTime){
		if(difference > doubleClickTime * .1 && difference <= doubleClickTime){
			return true;
		}
	}
	return false;
}
*/

bool isDoubleLeftClick(){
	if(g_leftDoubleClicked==false){
		if(abs(g_mousePos.x - g_previousLeftClickPos.x) <= 6 && abs(g_mousePos.y - g_previousLeftClickPos.y) <= 6){

			double difference = g_currentClickTime - g_previousClickTime;
			double doubleClickTime = GetDoubleClickTime() * MILISECONDS_TO_SECONDS;

			if(difference <= doubleClickTime){
				return true;
			}
		}
	}
	return false;
}


void hoverObject(){
	MultiSelect::resetDistance();
	SenseTugGlobals::node = 0;
	MultiSelect::withinSightsNode = 0;
	if(Viewpoint::withinSitesViewpoint){
		int a = 0;
	}
	Viewpoint::withinSitesViewpoint = 0;



	if(Viewpoint* selView = dynamic_cast<Viewpoint*>(g_selNode.getRaw())){
		if(selView->isOverPoint()){
			MultiSelect::withinSights = true;
			MultiSelect::withinSightsNode = selView;
			changeModellingCursor(g_lockedOnPointCursor);
			return;
		}
	}
	{

		if(MultiSelect::objectList.length && g_transformObj->isOverLever()){
			return;
		}

		g_scene->senseTug(Matrix(),Matrix(), 0);
		if(SenseTugGlobals::node){
			MultiSelect::withinSights = true;
			if(Art* part = dynamic_cast<Art*>(SenseTugGlobals::node)){
				MultiSelect::withinSightsNode = part->smls;
				g_transformObj->mode = TransformObj::SELECT;
			}
			else{
				MultiSelect::withinSightsNode =  SenseTugGlobals::node;
			}
			changeModellingCursor(g_handCursor);
		}
		else{
			changeModellingCursor(g_crossCursor);
		}
	}
}


void hoverCameraAnimMode(){
	MultiSelect::resetDistance();
	SenseTugGlobals::node = 0;
	MultiSelect::withinSightsNode = 0;
	if(Viewpoint::withinSitesViewpoint){
		int a = 0;
	}
	Viewpoint::withinSitesViewpoint = 0;



	{

		if(MultiSelect::objectList.length && g_transformObj->isOverLever()){ //dont think we need this
			return;
		}

		for(int k = g_viewpoints.length; -- k >= 0;){
			g_viewpoints[k]->senseTug(Matrix(),Matrix(), 0);
		}
	}
	if(SenseTugGlobals::node){
		MultiSelect::withinSights = true;
		MultiSelect::withinSightsNode =  SenseTugGlobals::node;
		changeModellingCursor(g_handCursor);
	}
	else{
		changeModellingCursor(g_crossCursor);
	}
}


void hoverWin3d(){
	if(g_rectangleSelectPointReady || g_rectangleUnselectPointReady || g_rectangleSelectPolyReady){
		return;
	}
	MultiSelect::withinSights = false;
	MultiSelect::withinSightsNode = 0;
	g_partCageClue->mode = 0;
	Window::checkIfMouseIsHoveringAWindowAndUpdate(g_mousePos);

	if(g_addBox.mousePhase == AddBox::READY_FOR_B || g_addBox.mousePhase == AddBox::READY_FOR_C){
		mouse2dSync();
		g_addBox.updateDimensions();
	}
	else if(g_addSphere.mousePhase == AddSphere::READY_FOR_RADIUS){
		mouse2dSync();
		g_addSphere.updateRadius();
	}
	else if(g_addLathe->mousePhase == AddLathe::SET_SPINDLE_ORIENTATION){
		mouse2dSync();
		if(g_addLathe->checkOrientationPoint(g_perspective)){
		}
	}
	else if(g_addLathe->mousePhase == AddLathe::INITIAL_POINTS){
		mouse2dSync();
		g_addLathe->checkOverInitPoint(g_perspective);
	}
	else if(g_addNurbsLine->mousePhase == AddNurbsLine::INITIAL_POINTS && g_addNurbsLine->line->point.length == 1){
		mouse2dSync();
		g_addNurbsLine->checkOverInitPoint(g_perspective);
	}
	else if(g_insertPatch.imousePhase == InsertPatch::NEUTRAL && g_deleteControlLine.mousePhase == DeleteControlLine::NEUTRAL && g_mergePatches.mousePhase == MergePatches::NEUTRAL){
		if(g_scene->mode & (MODELLING_MODE | ANIMATION_MODE | XRAY_EDIT_MODE | OWNERSHIP_MODE | NURBS_OWNERSHIP_MODE | SELECT_OBJECT_MODE) && tugFlag == false && g_modellingCursor == g_currentType2){
			if(g_scene->mode & SELECT_OBJECT_MODE){
				mouse2dSync();
				hoverObject();
				return;
			}
			else if(g_scene->mode & (XRAY_EDIT_MODE | OWNERSHIP_MODE | NURBS_OWNERSHIP_MODE)){
				if(Art* selPart = dynamic_cast<Art*>(g_selNode.getRaw())){
					mouse2dSync();
					selPart->smls->initSkeletonTug(g_perspective);
					return;
				}
			}
			else if(MultiSelect::pointList.length){
				mouse2dSync();
				if(g_transformCP->isOverLever()){
					//dLog("isOverLever()");
					return;
				}
			}
			if(Build* selBuild = dynamic_cast<Build*>(g_selNode.getRaw())){
				if(Seamless* smls = getSeamless2()){
					mouse2dSync();
					smls->initMultiCageTug(g_perspective);
				}
			}
			else if(g_scene->mode & ANIMATION_MODE){
				if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
					Seamless* smls = part->smls;
					mouse2dSync();
					g_partCageClue->mode = 0;

					if(g_vcArtAnimBase->initLeverTug(part, g_perspective)){
						g_partCageClue->mode = PartCageClue::BONE_ROTATION_LEVER;
						return;
					}
					else if(part->niki && part->niki->isOverPoint()){
						return;
					}
					else if(smls->initSkeletonTug(g_perspective)){
						return;
					}
					else{
						if(smls->animCP){
							smls->initMultiCageTug(g_perspective);
							if(dynamic_cast<NurbsPatch*>(MultiSelect::withinSightsNode.p)){
								g_partCageClue->mode = PartCageClue::MORPH_CP;
								return;
							}
						}
					}
				}
				else if(Viewpoint* selView = dynamic_cast<Viewpoint*>(g_selNode.getRaw())){
					mouse2dSync();
					if(selView->isOverPoint()){
						MultiSelect::withinSights = true;
						MultiSelect::withinSightsNode = selView;
						changeModellingCursor(g_lockedOnPointCursor);
						return;
					}
				}
				else if(NurbsTrack* nurbsTrack = dynamic_cast<NurbsTrack*>(g_selNode.getRaw())){
					mouse2dSync();
					if(nurbsTrack->isOverPoint()){
						MultiSelect::withinSights = true;
						MultiSelect::withinSightsNode = nurbsTrack;
						changeModellingCursor(g_lockedOnPointCursor);
						return;
					}
				}
				else if(NikInterpolator* niki = dynamic_cast<NikInterpolator*>(g_selNode.getRaw())){
					mouse2dSync();
					if(niki->isOverPoint()){
						MultiSelect::withinSights = true;
						MultiSelect::withinSightsNode = niki;
						changeModellingCursor(g_lockedOnPointCursor);
						return;
					}
				}
				if(g_showCamerasInAnimMode){
					hoverCameraAnimMode();
				}
				else{
					changeModellingCursor(g_crossCursor);
				}
			}
		}
	}
	//dLog("false");
}


BOOL mouse2dHandler(UINT msg,WPARAM wParam, LPARAM lParam){
	static POINT ptDownMousePos2d;
    switch( msg )
    {
	case WM_RBUTTONDOWN:{
		Button::repeatCounter=0;
		removeTypeFromList<Button>(g_callBackList);
		Window::removeToolTip();
		Window::removeMenu();

		g_mousePos = Vec2i(LOWORD(lParam), HIWORD(lParam));
		g_prevMousePos = g_mousePos;
		g_newMousePos = g_mousePos;


		Window::checkIfAWindowHasBeenRightClickedAndUpdate(g_mousePos);
		return TRUE;
	}
	case WM_MBUTTONDOWN:{
		Button::repeatCounter = 0;
		removeTypeFromList<Button>(g_callBackList);
		Window::removeToolTip();
		Window::removeMenu();

		g_mousePos = Vec2i(LOWORD(lParam), HIWORD(lParam));
		g_prevMousePos = g_mousePos;
		g_beginMousePos = g_mousePos;
		g_newMousePos = g_mousePos;
		g_beginViewPos = g_scene->translation;

		Window::checkIfAWindowHasBeenMidClickedAndUpdate(g_mousePos);
		return TRUE;
	}
	case WM_LBUTTONDOWN:{
		Button::repeatCounter = 0;
		g_previousClickTime = g_currentClickTime;
		g_currentClickTime = g_time;

		Window::removeToolTip();

		g_mousePos = Vec2i(LOWORD(lParam), HIWORD(lParam));
		g_prevMousePos = g_mousePos;
		g_beginMousePos = g_mousePos;
		g_newMousePos = g_mousePos;
		g_beginViewPos = g_scene->translation;

		//g_leftDoubleClicked = isDoubleLeftClick();
	//	dLog(String("g_leftDoubleClicked = " + String(g_leftDoubleClicked)));

		if(Button* button = Button::getHoverButton()){

			DownButton::setDownButton(button);

			button->win.onClickButton();
			Window::activeWindow = &button->win;
			button->action();
			//button->win.paint();//do we need this?
			if(g_dontCloseMenu == false){
				Window::removeMenu();
			}
		}
		else{
			//if(g_insertPatch.imousePhase == InsertPatch::LINE_MODE){
			//	g_insertPatch.imousePhase = InsertPatch::DIVERT_STANDBY;
			//	g_stw->cp->setupNodeWindow(g_insertComplete);
			//}
			//else if(g_deleteControlLine.mousePhase == 2){
			//	g_deleteControlLine.doLastPhase();
			//}
			//else{
				Window::removeMenu();
				Window::checkIfAWindowHasBeenClickedAndUpdate(g_mousePos);
			//}
		}


		g_dontCloseMenu = false;

		if(g_drag){
			g_trackMouseEvent.hwndTrack = hWndMain;
			TrackMouseEvent(&g_trackMouseEvent);
			SetCapture(hWndMain);
		}

		g_previousLeftClickPos = g_mousePos;

		return TRUE;
	}
//	case WM_LBUTTONDBLCLK:{ //why does this not catch double click?
//		break;
//	}
	case WM_TOUCH:{
		break;
	}
	case WM_GESTURENOTIFY:{
        GESTURECONFIG gc = {0,GC_ALLGESTURES,0};
        BOOL bResult = SetGestureConfig(hWndMain,0,1,&gc,sizeof(GESTURECONFIG));
            
        if(!bResult)
        {
            // an error
        }
        return DefWindowProc(hWndMain, WM_GESTURENOTIFY, wParam, lParam);
    }
	case WM_GESTURE:{
		int a = 0;
		break;
	}
	case WM_MBUTTONUP:{
		if(g_winAPI3DMode){
			release3dDrag();
		}
		break;
	}
	case WM_RBUTTONUP:{

		if(g_winAPI3DMode){
			release3dRightDrag();
		}
		break;
	}
	case WM_LBUTTONUP:{
		g_mousePos = Vec2i(LOWORD(lParam), HIWORD(lParam));

		if(g_drag){
			g_drag->endDrag(g_mousePos);
		}
		else if(Button* button = DownButton::getDownButton()){
			button->onButtonUp();
		}
		else{
			Window::activeWindow->release(g_mousePos);
		}
#ifdef PROG_ROB
		InputEvent::sceneTreeScrolPos=0;
		if(g_upStr!=""){
			dLog(g_upStr);
		}
		g_upStr="";

		InputEvent::sceneTreeScrolPos=g_stw->getScrollPosCollapsed();
#endif

		if(g_winAPI3DMode){
			release3dDrag();
		}



		return TRUE;
	}

	//case WM_NCMOUSEMOVE:
	case WM_MOUSEMOVE:{
		//if(g_useWinTab && g_penEnabled){
		//	return TRUE;
		//}
		if(g_mergePatches.mousePhase != 0){
			int a = 0;
		}
		short lw = LOWORD(lParam);
		short hw = HIWORD(lParam);
		//if(lw < 0 || lw >= g_3d->size.x || hw < 0 || hw >= g_3d->size.y){
		//	return TRUE;
		//}
		//if(lw < -100){
		//	int a = 0;
		//}
		int remOldMousePos = g_mousePos.y;
		if(g_mousePos.x == lw && g_mousePos.y == hw){
			return TRUE;
		}

		if(g_scene->mode & PAINT_MODE){
			if(g_scene->mode & BRUSH_MODE){
				g_brush->houseKeep();
			}
		}

		g_mousePos = Vec2i(lw, hw);
		g_newMousePos = g_mousePos;

		if(g_winAPI3DMode){
			return TRUE;
		}

		//if(g_winAPI3DMode){
			//dLog("WM_MOUSEMOVE");
			//moveMouse3d();
			//break;
		//}

		Button* hoverButton = Button::getHoverButton();
		MultiSelect::resetDistance();
		//if(g_insertPatch.imousePhase == LINE_MODE){
		//	mouse2dSync(hWndMain);
		//	//g_insertPatch.mouseOverPatch->getRatioPhaseForInsert();
		//	g_insertPatch.addInsertVisualClue();
		//}
		if(g_insertPatch.imousePhase == InsertPatch::MONITOR_DIRECTION){
			mouse2dSync();
			if(g_insertPatch.monitorDirection()){
				g_insertPatch.m_insert_tempDivertBackup = g_insertPatch.m_insert;
				g_insertPatch.m_insertOD_tempDivertBackup = g_insertPatch.m_insertOD;
				g_insertPatch.remLeftSide_tempDivertBackup = g_insertPatch.remLeftSide;

				g_insertPatch.beginInsertIdx_tempDivertBackup = g_insertPatch.beginInsertIdx;
				g_insertPatch.beginInsertIdxOD_tempDivertBackup = g_insertPatch.beginInsertIdxOD;
				g_insertPatch.imousePhase = InsertPatch::DIVERT_STANDBY;
			}
			return FALSE;
		}
		if(g_insertPatch.imousePhase == InsertPatch::STANDBY || g_insertPatch.imousePhase == InsertPatch::LINE_MODE || g_insertPatch.imousePhase == InsertPatch::DIVERT_STANDBY || g_addBone->mousePhase != AddBone::NEUTRAL){
			if(hoverButton){
				hoverButton->onHoverMove(g_mousePos - hoverButton->win.pos);
			}
			else{
				mouse2dSync();
				//SenseTugGlobals::initTug();
				if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
					Seamless* smls = patch->smls;
					if(g_insertPatch.imousePhase == InsertPatch::DIVERT_STANDBY){
						if(smls->addDiversionMark()){
							//g_insertPatch.imousePhase = InsertPatch::TEMP_DIVERT;
						}
					}
					else{
						smls->insertControlRowColum();
					}
				}
				else if(g_addBone->mousePhase != AddBone::NEUTRAL){
					if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
						mouse2dSync();
						g_addBone->initMouseOverLine(part->smls);
					}
				}
			}
		}
		if(g_mergePatches.mousePhase == MergePatches::STANDBY){
			if(hoverButton){
				hoverButton->onHoverMove(g_mousePos - hoverButton->win.pos);
			}
			else{
				mouse2dSync();
				g_mergePatches.smls->mergePatchesStandby();
			}
		}
		if(g_deleteControlLine.mousePhase == 1 || g_deleteControlLine.mousePhase == 2){
			if(hoverButton){
				hoverButton->onHoverMove(g_mousePos - hoverButton->win.pos);
			}
			else{
				mouse2dSync();
				if(g_scene->mode & ANIMATION_MODE){
					if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
						part->smls->deleteControlLine();
					}
				}
				else if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
					patch->smls->deleteControlLine();
				}
			}
		}
		else if(g_drag){
			if(g_mousePos.x >= 0){
				if(g_drag->triggered == false){
					if(abs(g_mousePos.x - g_drag->beginMousePos.x) > 6 || abs(g_mousePos.y - g_drag->beginMousePos.y) > 6){
						g_drag->triggered = true;
					}
				}
				if(g_drag->triggered){
					g_drag->doDrag(g_mousePos);
				}
			}
		}
		else if(hoverButton && hoverButton->onHoverMove(g_mousePos - hoverButton->win.pos)){
		}
		else{
			hoverWin3d();
		}
		break;
	}
		/*
	case WM_MOUSEMOVE:
		if(MK_LBUTTON&wParam){
			
			
			GetCursorPos( &ptCursor );
			g_scene->translation=g_scene->rotation.multVec(g_scene->translation);
			g_scene->translation.x+=.0025*(float)(ptCursor.x-ptDownMousePos2d.x);
			g_scene->translation=-g_scene->rotation*g_scene->translation;
			
			if(controlKeyState()){
				g_scene->translation=g_scene->rotation.multVec(g_scene->translation);
				g_scene->translation.z-=.0025*(float)(ptCursor.y-ptDownMousePos2d.y);
				g_scene->translation=-g_scene->rotation*g_scene->translation;
			}
			else{
				g_scene->translation=g_scene->rotation*g_scene->translation;
				g_scene->translation.y-=.0025*(float)(ptCursor.y-ptDownMousePos2d.y);
				g_scene->translation=-g_scene->rotation*g_scene->translation;
			}
			
			GetCursorPos( &ptDownMousePos2d );
			return TRUE;
		}
		if(MK_RBUTTON&wParam){
			GetCursorPos( &ptCursor );
			rotateScene(ptCursor.x-ptDownMousePos2d.x, ptCursor.y-ptDownMousePos2d.y);
			ptDownMousePos2d=ptCursor;
			return TRUE;
		}
		*/
	}
	return FALSE;
}

void updateNavKey();

bool keyDown(HWND &hWnd, UINT &message, WPARAM &wParam, LPARAM &lParam){
	if(Window::activeWindow == g_3d){
		static String str;
		float tran=(1/g_scene->scale.x)*.01;
		switch (wParam){
			case VK_LEFT:{
				if(g_leftKeyDown == false){
					g_leftKeyTime = g_time;
					g_leftKeyDown = true;
					updateNavKey();
				}
				else{
					g_leftKeyDown = true;
				}
				break;
			}
			case VK_RIGHT:{
				if(g_rightKeyDown == false){
					g_rightKeyTime = g_time;
					g_rightKeyDown = true;
					updateNavKey();
				}
				else{
					g_rightKeyDown=true;
				}
				break;
			}
			case VK_DOWN:{
				if(g_backwardKeyDown==false){
					g_backwardKeyTime=g_time;
					g_backwardKeyDown=true;
					updateNavKey();
				}
				else{
					g_backwardKeyDown=true;
				}
				break;
			}
			case VK_UP:{
				if(g_forwardKeyDown==false){
					g_forwardKeyTime=g_time;
					g_forwardKeyDown=true;
					updateNavKey();
				}
				else{
					g_forwardKeyDown=true;
				}
				break;
			}
			default:{
				return false;
			}
		}
		return true;
	}
	return false;
}

bool getOldColor(Color3f& oldColor){
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	if(part==0)return false;
	Seamless* smls=part->smls;
	Art* triPart=smls->actTriPart;
	if(triPart==0)return false;
	ColorEffect* colorEffect=dynamic_cast<ColorEffect*>(part->workEffect);
	if(colorEffect==0)return false;
	if(triPart->triangle.length <= smls->ActTri){
		return false;
	}
	M_Triangle& triangle=triPart->triangle;
	const M_Color3f& color = colorEffect->color;
	oldColor=color[triangle.refTexIndex(smls->ActTri*3+smls->ActTriCorner)];
	return true;
}
//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: The window's message handler
//-----------------------------------------------------------------------------

void keyboardRadialTug(Vec3f coord, Vec3f pt){
	Art* part = dynamic_cast<Art*>(g_selNode.getRaw());
	Seamless* smls=part->smls;
	M_Art& partList=smls->partList;

	g_radialTug->displacement.refValue()[0]=pt;

	g_radialTug->slope.refValue() = 1;
	if(g_settings->multiPartTug.refValue()){
		for(int k=partList.length; --k>=0;){
			Art* prt=partList[k];
			g_radialTugStem->part.refValue()[0]=prt;
			g_radialTug->location.refValue()[0]=part->matrix * (prt->imatrix * coord);
			g_radialTugStem->doBuild();
		}
	}
	else{
		g_radialTugStem->part.refValue()[0]=part;
		g_radialTug->location.refValue()[0]=coord;
		g_radialTugStem->doBuild();
	}
}

void openChatLog(){
	String url = g_appDataDir;
	url+="chat.txt";

	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory( &si, sizeof(si) );
	ZeroMemory( &pi, sizeof(pi) );
	si.cb = sizeof(si);
	// Start the child process.

	String commandLine='\"';
	getFolderPath(commandLine, CSIDL_WINDOWS);
	if(commandLine[commandLine.length-1]!='\\'){
		commandLine+='\\';
	}
	commandLine+="notepad.EXE\" ";
	commandLine+=url;

	if(CreateProcess(
		NULL, // No module name (use command line).
		const_cast<char*>(commandLine.c_str()), // Command line. 
		NULL,             // Process handle not inheritable. 
		NULL,             // Thread handle not inheritable. 
		FALSE,            // Set handle inheritance to FALSE. 
		0,                // No creation flags. 
		NULL,             // Use parent's environment block. 
		NULL,             // Use parent's starting directory. 
		&si,              // Pointer to STARTUPINFO structure.
		&pi )){             // Pointer to PROCESS_INFORMATION structure.    ) 
		
		//WaitForSingleObject( pi.hProcess, INFINITE );
		// Close process and thread handles.
		CloseHandle( pi.hProcess );
		CloseHandle( pi.hThread );
	}
}


void openWithHtmlBrowser(String url, const String htmlBrowser){
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory( &si, sizeof(si) );
	ZeroMemory( &pi, sizeof(pi) );
	si.cb = sizeof(si);
	// Start the child process.

	String commandLine='\"';
	if(htmlBrowser == ""){
		getFolderPath(commandLine, CSIDL_PROGRAM_FILES);
		if(commandLine[commandLine.length-1]!='\\'){
			commandLine+='\\';
		}
		commandLine+="Internet Explorer\\IEXPLORE.EXE\"";
	}
	else{
		commandLine = htmlBrowser;
	}
	commandLine += ' ';
	commandLine += url;

	if( CreateProcess(
		NULL, // No module name (use command line).
		const_cast<char*>(commandLine.c_str()), // Command line. 
		NULL,             // Process handle not inheritable. 
		NULL,             // Thread handle not inheritable. 
		FALSE,            // Set handle inheritance to FALSE. 
		0,                // No creation flags. 
		NULL,             // Use parent's environment block. 
		NULL,             // Use parent's starting directory. 
		&si,              // Pointer to STARTUPINFO structure.
		&pi )){             // Pointer to PROCESS_INFORMATION structure.    ) 
		
		//WaitForSingleObject( pi.hProcess, INFINITE );
		// Close process and thread handles.
		CloseHandle( pi.hProcess );
		CloseHandle( pi.hThread );
	}
}
void getFFmpegPath(String& path){
    char buffer[MAX_PATH];
	GetModuleFileName(NULL, buffer, MAX_PATH);
	path = buffer;

	String drive;
	String dir;
	String fname;
	String ext;
	splitPath(path, drive, dir, fname, ext);
	path = '\"';
	path += drive;
	path += dir;
	path += "ffmpeg\\ffmpeg.exe\"";
}
void openWithFFmpeg(String url){
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory( &si, sizeof(si));
	ZeroMemory( &pi, sizeof(pi));
	si.cb = sizeof(si);
	// Start the child process.
	String commandLine;
	getFFmpegPath(commandLine);


//	String commandLine='\"';
//	getFolderPath(commandLine, CSIDL_PROGRAM_FILES);
//	if(commandLine[commandLine.length-1]!='\\'){
//		commandLine+='\\';
//	}
//	commandLine+="Seamless3d\\ffmpeg.exe\"";
	commandLine+=url;
	//commandLine+='\"';

	if( CreateProcess(
		NULL, // No module name (use command line).
		const_cast<char*>(commandLine.c_str()), // Command line. 
		NULL,             // Process handle not inheritable. 
		NULL,             // Thread handle not inheritable. 
		FALSE,            // Set handle inheritance to FALSE. 
		0,                // No creation flags. 
		NULL,             // Use parent's environment block. 
		NULL,             // Use parent's starting directory. 
		&si,              // Pointer to STARTUPINFO structure.
		&pi )){             // Pointer to PROCESS_INFORMATION structure.    ) 
		
		//WaitForSingleObject( pi.hProcess, INFINITE );
		// Close process and thread handles.
		CloseHandle( pi.hProcess );
		CloseHandle( pi.hThread );
	}
}

void openWithWindowsMediaPlayer(String url){
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory( &si, sizeof(si) );
	ZeroMemory( &pi, sizeof(pi) );
	si.cb = sizeof(si);
	// Start the child process.

	String commandLine='\"';
	getFolderPath(commandLine, CSIDL_PROGRAM_FILES);
	if(commandLine[commandLine.length-1]!='\\'){
		commandLine+='\\';
	}
	commandLine+="Windows Media Player\\wmplayer.exe\" \"";
	commandLine+=url;
	commandLine+='\"';

	if( CreateProcess(
		NULL, // No module name (use command line).
		const_cast<char*>(commandLine.c_str()), // Command line. 
		NULL,             // Process handle not inheritable. 
		NULL,             // Thread handle not inheritable. 
		FALSE,            // Set handle inheritance to FALSE. 
		0,                // No creation flags. 
		NULL,             // Use parent's environment block. 
		NULL,             // Use parent's starting directory. 
		&si,              // Pointer to STARTUPINFO structure.
		&pi )){             // Pointer to PROCESS_INFORMATION structure.    ) 
		
		//WaitForSingleObject( pi.hProcess, INFINITE );
		// Close process and thread handles.
		CloseHandle( pi.hProcess );
		CloseHandle( pi.hThread );
	}
}
void openWithPovRay(String url){
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory( &si, sizeof(si) );
	ZeroMemory( &pi, sizeof(pi) );
	si.cb = sizeof(si);
	// Start the child process.

	String commandLine='\"';
	getFolderPath(commandLine, CSIDL_PROGRAM_FILES);
	if(commandLine[commandLine.length-1]!='\\'){
		commandLine+='\\';
	}
	commandLine+="POV-Ray for Windows v3.62\\bin\\pvengine.exe\" /EDIT \"";
	commandLine+=url;
	commandLine+='\"';

	if( CreateProcess(
		NULL, // No module name (use command line).
		const_cast<char*>(commandLine.c_str()), // Command line. 
		NULL,             // Process handle not inheritable. 
		NULL,             // Thread handle not inheritable. 
		FALSE,            // Set handle inheritance to FALSE. 
		0,                // No creation flags. 
		NULL,             // Use parent's environment block. 
		NULL,             // Use parent's starting directory. 
		&si,              // Pointer to STARTUPINFO structure.
		&pi )){             // Pointer to PROCESS_INFORMATION structure.    ) 
		
		CloseHandle( pi.hProcess );
		CloseHandle( pi.hThread );
	}
}

void returnToExitSeamless3d(){
	g_settings->writeINIFile();
	BackUp::deleteBackupFiles();
	DestroyWindow(hWndMain);
}

void riProcessMouseMessage( const RAWMOUSE *rmouse )
{
    // Store Mouse Button States
    // Check Mouse Position Relative Motion
    if ( MOUSE_MOVE_RELATIVE == rmouse->usFlags )   
    {
		int mxr = rmouse->lLastX ;      
		int myr = rmouse->lLastY ;
		g_mousePos.x += mxr;
		g_mousePos.y += myr;
		//dLog(g_mousePos.x);
    }
}

#ifdef WACOM_FILES
extern PACKET* g_wintabPktBuf;
#endif
extern int g_lastPacketIndex;
int g_packetBufLen = 0x200;


void initializeShortCuts();
void getMemberAddresses();
void convert_raw_bitmap_to_cpp();
void convert_raw_alpha_bitmap_to_cpp();

void genButtons();

#include <crtdbg.h>
//static _CrtMemState s1, s2, s3;

//void processWords(int);
//void printOutLines();


void saveMemToFiles(unsigned int width, unsigned int height);

//char nibleToHex(char par);


// Test.cpp
#include <iostream>

class Test
{
  public:
    template <typename T>
    void function () {
      std::cout << data << std::endl;
    }

    template <> void function<char>();
};



template <>
void Test::function<char> () {
  std::cout << 'Z';
}


extern "C" void CopyLocalToScratchPad_doOp_S_Float();

void ExportMovie::OpenOutputFolder::operator=(const bool& par){
	openWithHtmlBrowser(g_seamlessDir, "");
}

void ExportVrmlX3d::OpenOutputFolder::operator=(const bool& par){
	openWithHtmlBrowser(g_seamlessDir, "");
}
/*
inline void convertConcaveIFSIndexToTriIndex(const M_Vec3f& src_point, M_Int& index){
	M_Vec3f point;
	M_Int& tmpIndex=g_tmpIndex;
	tmpIndex.length=0;
	M_Int poly;
	for(int i = 0; i < index.length && index[i] >= 0; i++){
		poly+=index[i];
	}

	Rotation rot;

	for(int s = 0 ;;){
		int idx0 = poly[s++];
		if(s>=poly.length){
			s=0;
		}
		int idx1 = poly[s++];
		if(s>=poly.length){
			s=0;
		}
		int idx2 = poly[s++];
		if(s>=poly.length){
			s=0;
		}

		Vec3f v0 = src_point[idx0];
		Vec3f v1 = src_point[idx1];
		Vec3f v2 = src_point[idx2];

		if(thinTriangle(v0, v1, v2)){
			s -= 2;
			continue;
		}
		Vec3f normal = calcNormal(src_point[index[idx0]], src_point[index[idx1]], src_point[index[idx2]]);
		rot = Rotation(Vec3f(0, 0, 1), normal);
		break;
	}

	point.length = src_point.length;
	for(int i=0; i<point.length; i++){
		point[i] = rot * src_point[i];
	}

	int s=0;
	int c=0;
	while(poly.length>3){
		int idx0 = poly[s++];
		if(s>=poly.length){
			s=0;
			c++;
		}
		int idx1 = poly[s++];
		if(s>=poly.length){
			s=0;
			c++;
		}
		int idx2 = poly[s++];
		if(s>=poly.length){
			s=0;
			c++;
		}

		Vec3f v0 = point[idx0];
		Vec3f v1 = point[idx1];
		Vec3f v2 = point[idx2];

		if(!backFace(v0, v1, v2) || thinTriangle(v0, v1, v2) || vertexInsideTri(point, idx0, idx1, idx2)){
			if(c==3){
				s=0;
				c=0;
				rot=-rot;
				for(int i=0; i<point.length; i++){
					point[i] = rot * src_point[i];
				}
				poly.length=0;
				for(int i = 0; i < index.length && index[i] >= 0; i++){
					poly+=index[i];
				}
				tmpIndex.length=0;
				continue;
			}
			s -= 2;
			if(s < 0){
				s=poly.length + s;
			}
			continue;
		}
		tmpIndex+=idx0;
		tmpIndex+=idx1;
		tmpIndex+=idx2;
		tmpIndex+=-1;

		int g=s-2;
		if(g < 0){
			g=poly.length + g;
		}
		poly.deleteElement(g);
		s=0;
		c=0;
	}
	tmpIndex+=poly;
	index=tmpIndex;
}
*/

void getClientRect(Rect& rect){
	GetClientRect(hWndMain, reinterpret_cast<RECT*>(&rect));
}

void getCursorPos(Point& point){
	GetCursorPos(reinterpret_cast<POINT*>(&point));
}

void pivotPointToClipBoard(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if (OpenClipboard(hWndMain)){
			EmptyClipboard(); 

			// Allocate a global memory object for the text. 
		
			String str;
			part->smls->catIMatrix();
			(part->matrix * part->absolutePivotPoint).toString(str);

			HGLOBAL hglbCopy = GlobalAlloc(GMEM_MOVEABLE, str.length+1);
			if (hglbCopy == NULL) 
			{ 
				CloseClipboard(); 
			}
			else{
		
				// Lock the handle and copy the text to the buffer. 
			
				LPVOID lptstrCopy = GlobalLock(hglbCopy); 
				memcpy(lptstrCopy, str.c_str(), str.length+1); 

				GlobalUnlock(hglbCopy); 
			
				// Place the handle on the clipboard. 
			
				SetClipboardData(CF_TEXT, hglbCopy);
				CloseClipboard();
			}
		}
	}
}

void changeCursor(long requestType){
	if(g_currentType2 != requestType){
		static int g = 0;
		//if(g > 0){
		//	g = g;
		//}
		if(g_currentType2 == g_lockedOnPointCursor){
		//if(requestType == g_addPointCursor){
		//if(g_currentType2 == g_addPointCursor && requestType != g_lockedOnPointCursor){
			int g = 0;
		}

		SetClassLong(hWndMain, GCL_HCURSOR, (LONG)requestType);
		SetCursor((HCURSOR)requestType);

		g_currentType2 = requestType;
		if(requestType != g_sizeAllCursor){
			g_panViewMode = false;
		}
	}
}

float InputEvent::localMove(Vec2i endPos){
	ClientToScreen(hWndMain, reinterpret_cast<POINT*>(&endPos));
	return move(endPos);
}

void CaretWindow::copyToClipboard(bool cut){
	if(highlightPos<0)return;
	if(	caretPos!=highlightPos){
		if (OpenClipboard(hWndMain)){
			EmptyClipboard(); 
	
			// Allocate a global memory object for the text. 
		
			int highlightLen=abs(caretPos-highlightPos);
			int index=highlightPos;
			if(caretPos<highlightPos){
				index=caretPos;
			}

			HGLOBAL hglbCopy = GlobalAlloc(GMEM_MOVEABLE, abs(caretPos-highlightPos) + 1);
			if (hglbCopy == NULL) 
			{ 
				CloseClipboard(); 
			}
			else{
		
				// Lock the handle and copy the text to the buffer. 
			
				LPVOID lptstrCopy = GlobalLock(hglbCopy); 
				memcpy(lptstrCopy, &cellStr[index], highlightLen + 1); 

				((char*)lptstrCopy)[highlightLen] = 0;    // null character 
				GlobalUnlock(hglbCopy); 
			
				// Place the handle on the clipboard. 
			
				SetClipboardData(CF_TEXT, hglbCopy);
				CloseClipboard();

				if(cut){
					String str = cellStr;
					cellStr = str.substr(0, index);
					cellStr += str.substr(index + highlightLen);
					caretPos = highlightPos =index;
					onChange();
					win.paint();
				}
			}
		} 
	}
}
void CaretWindow::pasteFromClipboard(){
	if (!IsClipboardFormatAvailable(CF_TEXT)) 
		return; 
	if (!OpenClipboard(hWndMain)) 
		return; 

	HGLOBAL hglb = GetClipboardData(CF_TEXT); 
	if (hglb != NULL) 
	{ 
		LPTSTR lptstr = (LPTSTR)GlobalLock(hglb); 
		if (lptstr != NULL) 
		{ 
			// Call the application-defined ReplaceSelection 
			// function to insert the text and repaint the 
			// window. 
			deleteHighlight();

			int remLen=cellStr.length;
			cellStr=cellStr.replace(caretPos, 0, lptstr);
			
			GlobalUnlock(hglb); 
			caretPos+=cellStr.length-remLen;
			highlightPos=caretPos;

		} 
	} 
	CloseClipboard();

	updateCaret();
	onChange();
	win.paint();
}

bool tabKeyState(){
	if(0x8000 & GetKeyState(VK_TAB)){
		return true;
	}
	else return false;
}

bool enterKeyState(){
	if(0x8000 & GetKeyState(VK_RETURN)){
		return true;
	}
	else return false;
}

bool controlKeyState(){
	if(0x8000 & GetKeyState(VK_CONTROL)){
		return true;
	}
	else return false;
}
bool altKeyState(){
	if(0x8000 & GetKeyState(VK_MENU)){
		return true;
	}
	else return false;
}

bool shiftKeyState(){
	if(0x8000 & GetKeyState(VK_SHIFT)){
		return true;
	}
	else return false;
}

bool zKeyState(){
	if(0x8000 & GetKeyState('Z')){
		return true;
	}
	else return false;
}

bool aKeyState(){
	if(0x8000 & GetKeyState('A')){
		return true;
	}
	else return false;
}

bool sKeyState(){
	if(0x8000 & GetKeyState('S')){
		return true;
	}
	else return false;
}

bool dKeyState(){
	if(0x8000 & GetKeyState('D')){
		return true;
	}
	else return false;
}

bool r_controlKeyState(){
	if(0x8000 & GetKeyState(VK_RCONTROL)){
		return true;
	}
	else return false;
}
bool r_altKeyState(){
	if(0x8000 & GetKeyState(VK_RMENU)){
		return true;
	}
	else return false;
}
bool r_shiftKeyState(){
	if(0x8000 & GetKeyState(VK_RSHIFT)){
		return true;
	}
	else return false;
}
bool deleteKeyState(){
	if(0x8000 & GetKeyState(VK_DELETE)){
		return true;
	}
	else return false;
}

void BackUp::deleteBackupFiles(){
	String filename;
	_close(g_lockFile);
	g_lockFile = 0;

	filename = g_undoDir;
	filename += ourBackUpName;
	filename += ".lock";
	DeleteFile(filename.c_str());

	BOOL noError = FALSE;
	int i = 0;
	do{
		filename = g_undoDir;
		filename += ourBackUpName;
		filename += i++;
		filename += ".smls";
		noError = DeleteFile(filename.c_str());
	}while (noError);
	backedUpFileCntr = 0;
}


void BackUp::createUniqueBackupName(bool copy){
	String filename;
	int i = 0;
	while(1){
		ourBackUpName = i++;
		ourBackUpName += '_';
		filename = g_undoDir;
		filename += ourBackUpName;
		filename += "0.smls";
		if(FILE* file = fopen(filename.c_str(), "r+b")){
			fclose(file);
		}
		else{
			String lockName = g_undoDir;
			lockName += ourBackUpName;
			lockName += ".lock";
			if(g_lockFile > 0){
				_close(g_lockFile);
			}
			g_lockFile = 0;
			SetFileAttributes(lockName.c_str(), FILE_ATTRIBUTE_NORMAL);
			if(_sopen_s(&g_lockFile, lockName.c_str(), _O_RDWR | _O_CREAT | _O_EXCL, _SH_DENYRW, _S_IREAD | _S_IWRITE) == 0){
				//const String& file0 = g_settings->file.refValue()[0];
				_write(g_lockFile, g_currentFilename.c_str(), g_currentFilename.length);
				backupToCache(copy);
				break;
			}
			else{
				g_lockFile = 0;
			}
		}
	}
}

void BackUp::backupToCache(bool copy){
	static String filename;

	//changeCursor(IDC_WAIT);

	filename = g_undoDir;
	filename += ourBackUpName;
	filename += backedUpFileCntr;
	filename += ".smls";
	SetFileAttributes(filename.c_str(), FILE_ATTRIBUTE_NORMAL);
	if(copy && g_settings->file.length){
		CopyFile(g_settings->file.refValue()[0].c_str(), filename.c_str(), FALSE);
	}
	else{
		writeSmlsFile(filename);
	}
	backedUpFileCntrEnd = backedUpFileCntr;

	if(copy){
		smlsSaved.set();
	}
	//setTitleName();
}
void splitPath(String wholePath, String& drive, String& dir, String& fname, String& ext){
    char _dir[_MAX_DIR];
    char _fname[_MAX_FNAME];
    char _ext[_MAX_EXT];
    char _drive[_MAX_DRIVE];
    _splitpath( wholePath.c_str(), _drive, _dir, _fname, _ext );
    drive=_drive;
    dir=_dir;
    fname=_fname;
    ext=_ext;
}
String readFile(String filename){
    static int maxBufSize=0x4000;
    //dLog(String("read (")+=filename+")");
    if(filename=="") return "";
    int bufSize=0;
    char* buf;

    gzFile gzfile;


    while(1){
        buf=new char[maxBufSize+2];
        if((gzfile = gzopen(filename.c_str(), "rb"))!=NULL){
            bufSize=gzread( gzfile, buf, maxBufSize);
            gzclose(gzfile);
            if(bufSize==maxBufSize){
                maxBufSize*=2;
                delete buf;
            }
            else break;
        }
        else {
            dLog("Failed to open " + filename);
            break;
        }
    }
    buf[bufSize]=0;
    return buf;
}
bool String::write(String fileName)const{
    if(g_writeGzipped){
        gzFile gzfile;
        gzfile = gzopen(fileName.c_str(), "w+b");
        if(gzfile==NULL)return false;
        gzwrite( gzfile, const_cast<char*>(c_str()), length);
        gzclose(gzfile);
    }
    else{
        FILE* file;
        file = fopen(fileName.c_str(), "w+b");
        if(file==NULL)return false;
        fwrite( c_str(), length, 1, file);
        fclose(file);
    }
    return true;
}
void setNodeWinTitleBar(Node* node){
	static String str;

	str=node->getTypeStr();
	/*
	if(node->getType()>=g_userProtoBegin){
		str="ProtoInstance ";
		str+=node->getTypeStr();
	}
	else{
		str=node->getTypeStr();
	}
*/
	str+=' ';
	str+=node->name;
	SetWindowText(g_hDlg, str.c_str());

/*
		String str=nodeLookUpStr[windowNode->getType()];
		str+=' ';
		str+=windowNode->name;
		SetWindowText(hWnd, str.c_str());
*/

/*
	str=node->name;
	str+=' ';
	if(node->getType()>=g_userProtoBegin){
		str+="ProtoInstance";
	}
	else{
		str+=node->getTypeStr();
	}
*/
}

void Drag::endDrag(const Vec2i ptCursor){
	g_drag = 0;
	ReleaseCapture();
}
void releaseCapture(){
	ReleaseCapture();
}
